@@ -46,7 +46,10 @@ GetOptions(\%args, qw(
     static has-libtommath has-libatomic_ops
     has-dyncall has-libffi pkgconfig=s
     build=s host=s big-endian jit! enable-jit
-    prefix=s bindir=s libdir=s mastdir=s make-install asan ubsan valgrind telemeh show-autovect show-autovect-failed:s),
+    prefix=s bindir=s libdir=s mastdir=s
+    relocatable make-install asan ubsan
+    valgrind telemeh show-autovect
+    show-autovect-failed:s),
 
     'no-optimize|nooptimize' => sub { $args{optimize} = 0 },
     'no-debug|nodebug' => sub { $args{debug} = 0 },
@@ -68,7 +71,7 @@ Please specify another installation target by using --prefix=PATH.
 ENOTTOCWD
 }
 
-# Override default target directories with command line argumets
+# Override default target directories with command line arguments
 my @target_dirs = qw{bindir libdir mastdir};
 for my $target (@target_dirs) {
     $config{$target} = $args{$target} if $args{$target};
@@ -91,6 +94,11 @@ if (-d '.git') {
 $args{optimize}     = 3 if not defined $args{optimize} or $args{optimize} eq "";
 $args{debug}        = 3 if defined $args{debug} and $args{debug} eq "";
 
+# Relocatability is not supported on AIX and OpenBSD.
+if ( $args{relocatable} && ($^O eq 'aix' || $^O eq 'openbsd') ) {
+    hardfail('Relocatability is not supported on ' . $^O .
+    ".\n    Leave off the --relocatable flag to do a non-relocatable build.");
+}
 
 for (qw(coverage instrument static big-endian has-libtommath has-sha has-libuv
         has-libatomic_ops asan ubsan valgrind show-vec)) {
@@ -167,6 +175,43 @@ $config{ldinstflags}  = $config{ccinstflags} unless defined $config{ldinstflags}
 
 $config{as}           = $config{cc} unless defined $config{as};
 
+# If we're in macOS, let's verify that the toolchain is consistent.
+if ($^O eq 'darwin') {
+    my $gnu_count = 0;
+    my $gnu_toolchain = (exists $args{toolchain}) && ($args{toolchain} eq 'gnu');
+
+    unless ($gnu_toolchain) {
+        # When XCode toolchain is used then force use of XCode's make if
+        # available.
+        $config{make} = '/usr/bin/make' if -x '/usr/bin/make'; 
+    }
+
+    # Here are the tools that seem to cause trouble.
+    # If you see other ones, please add them to this list.
+    my @check_tools = qw/ar cc ld/;
+    for my $tool (map { `which $_` } @config{@check_tools}) {
+        chomp $tool;
+        system "grep -b 'gnu' '$tool'"; # Apple utilities don't match `gnu`
+        if ($? == 0) {
+            $gnu_count += 1;
+        }
+    }
+
+    ## For a GNU toolchain, make sure that they're all GNU.
+    if ($gnu_toolchain && $gnu_count != scalar @check_tools) {
+        print "\nNot all tools in the toolchain are GNU. Please correct this and retry.\n"
+            . "See README.markdown for more details.\n\n";
+        exit -1;
+    }
+
+    ## Otherwise, make sure that none of them are GNU
+    elsif (!$gnu_toolchain && $gnu_count != 0) {
+        print "\nGNU tools detected, despite this not being a GNU-oriented build.\n"
+            ." Please correct this and retry. See README.markdown for more details.\n\n";
+        exit -1;
+    }
+}
+
 # Probe the compiler.
 build::probe::compiler_usability(\%config, \%defaults);
 
@@ -178,6 +223,11 @@ if ($config{cc} eq 'gcc' && !$config{can_specific_werror}) {
     $config{ccmiscflags} =~ s/^ +$//;
 }
 
+# Disable RETGUARD on OpenBSD, since it breaks the legojit.
+if ($^O eq 'openbsd' && $args{jit} && $config{cc} eq 'clang') {
+	$config{ccmiscflags} .= ' -fno-ret-protector';
+}
+
 # Set the remaining ldmiscflags. Do this after probing for gcc -Werror probe to not miss that change for the linker.
 $config{ldmiscflags}  = $config{ccmiscflags} unless defined $config{ldmiscflags};
 
@@ -212,6 +262,7 @@ if (-e "$config{pkgconfig}") {
 
 # conditionally set include dirs and install rules
 $config{cincludes} = '' unless defined $config{cincludes};
+$config{moar_cincludes} = '' unless defined $config{moar_cincludes};
 $config{lincludes} = '' unless defined $config{lincludes};
 $config{install}   = '' unless defined $config{install};
 if ($args{'has-libuv'}) {
@@ -220,12 +271,12 @@ if ($args{'has-libuv'}) {
     setup_native_library('libuv') if $config{pkgconfig_works};
 }
 else {
-    $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libuv/include'
-                        . ' ' . $defaults{ccinc} . '3rdparty/libuv/src';
-    $config{install}   .= "\t\$(MKPATH) \$(DESTDIR)\$(PREFIX)/include/libuv\n"
-                        . "\t\$(MKPATH) \$(DESTDIR)\$(PREFIX)/include/libuv/uv\n"
-                        . "\t\$(CP) 3rdparty/libuv/include/*.h \$(DESTDIR)\$(PREFIX)/include/libuv\n"
-                        . "\t\$(CP) 3rdparty/libuv/include/uv/*.h \$(DESTDIR)\$(PREFIX)/include/libuv/uv\n";
+    $config{moar_cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libuv/include'
+                             . ' ' . $defaults{ccinc} . '3rdparty/libuv/src';
+    $config{install}   .= "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/libuv\"\n"
+                        . "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/libuv/uv\"\n"
+                        . "\t\$(CP) 3rdparty/libuv/include/*.h \"\$(DESTDIR)\$(PREFIX)/include/libuv\"\n"
+                        . "\t\$(CP) 3rdparty/libuv/include/uv/*.h \"\$(DESTDIR)\$(PREFIX)/include/libuv/uv\"\n";
 }
 
 if ($args{'has-libatomic_ops'}) {
@@ -234,43 +285,45 @@ if ($args{'has-libatomic_ops'}) {
     setup_native_library('atomic_ops') if $config{pkgconfig_works};
 }
 else {
-    $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libatomicops/src';
+    $config{moar_cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libatomicops/src';
     my $lao             = '$(DESTDIR)$(PREFIX)/include/libatomic_ops';
-    $config{install}   .= "\t\$(MKPATH) $lao/atomic_ops/sysdeps/armcc\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/gcc\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/hpc\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/ibmc\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/icc\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/loadstore\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/msftc\n"
-                        . "\t\$(MKPATH) $lao/atomic_ops/sysdeps/sunc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/*.h $lao\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/*.h $lao/atomic_ops\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/*.h $lao/atomic_ops/sysdeps\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/armcc/*.h $lao/atomic_ops/sysdeps/armcc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/gcc/*.h $lao/atomic_ops/sysdeps/gcc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/hpc/*.h $lao/atomic_ops/sysdeps/hpc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/ibmc/*.h $lao/atomic_ops/sysdeps/ibmc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/icc/*.h $lao/atomic_ops/sysdeps/icc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/loadstore/*.h $lao/atomic_ops/sysdeps/loadstore\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/msftc/*.h $lao/atomic_ops/sysdeps/msftc\n"
-                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/sunc/*.h $lao/atomic_ops/sysdeps/sunc\n";
+    $config{install}   .= "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/armcc\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/gcc\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/hpc\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/ibmc\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/icc\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/loadstore\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/msftc\"\n"
+                        . "\t\$(MKPATH) \"$lao/atomic_ops/sysdeps/sunc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/*.h \"$lao\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/*.h \"$lao/atomic_ops\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/*.h \"$lao/atomic_ops/sysdeps\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/armcc/*.h \"$lao/atomic_ops/sysdeps/armcc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/gcc/*.h \"$lao/atomic_ops/sysdeps/gcc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/hpc/*.h \"$lao/atomic_ops/sysdeps/hpc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/ibmc/*.h \"$lao/atomic_ops/sysdeps/ibmc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/icc/*.h \"$lao/atomic_ops/sysdeps/icc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/loadstore/*.h \"$lao/atomic_ops/sysdeps/loadstore\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/msftc/*.h \"$lao/atomic_ops/sysdeps/msftc\"\n"
+                        . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/sunc/*.h \"$lao/atomic_ops/sysdeps/sunc\"\n";
 }
 
 if ($args{'has-libtommath'}) {
     $defaults{-thirdparty}->{tom} = undef;
     unshift @{$config{usrlibs}}, 'tommath';
-    if (index($config{cincludes}, '-I/usr/local/include') == -1) {
-        $config{cincludes} = join(' ', $config{cincludes}, '-I/usr/local/include');
-    }
-    if (index($config{lincludes}, '-L/usr/local/lib') == -1) {
-        $config{lincludes} = join(' ', $config{lincludes}, '-L/usr/local/lib');
+    if (not $config{crossconf}) {
+        if (index($config{cincludes}, '-I/usr/local/include') == -1) {
+            $config{cincludes} = join(' ', $config{cincludes}, '-I/usr/local/include');
+        }
+        if (index($config{lincludes}, '-L/usr/local/lib') == -1) {
+            $config{lincludes} = join(' ', $config{lincludes}, '-L/usr/local/lib');
+        }
     }
 }
 else {
-    $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libtommath';
-    $config{install}   .= "\t\$(MKPATH) \$(DESTDIR)\$(PREFIX)/include/libtommath\n"
-                        . "\t\$(CP) 3rdparty/libtommath/*.h \$(DESTDIR)\$(PREFIX)/include/libtommath\n";
+    $config{moar_cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libtommath';
+    $config{install}   .= "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/libtommath\"\n"
+                        . "\t\$(CP) 3rdparty/libtommath/*.h \"\$(DESTDIR)\$(PREFIX)/include/libtommath\"\n";
 }
 
 if ($args{'has-libffi'}) {
@@ -300,22 +353,33 @@ elsif ($args{'has-dyncall'}) {
     $defaults{-thirdparty}->{dcb} = undef;
     $defaults{-thirdparty}->{dl}  = undef;
     $config{nativecall_backend} = 'dyncall';
-    if (index($config{cincludes}, '-I/usr/local/include') == -1) {
-        $config{cincludes} = join(' ', $config{cincludes}, '-I/usr/local/include');
-    }
-    if (index($config{lincludes}, '-L/usr/local/lib') == -1) {
-        $config{lincludes} = join(' ', $config{lincludes}, '-L/usr/local/lib');
+    if (not $config{crossconf}) {
+        if (index($config{cincludes}, '-I/usr/local/include') == -1) {
+            $config{cincludes} = join(' ', $config{cincludes}, '-I/usr/local/include');
+        }
+        if (index($config{lincludes}, '-L/usr/local/lib') == -1) {
+            $config{lincludes} = join(' ', $config{lincludes}, '-L/usr/local/lib');
+        }
     }
 }
 else {
     $config{nativecall_backend} = 'dyncall';
-    $config{cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/dyncall/dynload'
-                        . ' ' . $defaults{ccinc} . '3rdparty/dyncall/dyncall'
-                        . ' ' . $defaults{ccinc} . '3rdparty/dyncall/dyncallback';
-    $config{install}   .= "\t\$(MKPATH) \$(DESTDIR)\$(PREFIX)/include/dyncall\n"
-                        . "\t\$(CP) 3rdparty/dyncall/dynload/*.h \$(DESTDIR)\$(PREFIX)/include/dyncall\n"
-                        . "\t\$(CP) 3rdparty/dyncall/dyncall/*.h \$(DESTDIR)\$(PREFIX)/include/dyncall\n"
-                        . "\t\$(CP) 3rdparty/dyncall/dyncallback/*.h \$(DESTDIR)\$(PREFIX)/include/dyncall\n";
+    $config{moar_cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/dyncall/dynload'
+                             . ' ' . $defaults{ccinc} . '3rdparty/dyncall/dyncall'
+                             . ' ' . $defaults{ccinc} . '3rdparty/dyncall/dyncallback';
+    $config{install}   .= "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/dyncall\"\n"
+                        . "\t\$(CP) 3rdparty/dyncall/dynload/*.h \"\$(DESTDIR)\$(PREFIX)/include/dyncall\"\n"
+                        . "\t\$(CP) 3rdparty/dyncall/dyncall/*.h \"\$(DESTDIR)\$(PREFIX)/include/dyncall\"\n"
+                        . "\t\$(CP) 3rdparty/dyncall/dyncallback/*.h \"\$(DESTDIR)\$(PREFIX)/include/dyncall\"\n";
+}
+
+if ($config{pkgconfig_works} && has_native_library('libzstd')) {
+    setup_native_library('libzstd');
+    $config{heapsnapformat} = 3;
+}
+else {
+    print "did not find libzstd; will not use heap snapshot format version 3\n";
+    $config{heapsnapformat} = 2;
 }
 
 # mangle library names
@@ -325,6 +389,7 @@ $config{ldlibs} = join ' ',
     (map { sprintf $config{ldsys}, $_; } @{$config{syslibs}});
 $config{ldlibs} = ' -lasan ' . $config{ldlibs} if $args{asan} && $^O ne 'darwin' && $config{cc} ne 'clang';
 $config{ldlibs} = ' -lubsan ' . $config{ldlibs} if $args{ubsan} and $^O ne 'darwin';
+$config{ldlibs} = $config{ldlibs} . ' -lzstd ' if $config{heapsnapformat} == 3;
 # macro defs
 $config{ccdefflags} = join ' ', map { $config{ccdef} . $_ } @{$config{defs}};
 
@@ -369,6 +434,7 @@ push @cflags, '-DDEBUG_HELPERS' if $args{debug};
 push @cflags, '-DMVM_VALGRIND_SUPPORT' if $args{valgrind};
 push @cflags, '-DHAVE_TELEMEH' if $args{telemeh};
 push @cflags, '-DWORDS_BIGENDIAN' if $config{be}; # 3rdparty/sha1 needs it and it isnt set on mips;
+push @cflags, '-DMVM_HEAPSNAPSHOT_FORMAT=' . $config{heapsnapformat};
 push @cflags, $ENV{CFLAGS} if $ENV{CFLAGS};
 push @cflags, $ENV{CPPFLAGS} if $ENV{CPPFLAGS};
 $config{cflags} = join ' ', uniq(@cflags);
@@ -377,13 +443,23 @@ $config{cflags} = join ' ', uniq(@cflags);
 my @ldflags = ($config{ldmiscflags});
 push @ldflags, $config{ldoptiflags}  if $args{optimize};
 push @ldflags, $config{lddebugflags} if $args{debug};
-push @ldflags, $config{ldinstflags}       if $args{instrument};
+push @ldflags, $config{ldinstflags}  if $args{instrument};
 push @ldflags, $config{ld_covflags}  if $args{coverage};
-push @ldflags, $config{ldrpath}           if not $args{static} and $config{prefix} ne '/usr';
-push @ldflags, '-fsanitize=address' if $args{asan};
-push @ldflags, $ENV{LDFLAGS}  if $ENV{LDFLAGS};
+if (not $args{static} and $config{prefix} ne '/usr') {
+    push @ldflags, $config{ldrpath_relocatable} if  $args{relocatable};
+    push @ldflags, $config{ldrpath}             if !$args{relocatable};
+}
+push @ldflags, '-fsanitize=address'  if $args{asan};
+push @ldflags, $ENV{LDFLAGS}         if $ENV{LDFLAGS};
 $config{ldflags} = join ' ', @ldflags;
 
+$config{moarshared} = '';
+# Switch shared lib compiler flags in relocatable case.
+if (not $args{static} and $config{prefix} ne '/usr') {
+    $config{moarshared} = $config{moarshared_relocatable}   if  $args{relocatable};
+    $config{moarshared} = $config{moarshared_norelocatable} if !$args{relocatable};
+}
+
 # setup library names
 $config{moarlib} = sprintf $config{lib}, $NAME;
 $config{moardll} = sprintf $config{dll}, $NAME;
@@ -456,13 +532,14 @@ unless (defined $config{jit_obj}) {
 
 
 if ($config{cc} eq 'cl') {
-    $config{install}   .= "\t\$(MKPATH) \$(DESTDIR)\$(PREFIX)/include/msinttypes\n"
-                        . "\t\$(CP) 3rdparty/msinttypes/*.h \$(DESTDIR)\$(PREFIX)/include/msinttypes\n";
+    $config{install}   .= "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/msinttypes\"\n"
+                        . "\t\$(CP) 3rdparty/msinttypes/*.h \"\$(DESTDIR)\$(PREFIX)/include/msinttypes\"\n";
 }
 
 build::probe::C_type_bool(\%config, \%defaults);
 build::probe::computed_goto(\%config, \%defaults);
 build::probe::pthread_yield(\%config, \%defaults);
+build::probe::check_fn_malloc_trim(\%config, \%defaults);
 if ($^O eq 'aix') {
     build::probe::numbits(\%config, \%defaults);
     $config{ldflags} = join(',', $config{ldflags}, '-bmaxdata:0x80000000')
@@ -476,7 +553,7 @@ my $order = $config{be} ? 'big endian' : 'little endian';
 print "\n", <<TERM, "\n";
         make: $config{make}
      compile: $config{cc} $config{cflags}
-    includes: $config{cincludes}
+    includes: $config{cincludes} $config{moar_cincludes}
         link: $config{ld} $config{ldflags}
         libs: $config{ldlibs}
 
@@ -600,7 +677,7 @@ available make targets.
 TERM2
 
 if (!$failed && $args{'make-install'}) {
-    system($config{make}, 'install');
+    $failed = system($config{make}, 'install') >> 8;
 }
 print $folder_to_delete if $folder_to_delete;
 exit $failed;
@@ -862,6 +939,17 @@ sub setup_native_library {
     }
 }
 
+sub has_native_library {
+    my $library = shift;
+    my $result_exists = `$config{pkgconfig} --exists $library`;
+    if ($? == 0) {
+        return 1;
+    }
+    else {
+        return 0;
+    }
+}
+
 __END__
 
 =head1 SYNOPSIS
@@ -872,7 +960,7 @@ __END__
                    [--toolchain <toolchain>] [--compiler <compiler>]
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
                    [--debug] [--optimize] [--instrument]
-                   [--static] [--prefix]
+                   [--static] [--prefix <path>] [--relocatable]
                    [--has-libtommath] [--has-sha] [--has-libuv]
                    [--has-libatomic_ops]
                    [--asan] [--ubsan] [--no-jit]
@@ -881,8 +969,8 @@ __END__
     ./Configure.pl --build <build-triple> --host <host-triple>
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
                    [--debug] [--optimize] [--instrument]
-                   [--static] [--big-endian] [--prefix]
-                   [--make-install]
+                   [--static] [--prefix <path>] [--relocatable]
+                   [--big-endian] [--make-install]
 
 =head2 Use of environment variables
 
@@ -1014,6 +1102,12 @@ builds, the byte order is auto-detected.
 Install files in subdirectory /bin, /lib and /include of the supplied path.
 The default prefix is "install" if this option is not passed.
 
+=item --relocatable
+
+Search for the libmoar library relative to the executable path.
+(On AIX and OpenBSD MoarVM can not be built relocatable, since both OS'
+miss the necessary mechanism to make this work.)
+
 =item --bindir
 
 Install executable files in the supplied path.  The default is
