@@ -497,7 +497,9 @@ static const MVMREPROps CArray_this_repr = {
         MVM_REPR_DEFAULT_SET_DIMENSIONS,
         MVM_REPR_DEFAULT_GET_ELEM_STORAGE_SPEC,
         MVM_REPR_DEFAULT_POS_AS_ATOMIC,
-        MVM_REPR_DEFAULT_POS_AS_ATOMIC_MULTIDIM
+        MVM_REPR_DEFAULT_POS_AS_ATOMIC_MULTIDIM,
+        MVM_REPR_DEFAULT_POS_WRITE_BUF,
+        MVM_REPR_DEFAULT_POS_READ_BUF
     },    /* pos_funcs */
     MVM_REPR_DEFAULT_ASS_FUNCS,
     elems,
