@@ -16,17 +16,11 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
 
     MVMint32 mro_idx = MVM_repr_elems(tc, mro);
 
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&st);
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&mro);
-
     flat_list = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_array_type);
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&flat_list);
 
     class_list = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_array_type);
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&class_list);
 
     attr_map_list = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_array_type);
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&attr_map_list);
 
     /* Walk through the parents list. */
     while (mro_idx)
@@ -45,9 +39,7 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
             MVMObject *attr_map = NULL;
 
             if (MVM_iter_istrue(tc, attr_iter)) {
-                MVM_gc_root_temp_push(tc, (MVMCollectable **)&attr_iter);
                 attr_map = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_hash_type);
-                MVM_gc_root_temp_push(tc, (MVMCollectable **)&attr_map);
             }
 
             while (MVM_iter_istrue(tc, attr_iter)) {
@@ -72,22 +64,17 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
                 MVM_repr_push_o(tc, flat_list, attr);
             }
 
-            if (attr_map) {
-                MVM_gc_root_temp_pop_n(tc, 2);
-            }
-
             /* Add to class list and map list. */
             MVM_repr_push_o(tc, class_list, current_class);
             MVM_repr_push_o(tc, attr_map_list, attr_map);
         }
         else {
+            MVM_gc_allocate_gen2_default_clear(tc);
             MVM_exception_throw_adhoc(tc,
                 "CPPStruct representation does not support multiple inheritance");
         }
     }
 
-    MVM_gc_root_temp_pop_n(tc, 5);
-
     /* We can now form the name map. */
     num_classes = MVM_repr_elems(tc, class_list);
     result = (MVMCPPStructNameMap *) MVM_malloc(sizeof(MVMCPPStructNameMap) * (1 + num_classes));
@@ -165,6 +152,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
             MVMint32   align = ALIGNOF(void *);
 
             if (num_dimensions > 1) {
+                MVM_gc_allocate_gen2_default_clear(tc);
                 MVM_exception_throw_adhoc(tc,
                     "Only one dimensions supported in CPPStruct attribute");
             }
@@ -218,6 +206,11 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                     if (inlined) {
                         MVMCArrayREPRData *carray_repr_data = (MVMCArrayREPRData *)STABLE(type)->REPR_data;
+                        if (!carray_repr_data) {
+                            MVM_gc_allocate_gen2_default_clear(tc);
+                            MVM_exception_throw_adhoc(tc,
+                                "CPPStruct: can't inline a CArray attribute before its type's definition");
+                        }
                         bits                                = carray_repr_data->elem_size * 8;
                         repr_data->attribute_locations[i]  |= MVM_CSTRUCT_ATTR_INLINED;
 
@@ -245,9 +238,14 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                     if (inlined) {
                         MVMCStructREPRData *cstruct_repr_data = (MVMCStructREPRData *)STABLE(type)->REPR_data;
+                        if (!cstruct_repr_data) {
+                            MVM_gc_allocate_gen2_default_clear(tc);
+                            MVM_exception_throw_adhoc(tc,
+                                "CPPStruct: can't inline a CStruct attribute before its type's definition");
+                        }
                         bits                                  = cstruct_repr_data->struct_size * 8;
                         align                                 = cstruct_repr_data->struct_align;
-                        repr_data->attribute_locations[i]    |= MVM_CSTRUCT_ATTR_INLINED;
+                        repr_data->attribute_locations[i]    |= MVM_CPPSTRUCT_ATTR_INLINED;
                     }
                 }
                 else if (type_id == MVM_REPR_ID_MVMCPPStruct) {
@@ -257,6 +255,11 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                     if (inlined) {
                         MVMCPPStructREPRData *cppstruct_repr_data = (MVMCPPStructREPRData *)STABLE(type)->REPR_data;
+                        if (!cppstruct_repr_data) {
+                            MVM_gc_allocate_gen2_default_clear(tc);
+                            MVM_exception_throw_adhoc(tc,
+                                "CPPStruct: can't inline a CPPStruct attribute before its type's definition");
+                        }
                         bits                                      = cppstruct_repr_data->struct_size * 8;
                         align                                     = cppstruct_repr_data->struct_align;
                         repr_data->attribute_locations[i]        |= MVM_CPPSTRUCT_ATTR_INLINED;
@@ -265,13 +268,18 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                 else if (type_id == MVM_REPR_ID_MVMCUnion) {
                     /* It's a CUnion. */
                     repr_data->num_child_objs++;
-                    repr_data->attribute_locations[i] = (cur_obj_attr++ << MVM_CSTRUCT_ATTR_SHIFT) | MVM_CSTRUCT_ATTR_CUNION;
+                    repr_data->attribute_locations[i] = (cur_obj_attr++ << MVM_CPPSTRUCT_ATTR_SHIFT) | MVM_CPPSTRUCT_ATTR_CUNION;
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                     if (inlined) {
                         MVMCUnionREPRData *cunion_repr_data = (MVMCUnionREPRData *)STABLE(type)->REPR_data;
+                        if (!cunion_repr_data) {
+                            MVM_gc_allocate_gen2_default_clear(tc);
+                            MVM_exception_throw_adhoc(tc,
+                                "CPPStruct: can't inline a CUnion attribute before its type's definition");
+                        }
                         bits                                = cunion_repr_data->struct_size * 8;
                         align                               = cunion_repr_data->struct_align;
-                        repr_data->attribute_locations[i]  |= MVM_CSTRUCT_ATTR_INLINED;
+                        repr_data->attribute_locations[i]  |= MVM_CPPSTRUCT_ATTR_INLINED;
                     }
                 }
                 else if (type_id == MVM_REPR_ID_MVMCPointer) {
@@ -281,6 +289,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                 }
                 else {
+                    MVM_gc_allocate_gen2_default_clear(tc);
                     MVM_exception_throw_adhoc(tc,
                         "CPPStruct representation only handles attributes of type:\n"
                         "  (u)int8, (u)int16, (u)int32, (u)int64, (u)long, (u)longlong, num32, num64, (s)size_t, bool, Str\n"
@@ -288,11 +297,13 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMSTable *st,
                 }
             }
             else {
+                MVM_gc_allocate_gen2_default_clear(tc);
                 MVM_exception_throw_adhoc(tc,
                     "CPPStruct representation requires the types of all attributes to be specified");
             }
 
             if (bits % 8) {
+                MVM_gc_allocate_gen2_default_clear(tc);
                  MVM_exception_throw_adhoc(tc,
                     "CPPStruct only supports native types that are a multiple of 8 bits wide (was passed: %"PRId32")", bits);
             }
@@ -370,7 +381,9 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *repr_info) {
     /* Compute allocation strategy. */
     MVMCPPStructREPRData *repr_data = MVM_calloc(1, sizeof(MVMCPPStructREPRData));
     MVMObject *attr_info = MVM_repr_at_key_o(tc, repr_info, tc->instance->str_consts.attribute);
+    MVM_gc_allocate_gen2_default_set(tc);
     compute_allocation_strategy(tc, st, attr_info, repr_data);
+    MVM_gc_allocate_gen2_default_clear(tc);
     st->REPR_data = repr_data;
 }
 
@@ -448,42 +461,43 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                     /* No cached object. */
                     void *cobj = get_ptr_at_offset(body->cppstruct, repr_data->struct_offsets[slot]);
                     if (cobj) {
-                        MVMObject **child_objs = body->child_objs;
-                        if (type == MVM_CPPSTRUCT_ATTR_CARRAY) {
-                            obj = MVM_nativecall_make_carray(tc, typeobj, cobj);
-                        }
-                        else if(type == MVM_CPPSTRUCT_ATTR_CSTRUCT) {
-                            if (repr_data->attribute_locations[slot] & MVM_CPPSTRUCT_ATTR_INLINED)
-                                obj = MVM_nativecall_make_cstruct(tc, typeobj,
-                                    (char *)body->cppstruct + repr_data->struct_offsets[slot]);
-                            else
-                                obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
-                        }
-                        else if(type == MVM_CPPSTRUCT_ATTR_CPPSTRUCT) {
-                            if (repr_data->attribute_locations[slot] & MVM_CPPSTRUCT_ATTR_INLINED)
-                                obj = MVM_nativecall_make_cppstruct(tc, typeobj,
-                                    (char *)body->cppstruct + repr_data->struct_offsets[slot]);
-                            else
-                                obj = MVM_nativecall_make_cppstruct(tc, typeobj, cobj);
-                        }
-                        else if(type == MVM_CPPSTRUCT_ATTR_CUNION) {
-                            if (repr_data->attribute_locations[slot] & MVM_CPPSTRUCT_ATTR_INLINED)
-                                obj = MVM_nativecall_make_cunion(tc, typeobj,
-                                    (char *)body->cppstruct + repr_data->struct_offsets[slot]);
-                            else
-                                obj = MVM_nativecall_make_cunion(tc, typeobj, cobj);
-                        }
-                        else if(type == MVM_CPPSTRUCT_ATTR_CPTR) {
-                            obj = MVM_nativecall_make_cpointer(tc, typeobj, cobj);
-                        }
-                        else if(type == MVM_CPPSTRUCT_ATTR_STRING) {
-                            MVMROOT(tc, typeobj, {
-                                MVMString *str = MVM_string_utf8_decode(tc, tc->instance->VMString,
-                                    cobj, strlen(cobj));
-                                obj = MVM_repr_box_str(tc, typeobj, str);
-                            });
-                        }
-                        child_objs[real_slot] = obj;
+                        MVMROOT(tc, root, {
+                            if (type == MVM_CPPSTRUCT_ATTR_CARRAY) {
+                                obj = MVM_nativecall_make_carray(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CPPSTRUCT_ATTR_CSTRUCT) {
+                                if (repr_data->attribute_locations[slot] & MVM_CPPSTRUCT_ATTR_INLINED)
+                                    obj = MVM_nativecall_make_cstruct(tc, typeobj,
+                                        (char *)body->cppstruct + repr_data->struct_offsets[slot]);
+                                else
+                                    obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CPPSTRUCT_ATTR_CPPSTRUCT) {
+                                if (repr_data->attribute_locations[slot] & MVM_CPPSTRUCT_ATTR_INLINED)
+                                    obj = MVM_nativecall_make_cppstruct(tc, typeobj,
+                                        (char *)body->cppstruct + repr_data->struct_offsets[slot]);
+                                else
+                                    obj = MVM_nativecall_make_cppstruct(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CPPSTRUCT_ATTR_CUNION) {
+                                if (repr_data->attribute_locations[slot] & MVM_CPPSTRUCT_ATTR_INLINED)
+                                    obj = MVM_nativecall_make_cunion(tc, typeobj,
+                                        (char *)body->cppstruct + repr_data->struct_offsets[slot]);
+                                else
+                                    obj = MVM_nativecall_make_cunion(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CPPSTRUCT_ATTR_CPTR) {
+                                obj = MVM_nativecall_make_cpointer(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CPPSTRUCT_ATTR_STRING) {
+                                MVMROOT(tc, typeobj, {
+                                    MVMString *str = MVM_string_utf8_decode(tc, tc->instance->VMString,
+                                        cobj, strlen(cobj));
+                                    obj = MVM_repr_box_str(tc, typeobj, str);
+                                });
+                            }
+                        });
+                        MVM_ASSIGN_REF(tc, &(root->header), body->child_objs[real_slot], obj);
                     }
                     else {
                         obj = typeobj;
