@@ -1035,8 +1035,12 @@ static void deserialize_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMSerial
         for (i = 0; i < repr_data->num_attributes; i++) {
             MVMuint16 repr_id = MVM_serialization_read_int(tc, reader);
             MVMuint16 slot = MVM_serialization_read_int(tc, reader);
-            if (repr_id)
+            if (slot > repr_data->num_attributes)
+                MVM_exception_throw_adhoc(tc, "Serialization error: P6opaque's unbox slot out of range (slot %d > %d attributes).", slot, repr_data->num_attributes);
+            if (repr_id < MVM_REPR_MAX_COUNT)
                 repr_data->unbox_slots[repr_id] = slot;
+            else
+                MVM_exception_throw_adhoc(tc, "Serialization error: P6opaque's unbox slot repr id out of range.");
         }
     } else {
         repr_data->unbox_slots = NULL;
@@ -1183,11 +1187,22 @@ static void change_type(MVMThreadContext *tc, MVMObject *obj, MVMObject *new_typ
         MVM_exception_throw_adhoc(tc,
             "Cannot change the type of a %s type object", MVM_6model_get_debug_name(tc, obj));
 
+    /* If the type is already the same, nothing to do. */
+    if (STABLE(obj) == STABLE(new_type))
+        return;
+
     /* Ensure that the REPR of the new type is also P6opaque. */
     if (REPR(new_type)->ID != REPR(obj)->ID)
         MVM_exception_throw_adhoc(tc,
             "New type for %s must have a matching representation (P6opaque vs %s)", MVM_6model_get_debug_name(tc, obj), REPR(new_type)->name);
 
+    /* Ensure the target type is a mixin type. */
+    if (!STABLE(new_type)->is_mixin_type)
+        MVM_exception_throw_adhoc(tc,
+            "New type %s for %s is not a mixin type",
+            MVM_6model_get_debug_name(tc, new_type),
+            MVM_6model_get_debug_name(tc, obj));
+
     /* Ensure the MRO prefixes match up. */
     cur_map_entry = cur_repr_data->name_to_index_mapping;
     new_map_entry = new_repr_data->name_to_index_mapping;
@@ -1461,19 +1476,26 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             && obj_facts->flags & MVM_SPESH_FACT_CONCRETE) {
             MVMint64 slot = try_get_slot(tc, repr_data, ch_facts->type, name);
             if (slot >= 0 && !repr_data->flattened_stables[slot]) {
+                MVMint8 mixin = st->is_mixin_type;
                 add_slot_name_comment(tc, g, ins, name, ch_facts, st);
                 if (repr_data->auto_viv_values && repr_data->auto_viv_values[slot]) {
                     MVMObject *av_value = repr_data->auto_viv_values[slot];
                     if (IS_CONCRETE(av_value)) {
-                        ins->info = MVM_op_get_op(MVM_OP_sp_p6ogetvc_o);
+                        ins->info = MVM_op_get_op(mixin
+                                ? MVM_OP_sp_p6ogetvc_o
+                                : MVM_OP_sp_getvc_o);
                     }
                     else {
-                        ins->info = MVM_op_get_op(MVM_OP_sp_p6ogetvt_o);
+                        ins->info = MVM_op_get_op(mixin
+                                ? MVM_OP_sp_p6ogetvt_o
+                                : MVM_OP_sp_getvt_o);
                     }
                     if (opcode == MVM_OP_getattrs_o)
                         MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[3], ins);
                     MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
-                    ins->operands[2].lit_i16 = repr_data->attribute_offsets[slot];
+                    ins->operands[2].lit_i16 = mixin
+                        ? repr_data->attribute_offsets[slot]
+                        : sizeof(MVMObject) + repr_data->attribute_offsets[slot];
                     ins->operands[3].lit_i16 = MVM_spesh_add_spesh_slot(tc, g,
                         (MVMCollectable *)av_value);
                 }
@@ -1481,9 +1503,15 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
                     if (opcode == MVM_OP_getattrs_o)
                         MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[3], ins);
                     MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
-                    ins->info = MVM_op_get_op(MVM_OP_sp_p6oget_o);
-                    ins->operands[2].lit_i16 = repr_data->attribute_offsets[slot];
-                    MVM_spesh_manipulate_remove_handler_successors(tc, bb);
+                    if (mixin) {
+                        ins->info = MVM_op_get_op(MVM_OP_sp_p6oget_o);
+                        ins->operands[2].lit_i16 = repr_data->attribute_offsets[slot];
+                    }
+                    else {
+                        ins->info = MVM_op_get_op(MVM_OP_sp_get_o);
+                        ins->operands[2].lit_i16 = sizeof(MVMObject) +
+                            repr_data->attribute_offsets[slot];
+                    }
                 }
             }
         }
@@ -1501,11 +1529,15 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
                 MVMSTable      *flat_st = repr_data->flattened_stables[slot];
                 const MVMStorageSpec *flat_ss = flat_st->REPR->get_storage_spec(tc, flat_st);
                 add_slot_name_comment(tc, g, ins, name, ch_facts, st);
-                if (flat_st->REPR->ID == MVM_REPR_ID_P6int && flat_ss->bits == 64) {
+                if (flat_st->REPR->ID == MVM_REPR_ID_P6int &&
+                        (flat_ss->bits == 64 || flat_ss->bits == 32)) {
                     if (opcode == MVM_OP_getattrs_i)
                         MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[3], ins);
                     MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
-                    ins->info = MVM_op_get_op(MVM_OP_sp_p6oget_i);
+                    ins->info = MVM_op_get_op(
+                            flat_ss->bits == 64 ? MVM_OP_sp_p6oget_i
+                                                : MVM_OP_sp_p6oget_i32
+                            );
                     ins->operands[2].lit_i16 = repr_data->attribute_offsets[slot];
                 }
             }
@@ -1566,12 +1598,19 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             && obj_facts->flags & MVM_SPESH_FACT_CONCRETE) {
             MVMint64 slot = try_get_slot(tc, repr_data, ch_facts->type, name);
             if (slot >= 0 && !repr_data->flattened_stables[slot]) {
+                MVMint8 mixin = st->is_mixin_type;
                 add_slot_name_comment(tc, g, ins, name, ch_facts, st);
                 if (opcode == MVM_OP_bindattrs_o)
                     MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
                 MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[1], ins);
-                ins->info = MVM_op_get_op(MVM_OP_sp_p6obind_o);
-                ins->operands[1].lit_i16 = repr_data->attribute_offsets[slot];
+                if (mixin) {
+                    ins->info = MVM_op_get_op(MVM_OP_sp_p6obind_o);
+                    ins->operands[1].lit_i16 = repr_data->attribute_offsets[slot];
+                }
+                else {
+                    ins->info = MVM_op_get_op(MVM_OP_sp_bind_o);
+                    ins->operands[1].lit_i16 = sizeof(MVMObject) + repr_data->attribute_offsets[slot];
+                }
                 ins->operands[2] = ins->operands[3];
             }
         }
@@ -1588,12 +1627,16 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             if (slot >= 0 && repr_data->flattened_stables[slot]) {
                 MVMSTable      *flat_st = repr_data->flattened_stables[slot];
                 const MVMStorageSpec *flat_ss = flat_st->REPR->get_storage_spec(tc, flat_st);
-                if (flat_st->REPR->ID == MVM_REPR_ID_P6int && flat_ss->bits == 64) {
+                if (flat_st->REPR->ID == MVM_REPR_ID_P6int &&
+                        (flat_ss->bits == 64 || flat_ss->bits == 32)) {
                     add_slot_name_comment(tc, g, ins, name, ch_facts, st);
                     if (opcode == MVM_OP_bindattrs_i)
                         MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
                     MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[1], ins);
-                    ins->info = MVM_op_get_op(MVM_OP_sp_p6obind_i);
+                    ins->info = MVM_op_get_op(
+                            flat_ss->bits == 64 ? MVM_OP_sp_p6obind_i
+                                                : MVM_OP_sp_p6obind_i32
+                            );
                     ins->operands[1].lit_i16 = repr_data->attribute_offsets[slot];
                     ins->operands[2] = ins->operands[3];
                 }
@@ -1783,6 +1826,10 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             }
         }
         break;
+    default:
+        MVM_spesh_graph_add_comment(tc, g, ins, "reprop %s unsupported in P6Opaque %s",
+                ins->info->name,
+                MVM_6model_get_stable_debug_name(tc, st));
     }
 }
 
@@ -1824,13 +1871,15 @@ static const MVMREPROps P6opaque_this_repr = {
         shift,
         oslice,
         osplice,
-        MVM_REPR_DEFAULT_AT_POS_MULTIDIM,
-        MVM_REPR_DEFAULT_BIND_POS_MULTIDIM,
-        MVM_REPR_DEFAULT_DIMENSIONS,
-        MVM_REPR_DEFAULT_SET_DIMENSIONS,
+        MVM_REPR_DEFAULT_AT_POS_MULTIDIM_NO_MULTIDIM,
+        MVM_REPR_DEFAULT_BIND_POS_MULTIDIM_NO_MULTIDIM,
+        MVM_REPR_DEFAULT_DIMENSIONS_NO_MULTIDIM,
+        MVM_REPR_DEFAULT_SET_DIMENSIONS_NO_MULTIDIM,
         MVM_REPR_DEFAULT_GET_ELEM_STORAGE_SPEC,
         MVM_REPR_DEFAULT_POS_AS_ATOMIC,
-        MVM_REPR_DEFAULT_POS_AS_ATOMIC_MULTIDIM
+        MVM_REPR_DEFAULT_POS_AS_ATOMIC_MULTIDIM,
+        MVM_REPR_DEFAULT_POS_WRITE_BUF,
+        MVM_REPR_DEFAULT_POS_READ_BUF,
     },    /* pos_funcs */
     {
         at_key,
@@ -1880,6 +1929,16 @@ MVMuint16 MVM_p6opaque_get_bigint_offset(MVMThreadContext *tc, MVMSTable *st) {
     return 0;
 }
 
+/* Gets the attribute index given we know the slot offset. */
+MVMuint32 MVM_p6opaque_offset_to_attr_idx(MVMThreadContext *tc, MVMObject *type, size_t offset) {
+    MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)type->st->REPR_data;
+    MVMuint32 i;
+    for (i = 0; i < repr_data->num_attributes; i++)
+        if (repr_data->attribute_offsets[i] == offset)
+            return i;
+    MVM_oops(tc, "P6opaque: slot offset not found");
+}
+
 #ifdef DEBUG_HELPERS
 /* This is meant to be called in a debugging session and not used anywhere else.
  * Plese don't delete. */
@@ -1917,7 +1976,7 @@ static void dump_p6opaque(MVMThreadContext *tc, MVMObject *obj, int nested) {
                         MVMSTable * const attr_st = repr_data->flattened_stables[slot];
                         if (attr_st == NULL) {
                             MVMObject *value = get_obj_at_offset(data, offset);
-                            if (value != NULL) {
+                            if (value != NULL && REPR(value)->ID == MVM_REPR_ID_P6opaque) {
                                 fprintf(stderr, "=");
                                 dump_p6opaque(tc, value, 1);
                             }
