@@ -222,7 +222,7 @@ MVMString * MVM_coerce_i_s(MVMThreadContext *tc, MVMint64 i) {
         return result;
     }
     else {
-        MVM_exception_throw_adhoc(tc, "Could not stringify integer");
+        MVM_exception_throw_adhoc(tc, "Could not stringify integer (%"PRId64")", i);
     }
 }
 
@@ -248,7 +248,7 @@ MVMString * MVM_coerce_u_s(MVMThreadContext *tc, MVMuint64 i) {
         return result;
     }
     else {
-        MVM_exception_throw_adhoc(tc, "Could not stringify integer");
+        MVM_exception_throw_adhoc(tc, "Could not stringify integer (%"PRIu64")", i);
     }
 }
 
@@ -265,7 +265,7 @@ MVMString * MVM_coerce_n_s(MVMThreadContext *tc, MVMnum64 n) {
     else {
         char buf[64];
         if (dtoa_grisu3(n, buf, 64) < 0)
-            MVM_exception_throw_adhoc(tc, "Could not stringify number");
+            MVM_exception_throw_adhoc(tc, "Could not stringify number (%f)", n);
         else {
             MVMStringIndex len = strlen(buf);
             MVMGrapheme8 *blob = MVM_malloc(len);
@@ -321,7 +321,7 @@ void MVM_coerce_smart_stringify(MVMThreadContext *tc, MVMObject *obj, MVMRegiste
         else if (ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_NUM)
             res_reg->s = MVM_coerce_n_s(tc, REPR(obj)->box_funcs.get_num(tc, STABLE(obj), obj, OBJECT_BODY(obj)));
         else
-            MVM_exception_throw_adhoc(tc, "cannot stringify this");
+            MVM_exception_throw_adhoc(tc, "Cannot stringify this object of type %s (%s)", REPR(obj)->name, MVM_6model_get_stable_debug_name(tc, STABLE(obj)));
     }
 }
 
@@ -376,7 +376,54 @@ void MVM_coerce_smart_numify(MVMThreadContext *tc, MVMObject *obj, MVMRegister *
         else if (REPR(obj)->ID == MVM_REPR_ID_MVMHash)
             res_reg->n64 = (MVMnum64)REPR(obj)->elems(tc, STABLE(obj), obj, OBJECT_BODY(obj));
         else
-            MVM_exception_throw_adhoc(tc, "cannot numify this");
+            MVM_exception_throw_adhoc(tc, "Cannot numify this object of type %s (%s)", REPR(obj)->name, MVM_6model_get_stable_debug_name(tc, STABLE(obj)));
+    }
+}
+
+void MVM_coerce_smart_intify(MVMThreadContext *tc, MVMObject *obj, MVMRegister *res_reg) {
+    MVMObject *intmeth;
+
+    /* Handle null case. */
+    if (MVM_is_null(tc, obj)) {
+        res_reg->i64 = 0;
+    }
+
+    /* Check if there is an Int method. */
+    MVMROOT(tc, obj, {
+        intmeth = MVM_6model_find_method_cache_only(tc, obj,
+            tc->instance->str_consts.Int);
+    });
+
+    if (!MVM_is_null(tc, intmeth)) {
+        /* We need to do the invocation; just set it up with our result reg as
+         * the one for the call. */
+        MVMObject *code = MVM_frame_find_invokee(tc, intmeth, NULL);
+        MVMCallsite *inv_arg_callsite = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_INV_ARG);
+
+        MVM_args_setup_thunk(tc, res_reg, MVM_RETURN_INT, inv_arg_callsite);
+        tc->cur_frame->args[0].o = obj;
+        STABLE(code)->invoke(tc, code, inv_arg_callsite, tc->cur_frame->args);
+        return;
+    }
+
+    /* Otherwise, guess something appropriate. */
+    if (!IS_CONCRETE(obj)) {
+        res_reg->i64 = 0;
+    }
+    else {
+        const MVMStorageSpec *ss = REPR(obj)->get_storage_spec(tc, STABLE(obj));
+        if (ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_INT)
+            res_reg->i64 = REPR(obj)->box_funcs.get_int(tc, STABLE(obj), obj, OBJECT_BODY(obj));
+        else if (ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_NUM)
+            res_reg->i64 = (MVMint64)REPR(obj)->box_funcs.get_num(tc, STABLE(obj), obj, OBJECT_BODY(obj));
+        else if (ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_STR)
+            res_reg->i64 = MVM_coerce_s_i(tc, REPR(obj)->box_funcs.get_str(tc, STABLE(obj), obj, OBJECT_BODY(obj)));
+        else if (REPR(obj)->ID == MVM_REPR_ID_VMArray)
+            res_reg->i64 = REPR(obj)->elems(tc, STABLE(obj), obj, OBJECT_BODY(obj));
+        else if (REPR(obj)->ID == MVM_REPR_ID_MVMHash)
+            res_reg->i64 = REPR(obj)->elems(tc, STABLE(obj), obj, OBJECT_BODY(obj));
+        else
+            MVM_exception_throw_adhoc(tc, "Cannot intify this object of type %s (%s)", REPR(obj)->name, MVM_6model_get_stable_debug_name(tc, STABLE(obj)));
     }
 }
 
@@ -400,7 +447,7 @@ MVMint64 MVM_coerce_simple_intify(MVMThreadContext *tc, MVMObject *obj) {
         else if (REPR(obj)->ID == MVM_REPR_ID_MVMHash)
             return REPR(obj)->elems(tc, STABLE(obj), obj, OBJECT_BODY(obj));
         else
-            MVM_exception_throw_adhoc(tc, "cannot intify this");
+            MVM_exception_throw_adhoc(tc, "Cannot intify this object of type %s (%s)", REPR(obj)->name, MVM_6model_get_stable_debug_name(tc, STABLE(obj)));
     }
 }
 
