@@ -360,6 +360,7 @@ static void run_handler(MVMThreadContext *tc, LocatedHandler lh, MVMObject *ex_o
         /* Create active handler record. */
         MVMActiveHandler *ah = MVM_malloc(sizeof(MVMActiveHandler));
         MVMFrame *cur_frame = tc->cur_frame;
+        MVMFrame *pres_frame;
         MVMObject *handler_code;
 
         /* Ensure we have an exception object. */
@@ -371,6 +372,14 @@ static void run_handler(MVMThreadContext *tc, LocatedHandler lh, MVMObject *ex_o
             MVM_ASSIGN_REF(tc, &(ex_obj->header), ((MVMException *)ex_obj)->body.payload, payload);
         }
 
+        /* Preserve the frame caller chain, so we can do backtraces. */
+        pres_frame = ((MVMException *)ex_obj)->body.origin;
+        while (pres_frame) {
+            MVMFrameExtra *extra = MVM_frame_extra(tc, pres_frame);
+            extra->caller_info_needed = 1;
+            pres_frame = pres_frame->caller;
+        }
+
         /* Find frame to invoke. */
         handler_code = MVM_frame_find_invokee(tc, lh.frame->work[lh.handler->block_reg].o, NULL);
 
@@ -590,7 +599,9 @@ MVMObject * MVM_exception_backtrace_strings(MVMThreadContext *tc, MVMObject *ex_
     else
         MVM_exception_throw_adhoc(tc, "Op 'backtracestrings' needs an exception object");
 
-    arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
+    MVMROOT(tc, ex, {
+        arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
+    });
     cur_frame = ex->body.origin;
 
     MVMROOT2(tc, arr, cur_frame, {
