@@ -81,6 +81,9 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
         MVM_ASSIGN_REF(tc, &(static_frame->common.header), static_frame_body->spesh,
             MVM_repr_alloc_init(tc, tc->instance->StaticFrameSpesh));
         MVM_gc_allocate_gen2_default_clear(tc);
+
+        /* We now have at least instrumentation level 1. */
+        static_frame->body.instrumentation_level = 1;
     }
 
     /* Unlock, now we're finished. */
@@ -92,27 +95,39 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
  * verify it. It may also be because we need to instrument the code for
  * profiling. */
 static void instrumentation_level_barrier(MVMThreadContext *tc, MVMStaticFrame *static_frame) {
-    /* Prepare and verify if needed. */
-    if (static_frame->body.instrumentation_level == 0)
-        prepare_and_verify_static_frame(tc, static_frame);
+    MVMCompUnit *cu = static_frame->body.cu;
+    MVMROOT2(tc, static_frame, cu, {
+        /* Obtain mutex, so we don't end up with instrumentation races. */
+        MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
 
-    /* Mark frame as being at the current instrumentation level. */
-    static_frame->body.instrumentation_level = tc->instance->instrumentation_level;
+        /* Prepare and verify if needed. */
+        if (static_frame->body.instrumentation_level == 0)
+            prepare_and_verify_static_frame(tc, static_frame);
 
-    /* Add profiling instrumentation if needed. */
-    if (tc->instance->profiling)
-        MVM_profile_instrument(tc, static_frame);
-    else if (tc->instance->cross_thread_write_logging)
-        MVM_cross_thread_write_instrument(tc, static_frame);
-    else if (tc->instance->coverage_logging)
-        MVM_line_coverage_instrument(tc, static_frame);
-    else if (tc->instance->debugserver)
-        MVM_breakpoint_instrument(tc, static_frame);
-    else
-        /* XXX uninstrumenting is currently turned off, due to multithreading
-         * woes. If you add an instrumentation that has to be "turned off"
-         * again at some point, a solution for this problem must be found. */
-        MVM_profile_ensure_uninstrumented(tc, static_frame);
+        /* Re-check instrumentation level in case of races. */
+        if (static_frame->body.instrumentation_level != tc->instance->instrumentation_level) {
+            /* Add profiling instrumentation if needed. */
+            if (tc->instance->profiling)
+                MVM_profile_instrument(tc, static_frame);
+            else if (tc->instance->cross_thread_write_logging)
+                MVM_cross_thread_write_instrument(tc, static_frame);
+            else if (tc->instance->coverage_logging)
+                MVM_line_coverage_instrument(tc, static_frame);
+            else if (tc->instance->debugserver)
+                MVM_breakpoint_instrument(tc, static_frame);
+            else
+                /* XXX uninstrumenting is currently turned off, due to multithreading
+                 * woes. If you add an instrumentation that has to be "turned off"
+                 * again at some point, a solution for this problem must be found. */
+                MVM_profile_ensure_uninstrumented(tc, static_frame);
+
+            /* Mark frame as being at the current instrumentation level. */
+            static_frame->body.instrumentation_level = tc->instance->instrumentation_level;
+        }
+
+        /* Release the lock. */
+        MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
+    });
 }
 
 /* Called when the GC destroys a frame. Since the frame may have been alive as
@@ -158,7 +173,7 @@ static MVMFrame * create_context_only(MVMThreadContext *tc, MVMStaticFrame *stat
      * is vivified to prevent the clone (which is what creates the correct
      * BEGIN/INIT semantics). */
     if (static_frame->body.env_size) {
-        frame->env = MVM_fixed_size_alloc(tc, tc->instance->fsa, static_frame->body.env_size);
+        frame->env = MVM_fixed_size_alloc_zeroed(tc, tc->instance->fsa, static_frame->body.env_size);
         frame->allocd_env = static_frame->body.env_size;
         if (autoclose) {
             MVMROOT2(tc, frame, static_frame, {
@@ -343,17 +358,14 @@ static MVMFrame * allocate_heap_frame(MVMThreadContext *tc, MVMStaticFrame *stat
     }
     work_size = spesh_cand ? spesh_cand->work_size : static_frame_body->work_size;
     if (work_size) {
-        /* Fill up all object registers with a pointer to our VMNull object */
-        if (spesh_cand && spesh_cand->local_types) {
-            MVMuint32 num_locals = spesh_cand->num_locals;
-            MVMuint16 *local_types = spesh_cand->local_types;
-            MVMuint32 i;
+        if (spesh_cand) {
+            /* We make sure to initialize anything that can be read before
+             * being initialized with VMNull at frame entyr, so we can
+             * just provide zeroed memory here. */
             frame->work = MVM_fixed_size_alloc_zeroed(tc, tc->instance->fsa, work_size);
-            for (i = 0; i < num_locals; i++)
-                if (local_types[i] == MVM_reg_obj)
-                    frame->work[i].o = tc->instance->VMNull;
         }
         else {
+            /* Copy in a stub frame work area. */
             frame->work = MVM_fixed_size_alloc(tc, tc->instance->fsa, work_size);
             memcpy(frame->work, static_frame_body->work_initial,
                 sizeof(MVMRegister) * static_frame_body->num_locals);
@@ -397,16 +409,18 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
     /* Ensure we have an outer if needed. This is done ahead of allocating the
      * new frame, since an autoclose will force the callstack on to the heap. */
     if (outer) {
-        /* We were provided with an outer frame and it will already have had
-         * its reference count incremented; just ensure that it is based on the
+        /* We were provided with an outer frame. Ensure that it is based on the
          * correct static frame (compare on bytecode address to cope with
          * nqp::freshcoderef). */
-        if (MVM_UNLIKELY(outer->static_info->body.orig_bytecode != static_frame->body.outer->body.orig_bytecode)) {
+        if (MVM_UNLIKELY(static_frame->body.outer == 0 || outer->static_info->body.orig_bytecode != static_frame->body.outer->body.orig_bytecode)) {
             char *frame_cuuid = MVM_string_utf8_encode_C_string(tc, static_frame->body.cuuid);
             char *frame_name;
             char *outer_cuuid = MVM_string_utf8_encode_C_string(tc, outer->static_info->body.cuuid);
             char *outer_name;
-            char *frame_outer_cuuid = MVM_string_utf8_encode_C_string(tc, static_frame->body.outer->body.cuuid);
+            char *frame_outer_cuuid = MVM_string_utf8_encode_C_string(tc,
+                    static_frame->body.outer
+                        ? static_frame->body.outer->body.cuuid
+                        : tc->instance->str_consts.empty);
             char *frame_outer_name;
 
             char *waste[7] = { frame_cuuid, outer_cuuid, frame_outer_cuuid, NULL, NULL, NULL, NULL };
@@ -428,7 +442,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
                 outer_name = "<anonymous static frame>";
             }
 
-            if (static_frame->body.outer->body.name) {
+            if (static_frame->body.outer && static_frame->body.outer->body.name) {
                 frame_outer_name = MVM_string_utf8_encode_C_string(tc, static_frame->body.outer->body.name);
                 waste[waste_counter++] = frame_outer_name;
             }
@@ -628,7 +642,7 @@ MVMFrame * MVM_frame_move_to_heap(MVMThreadContext *tc, MVMFrame *frame) {
     MVMFrame *update_caller = NULL;
     MVMFrame *result = NULL;
     MVM_CHECK_CALLER_CHAIN(tc, cur_to_promote);
-    MVMROOT3(tc, new_cur_frame, update_caller, result, {
+    MVMROOT4(tc, new_cur_frame, update_caller, cur_to_promote, result, {
         while (cur_to_promote) {
             /* Allocate a heap frame. */
             MVMFrame *promoted = MVM_gc_allocate_frame(tc);
@@ -701,6 +715,7 @@ MVMFrame * MVM_frame_move_to_heap(MVMThreadContext *tc, MVMFrame *frame) {
                     if (cur_to_promote == tc->thread_entry_frame)
                         tc->thread_entry_frame = promoted;
                     cur_to_promote = NULL;
+                    MVM_gc_write_barrier(tc, (MVMCollectable*)promoted, (MVMCollectable*)promoted->caller);
                 }
             }
             else {
@@ -849,6 +864,7 @@ MVMFrame * MVM_frame_create_for_deopt(MVMThreadContext *tc, MVMStaticFrame *stat
 static MVMuint64 remove_one_frame(MVMThreadContext *tc, MVMuint8 unwind) {
     MVMFrame *returner = tc->cur_frame;
     MVMFrame *caller   = returner->caller;
+    MVMuint32 need_caller;
     MVM_ASSERT_NOT_FROMSPACE(tc, caller);
 
     /* Clear up any extra frame data. */
@@ -856,6 +872,7 @@ static MVMuint64 remove_one_frame(MVMThreadContext *tc, MVMuint8 unwind) {
         MVMFrameExtra *e = returner->extra;
         if (e->continuation_tags)
             MVM_continuation_free_tags(tc, returner);
+        need_caller = e->caller_info_needed;
         /* Preserve the extras if the frame has been used in a ctx operation
          * and marked with caller info. */
         if (!(e->caller_deopt_idx || e->caller_jit_position)) {
@@ -863,6 +880,9 @@ static MVMuint64 remove_one_frame(MVMThreadContext *tc, MVMuint8 unwind) {
             returner->extra = NULL;
         }
     }
+    else {
+        need_caller = 0;
+    }
 
     /* Clean up frame working space. */
     if (returner->work) {
@@ -888,6 +908,8 @@ static MVMuint64 remove_one_frame(MVMThreadContext *tc, MVMuint8 unwind) {
      * in dynamic scope. */
     else {
         returner->work = NULL;
+        if (!need_caller)
+            returner->caller = NULL;
     }
 
     /* Switch back to the caller frame if there is one. */
@@ -1131,7 +1153,6 @@ MVMObject * MVM_frame_get_code_object(MVMThreadContext *tc, MVMCode *code) {
 
 /* Given the specified code object, sets its outer to the current scope. */
 void MVM_frame_capturelex(MVMThreadContext *tc, MVMObject *code) {
-    MVMCode *code_obj = (MVMCode *)code;
     MVMFrame *captured;
     if (MVM_UNLIKELY(REPR(code)->ID != MVM_REPR_ID_MVMCode))
         MVM_exception_throw_adhoc(tc,
@@ -1139,7 +1160,7 @@ void MVM_frame_capturelex(MVMThreadContext *tc, MVMObject *code) {
     MVMROOT(tc, code, {
         captured = MVM_frame_force_to_heap(tc, tc->cur_frame);
     });
-    MVM_ASSIGN_REF(tc, &(code->header), code_obj->body.outer, captured);
+    MVM_ASSIGN_REF(tc, &(code->header), ((MVMCode*)code)->body.outer, captured);
 }
 
 /* This is used for situations in Perl 6 like:
@@ -1157,10 +1178,9 @@ void MVM_frame_capturelex(MVMThreadContext *tc, MVMObject *code) {
  * $x.
  */
 void MVM_frame_capture_inner(MVMThreadContext *tc, MVMObject *code) {
-    MVMCode *code_obj = (MVMCode *)code;
     MVMFrame *outer;
     MVMROOT(tc, code, {
-        MVMStaticFrame *sf_outer = code_obj->body.sf->body.outer;
+        MVMStaticFrame *sf_outer = ((MVMCode*)code)->body.sf->body.outer;
         MVMROOT(tc, sf_outer, {
             outer = create_context_only(tc, sf_outer, (MVMObject *)sf_outer->body.static_code, 1);
         });
@@ -1169,7 +1189,7 @@ void MVM_frame_capture_inner(MVMThreadContext *tc, MVMObject *code) {
             MVM_ASSIGN_REF(tc, &(outer->header), outer->outer, outer_outer);
         });
     });
-    MVM_ASSIGN_REF(tc, &(code->header), code_obj->body.outer, outer);
+    MVM_ASSIGN_REF(tc, &(code->header), ((MVMCode*)code)->body.outer, outer);
 }
 
 /* Given the specified code object, copies it and returns a copy which
@@ -1234,8 +1254,15 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
         MVMint32 scid, objid;
         if (MVM_bytecode_find_static_lexical_scref(tc, effective_sf->body.cu,
                 effective_sf, effective_idx, &scid, &objid)) {
-            MVMSerializationContext *sc = MVM_sc_get_sc(tc, effective_sf->body.cu, scid);
+            MVMSerializationContext *sc;
             MVMObject *resolved;
+            /* XXX This really ought to go into the bytecode validator
+             * instead for better performance and earlier crashing */
+            if (effective_sf->body.cu->body.num_scs <= scid) {
+                MVM_exception_throw_adhoc(tc,
+                    "Bytecode corruption: illegal sc dependency of lexical: %d > %d", scid, effective_sf->body.cu->body.num_scs);
+            }
+            sc = MVM_sc_get_sc(tc, effective_sf->body.cu, scid);
             if (sc == NULL)
                 MVM_exception_throw_adhoc(tc,
                     "SC not yet resolved; lookup failed");
