@@ -33,6 +33,7 @@ typedef struct {
 /* Various common string constants. */
 struct MVMStringConsts {
     MVMString *empty;
+    MVMString *Int;
     MVMString *Str;
     MVMString *Num;
     MVMString *integer;
@@ -114,6 +115,17 @@ struct MVMObjectId {
     UT_hash_handle hash_handle;
 };
 
+struct MVMEventSubscriptions {
+    uv_mutex_t mutex_event_subscription;
+
+    MVMObject *subscription_queue;
+
+    MVMObject *GCEvent;
+    MVMObject *SpeshOverviewEvent;
+
+    MVMuint64 vm_startup_time;
+};
+
 /* Represents a MoarVM instance. */
 struct MVMInstance {
     /************************************************************************
@@ -175,6 +187,11 @@ struct MVMInstance {
      * join in, but this lets them wait efficieintly. */
     uv_cond_t cond_blocked_can_continue;
 
+    /* Whether the coordinator considers the run to be fully completed,
+     * including cleanup of exited threads. */
+    AO_t gc_completed;
+    uv_cond_t cond_gc_completed;
+
     /* The number of threads that have yet to acknowledge the finish. */
     AO_t gc_ack;
 
@@ -246,6 +263,12 @@ struct MVMInstance {
      * than per STable, as this doesn't happen often. */
     uv_mutex_t mutex_parameterization_add;
 
+    /************************************************************************
+     * Configuration programs (for spesh, profiler, ...)
+     ************************************************************************/
+
+    MVMConfigurationProgram *confprog;
+
     /************************************************************************
      * Specializer (dynamic optimization)
      ************************************************************************/
@@ -336,6 +359,7 @@ struct MVMInstance {
     MVMObject        *event_loop_permit_queue;
     MVMObject        *event_loop_cancel_queue;
     MVMObject        *event_loop_active;
+    MVMObject        *event_loop_free_indices;
     uv_async_t       *event_loop_wakeup;
 
     /* Standard file handles. */
@@ -425,6 +449,8 @@ struct MVMInstance {
     MVMObject *SpeshLog;
     MVMObject *StaticFrameSpesh;
 
+    MVMObject *SpeshPluginState;
+
     /* Set of bootstrapping types. */
     MVMBootTypes boot_types;
 
@@ -501,6 +527,9 @@ struct MVMInstance {
     FILE *coverage_log_fh;
     MVMuint32  coverage_control;
 
+    /* The time it takes to run the profiler instrumentation. */
+    MVMuint64 profiling_overhead;
+
     /************************************************************************
      * Debugging
      ************************************************************************/
@@ -519,4 +548,10 @@ struct MVMInstance {
     /* Hash Secrets which is used as the hash seed. This is to avoid denial of
      * service type attacks. */
     MVMuint64 hashSecrets[2];
+
+    /************************************************************************
+     * VM Event subscription
+     ************************************************************************/
+
+    MVMEventSubscriptions subscriptions;
 };
