@@ -283,7 +283,7 @@ double MVM_nativecall_unmarshal_double(MVMThreadContext *tc, MVMObject *value) {
     return (double)MVM_repr_get_num(tc, value);
 }
 
-char * MVM_nativecall_unmarshal_string(MVMThreadContext *tc, MVMObject *value, MVMint16 type, MVMint16 *free) {
+char * MVM_nativecall_unmarshal_string(MVMThreadContext *tc, MVMObject *value, MVMint16 type, MVMint16 *free, MVMint16 unmarshal_kind) {
     if (IS_CONCRETE(value)) {
         MVMString *value_str = MVM_repr_get_str(tc, value);
 
@@ -317,47 +317,62 @@ char * MVM_nativecall_unmarshal_string(MVMThreadContext *tc, MVMObject *value, M
     }
 }
 
-void * MVM_nativecall_unmarshal_cstruct(MVMThreadContext *tc, MVMObject *value) {
+static void unmarshal_error(MVMThreadContext *tc, char *desired_repr, MVMObject *value, MVMint16 unmarshal_kind) {
+    if (unmarshal_kind == MVM_NATIVECALL_UNMARSHAL_KIND_GENERIC) {
+        MVM_exception_throw_adhoc(tc,
+            "NativeCall conversion expected type with %s representation, but got a %s (%s)", desired_repr, REPR(value)->name, MVM_6model_get_debug_name(tc, value));
+    }
+    else if (unmarshal_kind == MVM_NATIVECALL_UNMARSHAL_KIND_RETURN) {
+        MVM_exception_throw_adhoc(tc,
+            "Expected return value with %s representation, but got a %s (%s)", desired_repr, REPR(value)->name, MVM_6model_get_debug_name(tc, value));
+    }
+    else if (unmarshal_kind == MVM_NATIVECALL_UNMARSHAL_KIND_NATIVECAST) {
+        MVM_exception_throw_adhoc(tc,
+            "NativeCast expected value with %s representation, but got a %s (%s)", desired_repr, REPR(value)->name, MVM_6model_get_debug_name(tc, value));
+    }
+    else {
+        MVM_exception_throw_adhoc(tc,
+            "Native call expected argument %d with %s representation, but got a %s (%s)", unmarshal_kind + 1, desired_repr, REPR(value)->name, MVM_6model_get_debug_name(tc, value));
+    }
+}
+
+void * MVM_nativecall_unmarshal_cstruct(MVMThreadContext *tc, MVMObject *value, MVMint16 unmarshal_kind) {
     if (!IS_CONCRETE(value))
         return NULL;
     else if (REPR(value)->ID == MVM_REPR_ID_MVMCStruct)
         return ((MVMCStruct *)value)->body.cstruct;
     else
-        MVM_exception_throw_adhoc(tc,
-            "Native call expected return type with CStruct representation, but got a %s (%s)", REPR(value)->name, MVM_6model_get_debug_name(tc, value));
+        unmarshal_error(tc, "CStruct", value, unmarshal_kind);
 }
 
-void * MVM_nativecall_unmarshal_cppstruct(MVMThreadContext *tc, MVMObject *value) {
+void * MVM_nativecall_unmarshal_cppstruct(MVMThreadContext *tc, MVMObject *value, MVMint16 unmarshal_kind) {
     if (!IS_CONCRETE(value))
         return NULL;
     else if (REPR(value)->ID == MVM_REPR_ID_MVMCPPStruct)
         return ((MVMCPPStruct *)value)->body.cppstruct;
     else
-        MVM_exception_throw_adhoc(tc,
-            "Native call expected return type with CPPStruct representation, but got a %s (%s)", REPR(value)->name, MVM_6model_get_debug_name(tc, value));
+        unmarshal_error(tc, "CPPStruct", value, unmarshal_kind);
 }
 
-void * MVM_nativecall_unmarshal_cpointer(MVMThreadContext *tc, MVMObject *value) {
+void * MVM_nativecall_unmarshal_cpointer(MVMThreadContext *tc, MVMObject *value, MVMint16 unmarshal_kind) {
     if (!IS_CONCRETE(value))
         return NULL;
     else if (REPR(value)->ID == MVM_REPR_ID_MVMCPointer)
         return ((MVMCPointer *)value)->body.ptr;
     else
-        MVM_exception_throw_adhoc(tc,
-            "Native call expected return type with CPointer representation, but got a %s (%s)", REPR(value)->name, MVM_6model_get_debug_name(tc, value));
+        unmarshal_error(tc, "CPointer", value, unmarshal_kind);
 }
 
-void * MVM_nativecall_unmarshal_carray(MVMThreadContext *tc, MVMObject *value) {
+void * MVM_nativecall_unmarshal_carray(MVMThreadContext *tc, MVMObject *value, MVMint16 unmarshal_kind) {
     if (!IS_CONCRETE(value))
         return NULL;
     else if (REPR(value)->ID == MVM_REPR_ID_MVMCArray)
         return ((MVMCArray *)value)->body.storage;
     else
-        MVM_exception_throw_adhoc(tc,
-            "Native call expected return type with CArray representation, but got a %s (%s)", REPR(value)->name, MVM_6model_get_debug_name(tc, value));
+        unmarshal_error(tc, "CArray", value, unmarshal_kind);
 }
 
-void * MVM_nativecall_unmarshal_vmarray(MVMThreadContext *tc, MVMObject *value) {
+void * MVM_nativecall_unmarshal_vmarray(MVMThreadContext *tc, MVMObject *value, MVMint16 unmarshal_kind) {
     if (!IS_CONCRETE(value))
         return NULL;
     else if (REPR(value)->ID == MVM_REPR_ID_VMArray) {
@@ -367,18 +382,16 @@ void * MVM_nativecall_unmarshal_vmarray(MVMThreadContext *tc, MVMObject *value)
         return ((char *)body->slots.any) + start_pos;
     }
     else
-        MVM_exception_throw_adhoc(tc,
-            "Native call expected object with Array representation, but got a %s (%s)", REPR(value)->name, MVM_6model_get_debug_name(tc, value));
+        unmarshal_error(tc, "VMArray", value, unmarshal_kind);
 }
 
-void * MVM_nativecall_unmarshal_cunion(MVMThreadContext *tc, MVMObject *value) {
+void * MVM_nativecall_unmarshal_cunion(MVMThreadContext *tc, MVMObject *value, MVMint16 unmarshal_kind) {
     if (!IS_CONCRETE(value))
         return NULL;
     else if (REPR(value)->ID == MVM_REPR_ID_MVMCUnion)
         return ((MVMCUnion *)value)->body.cunion;
     else
-        MVM_exception_throw_adhoc(tc,
-            "Native call expected return type with CUnion representation, but got a %s (%s)", REPR(value)->name, MVM_6model_get_debug_name(tc, value));
+        unmarshal_error(tc, "CUnion", value, unmarshal_kind);
 }
 
 #ifdef _WIN32
@@ -519,6 +532,12 @@ MVMJitGraph *MVM_nativecall_jit_graph_for_caller_code(
                         ? is_rw ? MVM_JIT_ARG_I64_RW : MVM_JIT_ARG_I64
                         : is_rw ? MVM_JIT_PARAM_I64_RW : MVM_JIT_PARAM_I64;
                     break;
+                case MVM_NATIVECALL_ARG_DOUBLE:
+                    if (is_rw) goto fail;
+                    arg_type = dst == -1
+                        ? MVM_JIT_ARG_DOUBLE
+                        : MVM_JIT_PARAM_DOUBLE;
+                    break;
                 case MVM_NATIVECALL_ARG_CPOINTER:
                     if (is_rw) goto fail;
                     arg_type = dst == -1 ? MVM_JIT_ARG_PTR : MVM_JIT_PARAM_PTR;
@@ -614,6 +633,40 @@ MVMJitCode *create_caller_code(MVMThreadContext *tc, MVMNativeCallBody *body) {
     return jitcode;
 }
 
+void MVM_nativecall_setup(MVMThreadContext *tc, MVMNativeCallBody *body, unsigned int interval_id) {
+    /* Try to load the library. */
+    body->lib_handle = MVM_nativecall_load_lib(body->lib_name[0] ? body->lib_name : NULL);
+
+    if (!body->lib_handle) {
+        char *waste[] = { body->lib_name, NULL };
+        MVM_free(body->sym_name);
+        body->lib_name = NULL;
+        body->sym_name = NULL;
+        if (interval_id)
+            MVM_telemetry_interval_stop(tc, interval_id, "error building native call");
+        MVM_exception_throw_adhoc_free(tc, waste, "Cannot locate native library '%s': %s", body->lib_name, dlerror());
+    }
+
+    if (!body->entry_point) {
+        body->entry_point = MVM_nativecall_find_sym(body->lib_handle, body->sym_name);
+        if (!body->entry_point) {
+            char *waste[] = { body->sym_name, body->lib_name, NULL };
+            body->lib_name = NULL;
+            body->sym_name = NULL;
+            if (interval_id)
+                MVM_telemetry_interval_stop(tc, interval_id, "error building native call");
+            MVM_exception_throw_adhoc_free(tc, waste, "Cannot locate symbol '%s' in native library '%s'",
+                body->sym_name, body->lib_name);
+        }
+    }
+
+    if (tc->instance->jit_enabled) {
+        body->jitcode = create_caller_code(tc, body);
+    }
+    else
+        body->jitcode = NULL;
+}
+
 /* Builds up a native call site out of the supplied arguments. */
 MVMint8 MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *lib,
         MVMString *sym, MVMString *conv, MVMObject *arg_info, MVMObject *ret_info) {
@@ -630,33 +683,17 @@ MVMint8 MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *l
     /* Initialize the object; grab native call part of its body. */
     MVMNativeCallBody *body = MVM_nativecall_get_nc_body(tc, site);
 
-    /* Try to load the library. */
     body->lib_name = lib_name;
-    body->lib_handle = MVM_nativecall_load_lib(lib_name[0] ? lib_name : NULL);
-
-    if (!body->lib_handle) {
-        char *waste[] = { lib_name, NULL };
-        MVM_free(sym_name);
-        MVM_telemetry_interval_stop(tc, interval_id, "error building native call");
-        MVM_exception_throw_adhoc_free(tc, waste, "Cannot locate native library '%s': %s", lib_name, dlerror());
-    }
 
     /* Try to locate the symbol. */
-    if (entry_point_o) {
-        body->entry_point = MVM_nativecall_unmarshal_cpointer(tc, entry_point_o);
+    if (entry_point_o && !MVM_is_null(tc, entry_point_o)) {
+        body->entry_point = MVM_nativecall_unmarshal_cpointer(tc, entry_point_o, MVM_NATIVECALL_UNMARSHAL_KIND_GENERIC);
         body->sym_name    = sym_name;
         keep_sym_name     = 1;
     }
 
     if (!body->entry_point) {
-        body->entry_point = MVM_nativecall_find_sym(body->lib_handle, sym_name);
-        if (!body->entry_point) {
-            char *waste[] = { sym_name, lib_name, NULL };
-            MVM_telemetry_interval_stop(tc, interval_id, "error building native call");
-            MVM_exception_throw_adhoc_free(tc, waste, "Cannot locate symbol '%s' in native library '%s'",
-                sym_name, lib_name);
-        }
-        body->sym_name = sym_name;
+        body->sym_name    = sym_name;
         keep_sym_name     = 1;
     }
 
@@ -696,11 +733,8 @@ MVMint8 MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *l
 #ifdef HAVE_LIBFFI
     body->ffi_ret_type = MVM_nativecall_get_ffi_type(tc, body->ret_type);
 #endif
-    if (tc->instance->jit_enabled) {
-        body->jitcode = create_caller_code(tc, body);
-    }
-    else
-        body->jitcode = NULL;
+
+    MVM_nativecall_setup(tc, body, interval_id);
 
     MVM_telemetry_interval_stop(tc, interval_id, "nativecall built");
 
@@ -897,20 +931,20 @@ MVMObject * MVM_nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec, MV
         return target_type;
 
     if (REPR(source)->ID == MVM_REPR_ID_MVMCStruct)
-        data_body = MVM_nativecall_unmarshal_cstruct(tc, source);
+        data_body = MVM_nativecall_unmarshal_cstruct(tc, source, MVM_NATIVECALL_UNMARSHAL_KIND_NATIVECAST);
     else if (REPR(source)->ID == MVM_REPR_ID_MVMCPPStruct)
-        data_body = MVM_nativecall_unmarshal_cppstruct(tc, source);
+        data_body = MVM_nativecall_unmarshal_cppstruct(tc, source, MVM_NATIVECALL_UNMARSHAL_KIND_NATIVECAST);
     else if (REPR(source)->ID == MVM_REPR_ID_MVMCUnion)
-        data_body = MVM_nativecall_unmarshal_cunion(tc, source);
+        data_body = MVM_nativecall_unmarshal_cunion(tc, source, MVM_NATIVECALL_UNMARSHAL_KIND_NATIVECAST);
     else if (REPR(source)->ID == MVM_REPR_ID_MVMCPointer)
-        data_body = MVM_nativecall_unmarshal_cpointer(tc, source);
+        data_body = MVM_nativecall_unmarshal_cpointer(tc, source, MVM_NATIVECALL_UNMARSHAL_KIND_NATIVECAST);
     else if (REPR(source)->ID == MVM_REPR_ID_MVMCArray)
-        data_body = MVM_nativecall_unmarshal_carray(tc, source);
+        data_body = MVM_nativecall_unmarshal_carray(tc, source, MVM_NATIVECALL_UNMARSHAL_KIND_NATIVECAST);
     else if (REPR(source)->ID == MVM_REPR_ID_VMArray)
-        data_body = MVM_nativecall_unmarshal_vmarray(tc, source);
+        data_body = MVM_nativecall_unmarshal_vmarray(tc, source, MVM_NATIVECALL_UNMARSHAL_KIND_NATIVECAST);
     else
         MVM_exception_throw_adhoc(tc,
-            "Native call expected return type with CPointer, CStruct, CArray, or VMArray representation, but got a %s (%s)", REPR(source)->name, MVM_6model_get_debug_name(tc, source));
+            "Native call cast expected return type with CPointer, CStruct, CArray, or VMArray representation, but got a %s (%s)", REPR(source)->name, MVM_6model_get_debug_name(tc, source));
     return nativecall_cast(tc, target_spec, target_type, data_body);
 }
 
@@ -1005,15 +1039,15 @@ void MVM_nativecall_refresh(MVMThreadContext *tc, MVMObject *cthingy) {
         MVMint64            i;
 
         for (i = 0; i < repr_data->num_attributes; i++) {
-            MVMint32 kind = repr_data->attribute_locations[i] & MVM_CSTRUCT_ATTR_MASK;
-            MVMint32 slot = repr_data->attribute_locations[i] >> MVM_CSTRUCT_ATTR_SHIFT;
-            void *cptr = NULL;   /* The pointer in the C storage. */
+            MVMint32  kind = repr_data->attribute_locations[i] & MVM_CSTRUCT_ATTR_MASK;
+            MVMint32  slot = repr_data->attribute_locations[i] >> MVM_CSTRUCT_ATTR_SHIFT;
+            void *cptr   = NULL; /* Address of the struct member holding the pointer in the C storage. */
             void *objptr = NULL; /* The pointer in the object representing the C object. */
 
             if (kind == MVM_CSTRUCT_ATTR_IN_STRUCT || !body->child_objs[slot])
                 continue;
 
-            cptr = (void*)((uintptr_t)storage + (uintptr_t)repr_data->struct_offsets[i]);
+            cptr = (void *)(storage + (uintptr_t)repr_data->struct_offsets[i]);
             if (IS_CONCRETE(body->child_objs[slot])) {
                 switch (kind) {
                     case MVM_CSTRUCT_ATTR_CARRAY:
@@ -1065,7 +1099,7 @@ void MVM_nativecall_refresh(MVMThreadContext *tc, MVMObject *cthingy) {
             if (kind == MVM_CPPSTRUCT_ATTR_IN_STRUCT || !body->child_objs[slot])
                 continue;
 
-            cptr = (void*)((uintptr_t)storage + (uintptr_t)repr_data->struct_offsets[i]);
+            cptr = (void *)(storage + (uintptr_t)repr_data->struct_offsets[i]);
             if (IS_CONCRETE(body->child_objs[slot])) {
                 switch (kind) {
                     case MVM_CPPSTRUCT_ATTR_CARRAY:
