@@ -1,6 +1,9 @@
 #define MVMGCSTATUS_MASK 3
 #define MVMSUSPENDSTATUS_MASK 12
 
+#define MVM_NUM_TEMP_BIGINTS 3
+#include "tommath.h"
+
 /* Possible values for the thread execution interrupt flag. */
 typedef enum {
     /* Indicates that the thread is currently executing, and should
@@ -236,8 +239,9 @@ struct MVMThreadContext {
 
     /* We try to do better at OSR by creating a fresh log when we enter a new
      * compilation unit. However, for things that EVAL or do a ton of BEGIN,
-     * this does more harm than good. Use this to throttle it back. */
-    MVMuint32 num_compunit_extra_logs;
+     * we risk high memory use. Use this to throttle it by limiting the number
+     * of such extra logs that might exist at a time. */
+    AO_t num_compunit_extra_logs;
 
     /* The current specialization correlation ID, used in logging. */
     MVMuint32 spesh_cid;
@@ -261,6 +265,10 @@ struct MVMThreadContext {
     MVMObject *plugin_guard_args;
     MVMuint32 num_plugin_guards;
 
+    MVMSpeshPluginGuard *temp_plugin_guards;
+    MVMObject *temp_plugin_guard_args;
+    MVMuint32 temp_num_plugin_guards;
+
     /************************************************************************
      * Per-thread state held by assorted VM subsystems
      ************************************************************************/
@@ -295,6 +303,10 @@ struct MVMThreadContext {
     /* Random number generator state. */
     MVMuint64 rand_state[2];
 
+    /* Temporary big integers for when we need to upgrade operands in order
+     * to perform an operation. */
+    mp_int *temp_bigints[MVM_NUM_TEMP_BIGINTS];
+
     /* NFA evaluator memory cache, to avoid many allocations; see NFA.c. */
     MVMuint32 *nfa_done;
     MVMuint32 *nfa_curst;
@@ -326,5 +338,6 @@ struct MVMThreadContext {
 MVMThreadContext * MVM_tc_create(MVMThreadContext *parent, MVMInstance *instance);
 void MVM_tc_destroy(MVMThreadContext *tc);
 void MVM_tc_set_ex_release_mutex(MVMThreadContext *tc, uv_mutex_t *mutex);
+void MVM_tc_set_ex_release_atomic(MVMThreadContext *tc, AO_t *flag);
 void MVM_tc_release_ex_release_mutex(MVMThreadContext *tc);
 void MVM_tc_clear_ex_release_mutex(MVMThreadContext *tc);
