@@ -166,7 +166,7 @@ printf(" %u %s %.2s\n", val->cur_instr, info->name, info->mark);
 
     val->labels[pos] |= MVM_BC_op_boundary;
     val->cur_info     = info;
-    val->cur_mark     = info->mark;
+    val->cur_mark     = MVM_op_get_mark(opcode);
     val->cur_op      += 2;
     val->cur_instr   += 1;
 }
@@ -468,6 +468,9 @@ static void validate_sequence(Validator *val) {
         type = val->cur_mark[0];
         id   = val->cur_mark[1];
 
+        if (val->cur_info->specializable)
+            val->frame->body.specializable = 1;
+
         switch (type) {
             case MARK_special:
                 if (id == seq_id)
@@ -621,6 +624,9 @@ static void validate_block(Validator *val) {
         type = val->cur_mark[0];
         id   = val->cur_mark[1];
 
+        if (val->cur_info->specializable)
+            val->frame->body.specializable = 1;
+
         if (id != block_id)
             fail(val, MSG(val, "expected instruction marked '%c' but got '%c'"),
                     block_id, id);
@@ -644,6 +650,7 @@ terminate_block:
         case 'a':
             validate_operands(val);
             ensure_no_remaining_positionals(val);
+            ensure_no_remaining_args(val);
             break;
     }
 }
