@@ -1,5 +1,6 @@
 #include "moar.h"
 #include <platform/threads.h>
+#include "platform/malloc_trim.h"
 
 /* If we have the job of doing GC for a thread, we add it to our work
  * list. */
@@ -182,13 +183,14 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
         }
 
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
-            "Thread %d run %d : Co-ordinator handling fixed-size allocator safepoint frees\n");
-        MVM_fixed_size_safepoint(tc, tc->instance->fsa);
-        MVM_alloc_safepoint(tc);
-
+            "Thread %d run %d : Co-ordinator heapsnapshot or instrumented profiler data output\n");
         MVM_profile_dump_instrumented_data(tc);
         MVM_profile_heap_take_snapshot(tc);
 
+        GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
+            "Thread %d run %d : Co-ordinator handling fixed-size allocator safepoint frees\n");
+        MVM_fixed_size_safepoint(tc, tc->instance->fsa);
+        MVM_alloc_safepoint(tc);
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
             "Thread %d run %d : Co-ordinator signalling in-trays clear\n");
         uv_mutex_lock(&tc->instance->mutex_gc_orchestrate);
@@ -218,17 +220,20 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
             MVM_gc_gen2_transfer(other, tc);
             GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
                 "Thread %d run %d : destroying thread %d\n", other->thread_id);
-            MVM_tc_destroy(other);
             tc->gc_work[i].tc = thread_obj->body.tc = NULL;
+            MVM_tc_destroy(other);
             MVM_store(&thread_obj->body.stage, MVM_thread_stage_destroyed);
         }
         else {
             /* Free gen2 unmarked if full collection. */
             if (gen == MVMGCGenerations_Both) {
+                /* Tell malloc implementation to free empty pages to kernel.
+                 * Currently only activated for Linux. */
+                MVM_malloc_trim();
                 GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
                     "Thread %d run %d : freeing gen2 of thread %d\n",
                     other->thread_id);
-                MVM_gc_collect_free_gen2_unmarked(other, 0);
+                MVM_gc_collect_free_gen2_unmarked(tc, other, 0);
             }
 
             /* Contribute this thread's promoted bytes. */
@@ -238,7 +243,7 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
             GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
                 "Thread %d run %d : collecting nursery uncopied of thread %d\n",
                 other->thread_id);
-            MVM_gc_collect_free_nursery_uncopied(other, tc->gc_work[i].limit);
+            MVM_gc_collect_free_nursery_uncopied(tc, other, tc->gc_work[i].limit);
 
             /* Handle exited threads. */
             if (MVM_load(&thread_obj->body.stage) == MVM_thread_stage_exited) {
@@ -249,12 +254,29 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
                     other->thread_id, (int)MVM_load(&thread_obj->body.stage));
             }
 
+#if MVM_GC_DEBUG >= 3
+                memset(other->nursery_fromspace, 0xef, other->nursery_fromspace_size);
+#endif
+
             /* Mark thread free to continue. */
             MVM_cas(&other->gc_status, MVMGCStatus_STOLEN, MVMGCStatus_UNABLE);
             MVM_cas(&other->gc_status, MVMGCStatus_INTERRUPT, MVMGCStatus_NONE);
         }
     }
 
+    if (is_coordinator) {
+        uv_mutex_lock(&tc->instance->mutex_gc_orchestrate);
+        MVM_store(&tc->instance->gc_completed, 1);
+        uv_cond_broadcast(&tc->instance->cond_gc_completed);
+        uv_mutex_unlock(&tc->instance->mutex_gc_orchestrate);
+    }
+    else {
+        uv_mutex_lock(&tc->instance->mutex_gc_orchestrate);
+        while (!MVM_load(&tc->instance->gc_completed))
+            uv_cond_wait(&tc->instance->cond_gc_completed, &tc->instance->mutex_gc_orchestrate);
+        uv_mutex_unlock(&tc->instance->mutex_gc_orchestrate);
+    }
+
     /* Signal acknowledgement of completing the cleanup,
      * except for STables, and if we're the final to do
      * so, free the STables, which have been linked. */
@@ -383,8 +405,16 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
     MVMuint8   gen;
     MVMuint32  i, n;
 
+    MVMuint8 is_coordinator;
+
+    MVMuint64 start_time;
+
     unsigned int interval_id;
 
+    MVMObject *subscription_queue = NULL;
+
+    is_coordinator = what_to_do == MVMGCWhatToDo_All;
+
 #if MVM_GC_DEBUG
     if (tc->in_spesh)
         MVM_panic(1, "Must not GC when in the specializer/JIT\n");
@@ -399,6 +429,9 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
         interval_id = MVM_telemetry_interval_start(tc, "start minor collection");
     }
 
+    if (is_coordinator)
+        start_time = uv_hrtime();
+
     /* Do GC work for ourselves and any work threads. */
     for (i = 0, n = tc->gc_work_count ; i < n; i++) {
         MVMThreadContext *other = tc->gc_work[i].tc;
@@ -406,11 +439,51 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : starting collection for thread %d\n",
             other->thread_id);
         other->gc_promoted_bytes = 0;
+        if (tc->instance->profiling)
+            MVM_profiler_log_gen2_roots(tc, other->num_gen2roots, other);
         MVM_gc_collect(other, (other == tc ? what_to_do : MVMGCWhatToDo_NoInstance), gen);
     }
 
     /* Wait for everybody to agree we're done. */
-    finish_gc(tc, gen, what_to_do == MVMGCWhatToDo_All);
+    finish_gc(tc, gen, is_coordinator);
+
+    /* Finally, as the very last thing ever, the coordinator pushes a bit of
+     * info into the subscription queue (if it is set) */
+
+    subscription_queue = tc->instance->subscriptions.subscription_queue;
+
+    if (is_coordinator && subscription_queue && tc->instance->subscriptions.GCEvent) {
+        MVMuint64 end_time = uv_hrtime();
+        MVMObject *instance = MVM_repr_alloc(tc, tc->instance->subscriptions.GCEvent);
+        MVMThread *cur_thread;
+
+        MVMArray *arrobj = (MVMArray *)instance;
+        MVMuint64 *data;
+
+        MVM_repr_pos_set_elems(tc, instance, 9);
+
+        data = arrobj->body.slots.u64;
+
+        data[0] = MVM_load(&tc->instance->gc_seq_number);
+        data[1] = start_time / 1000;
+        data[2] = (start_time - tc->instance->subscriptions.vm_startup_time) / 1000;
+        data[3] = (end_time - start_time) / 1000;
+        data[4] = gen == MVMGCGenerations_Both;
+        data[5] = tc->gc_promoted_bytes;
+        data[6] = MVM_load(&tc->instance->gc_promoted_bytes_since_last_full);
+        data[7] = tc->thread_id;
+        data[8] = 0;
+
+        uv_mutex_lock(&tc->instance->mutex_threads);
+        cur_thread = tc->instance->threads;
+        while (cur_thread) {
+            data[8] += cur_thread->body.tc->num_gen2roots;
+            cur_thread = cur_thread->body.next;
+        }
+        uv_mutex_unlock(&tc->instance->mutex_threads);
+
+        MVM_repr_push_o(tc, tc->instance->subscriptions.subscription_queue, instance);
+    }
 
     MVM_telemetry_interval_stop(tc, interval_id, "finished run_gc");
 }
@@ -463,6 +536,7 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
         /* Flag that we didn't agree on this run that all the in-trays are
          * cleared (a responsibility of the co-ordinator. */
         MVM_store(&tc->instance->gc_intrays_clearing, 1);
+        MVM_store(&tc->instance->gc_completed, 0);
 
         /* We'll take care of our own work. */
         add_work(tc, tc);
@@ -583,10 +657,6 @@ void MVM_gc_enter_from_interrupt(MVMThreadContext *tc) {
 
     MVM_telemetry_timestamp(tc, "gc_enter_from_interrupt");
 
-    /* If profiling, record that GC is starting. */
-    if (tc->instance->profiling)
-        MVM_profiler_log_gc_start(tc, is_full_collection(tc), 0);
-
     /* We'll certainly take care of our own work. */
     tc->gc_work_count = 0;
     add_work(tc, tc);
@@ -602,6 +672,13 @@ void MVM_gc_enter_from_interrupt(MVMThreadContext *tc) {
     uv_cond_broadcast(&tc->instance->cond_gc_start);
     uv_mutex_unlock(&tc->instance->mutex_gc_orchestrate);
 
+    /* If profiling, record that GC is starting.
+     * We don't do this before add_work, because the gc_sequence_number
+     * may not yet have been increased, leading to duplicate entries
+     * in the gc parts of the profiler */
+    if (tc->instance->profiling)
+        MVM_profiler_log_gc_start(tc, is_full_collection(tc), 0);
+
     /* Wait for all threads to indicate readiness to collect. */
     GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : Waiting for other threads\n");
     uv_mutex_lock(&tc->instance->mutex_gc_orchestrate);
@@ -669,8 +746,8 @@ void MVM_gc_global_destruction(MVMThreadContext *tc) {
     tc->nursery_tospace = nursery_tmp;
 
     /* Run the objects' finalizers */
-    MVM_gc_collect_free_nursery_uncopied(tc, tc->nursery_alloc);
+    MVM_gc_collect_free_nursery_uncopied(tc, tc, tc->nursery_alloc);
     MVM_gc_root_gen2_cleanup(tc);
-    MVM_gc_collect_free_gen2_unmarked(tc, 1);
+    MVM_gc_collect_free_gen2_unmarked(tc, tc, 1);
     MVM_gc_collect_free_stables(tc);
 }
