@@ -2,6 +2,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <moar.h>
+#include "platform/io.h"
 
 #if MVM_TRACING
 #  define TRACING_OPT "[--tracing] "
@@ -226,7 +227,7 @@ int wmain(int argc, wchar_t *wargv[])
 
             case OPT_DEBUGPORT: {
                 MVMint64 port;
-                char *portstr = argv[argi] + strlen("--debugport=") + 1;
+                char *portstr = argv[argi] + strlen("--debug-port=");
                 char *endptr;
                 port = strtoll(portstr, &endptr, 10);
                 if (*endptr != '\0') {
@@ -258,7 +259,7 @@ int wmain(int argc, wchar_t *wargv[])
              getpid()
 #endif
              );
-        fp = fopen(path, "w");
+        fp = MVM_platform_fopen(path, "w");
         if (fp) {
             MVM_telemetry_init(fp);
             telemeh_inited = 1;
