@@ -57,23 +57,23 @@ int MVM_bigint_mp_set_uint64(mp_int * a, MVMuint64 b) {
  *  https://github.com/czurnieden/libtommath/blob/master/bn_mp_get_double.c
  *  and slightly modified to fit MoarVM's setup.
  */
-static const int mp_get_double_digits_needed
+static const int MVM_mp_get_double_digits_needed
 = ((MANTISSA_BITS_IN_DOUBLE + DIGIT_BIT) / DIGIT_BIT) + 1;
-static const double mp_get_double_multiplier = (double)(MP_MASK + 1);
+static const double MVM_mp_get_double_multiplier = (double)(MP_MASK + 1);
 
-static MVMnum64 mp_get_double(mp_int *a, int shift) {
+static MVMnum64 MVM_mp_get_double_shift(mp_int *a, int shift) {
     MVMnum64 d;
     int i, limit, final_shift;
     d = 0.0;
 
     mp_clamp(a);
     i = a->used;
-    limit = (i <= mp_get_double_digits_needed)
-        ? 0 : i - mp_get_double_digits_needed;
+    limit = (i <= MVM_mp_get_double_digits_needed)
+        ? 0 : i - MVM_mp_get_double_digits_needed;
 
     while (i-- > limit) {
         d += a->dp[i];
-        d *= mp_get_double_multiplier;
+        d *= MVM_mp_get_double_multiplier;
     }
 
     if (a->sign == MP_NEG)
@@ -161,38 +161,40 @@ static int can_be_smallint(const mp_int *i) {
 
 /* Forces a bigint, even if we only have a smallint. Takes a parameter that
  * indicates where to allocate a temporary mp_int if needed. */
-static mp_int * force_bigint(const MVMP6bigintBody *body, mp_int **tmp) {
+static mp_int * force_bigint(MVMThreadContext *tc, const MVMP6bigintBody *body, int idx) {
     if (MVM_BIGINT_IS_BIG(body)) {
         return body->u.bigint;
     }
     else {
-        MVMint64 value = body->u.smallint.value;
-        mp_int *i = MVM_malloc(sizeof(mp_int));
-        mp_init(i);
-        if (value >= 0) {
-            mp_set_long(i, value);
+        if (sizeof(mp_digit) > 4) {
+            MVMint64 value = body->u.smallint.value;
+            mp_int *i = tc->temp_bigints[idx];
+            if (value >= 0) {
+                mp_digit d = value;
+                mp_set(i, d);
+            }
+            else {
+                mp_digit d = -value;
+                mp_set(i, d);
+                mp_neg(i, i);
+            }
+            return i;
         }
         else {
-            mp_set_long(i, -value);
-            mp_neg(i, i);
+            MVMint32 value = body->u.smallint.value;
+            mp_int *i = tc->temp_bigints[idx];
+            if (value >= 0) {
+                mp_set_int(i, value);
+            }
+            else {
+                mp_set_int(i, -value);
+                mp_neg(i, i);
+            }
+            return i;
         }
-        while (*tmp)
-            tmp++;
-        *tmp = i;
-        return i;
     }
 }
 
-/* Clears an array that may contain tempory big ints. */
-static void clear_temp_bigints(mp_int *const *ints, MVMint32 n) {
-    MVMint32 i;
-    for (i = 0; i < n; i++)
-        if (ints[i]) {
-            mp_clear(ints[i]);
-            MVM_free(ints[i]);
-        }
-}
-
 /* Stores an int64 in a bigint result body, either as a 32-bit smallint if it
  * is in range, or a big integer if not. */
 static void store_int64_result(MVMP6bigintBody *body, MVMint64 result) {
@@ -357,7 +359,6 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
 MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) { \
     MVMP6bigintBody *ba, *bb, *bc; \
     MVMObject *result; \
-    mp_int *tmp[2] = { NULL, NULL }; \
     mp_int *ia, *ib, *ic; \
     MVMROOT2(tc, a, b, { \
         result = MVM_repr_alloc_init(tc, result_type);\
@@ -365,13 +366,12 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
     ba = get_bigint_body(tc, a); \
     bb = get_bigint_body(tc, b); \
     bc = get_bigint_body(tc, result); \
-    ia = force_bigint(ba, tmp); \
-    ib = force_bigint(bb, tmp); \
+    ia = force_bigint(tc, ba, 0); \
+    ib = force_bigint(tc, bb, 1); \
     ic = MVM_malloc(sizeof(mp_int)); \
     mp_init(ic); \
     mp_##opname(ia, ib, ic); \
     store_bigint_result(bc, ic); \
-    clear_temp_bigints(tmp, 2); \
     adjust_nursery(tc, bc); \
     return result; \
 }
@@ -379,15 +379,13 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
 #define MVM_BIGINT_BINARY_OP_SIMPLE(opname, SMALLINT_OP) \
 void MVM_bigint_fallback_##opname(MVMThreadContext *tc, MVMP6bigintBody *ba, MVMP6bigintBody *bb, \
                                   MVMP6bigintBody *bc) { \
-    mp_int *tmp[2] = { NULL, NULL }; \
     mp_int *ia, *ib, *ic; \
-    ia = force_bigint(ba, tmp); \
-    ib = force_bigint(bb, tmp); \
+    ia = force_bigint(tc, ba, 0); \
+    ib = force_bigint(tc, bb, 1); \
     ic = MVM_malloc(sizeof(mp_int)); \
     mp_init(ic); \
     mp_##opname(ia, ib, ic); \
     store_bigint_result(bc, ic); \
-    clear_temp_bigints(tmp, 2); \
     adjust_nursery(tc, bc); \
 } \
 MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) { \
@@ -396,7 +394,6 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
     ba = get_bigint_body(tc, a); \
     bb = get_bigint_body(tc, b); \
     if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) { \
-        mp_int *tmp[2] = { NULL, NULL }; \
         mp_int *ia, *ib, *ic; \
         MVMROOT2(tc, a, b, { \
             result = MVM_repr_alloc_init(tc, result_type);\
@@ -404,13 +401,12 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
         ba = get_bigint_body(tc, a); \
         bb = get_bigint_body(tc, b); \
         bc = get_bigint_body(tc, result); \
-        ia = force_bigint(ba, tmp); \
-        ib = force_bigint(bb, tmp); \
+        ia = force_bigint(tc, ba, 0); \
+        ib = force_bigint(tc, bb, 1); \
         ic = MVM_malloc(sizeof(mp_int)); \
         mp_init(ic); \
         mp_##opname(ia, ib, ic); \
         store_bigint_result(bc, ic); \
-        clear_temp_bigints(tmp, 2); \
         adjust_nursery(tc, bc); \
     } \
     else { \
@@ -430,33 +426,32 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
 
 #define MVM_BIGINT_BINARY_OP_2(opname, SMALLINT_OP) \
 MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) { \
-    MVMP6bigintBody *ba = get_bigint_body(tc, a); \
-    MVMP6bigintBody *bb = get_bigint_body(tc, b); \
-    MVMP6bigintBody *bc; \
     MVMObject *result; \
     MVMROOT2(tc, a, b, { \
         result = MVM_repr_alloc_init(tc, result_type);\
     }); \
-    bc = get_bigint_body(tc, result); \
-    if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) { \
-        mp_int *tmp[2] = { NULL, NULL }; \
-        mp_int *ia = force_bigint(ba, tmp); \
-        mp_int *ib = force_bigint(bb, tmp); \
-        mp_int *ic = MVM_malloc(sizeof(mp_int)); \
-        mp_init(ic); \
-        two_complement_bitop(ia, ib, ic, mp_##opname); \
-        store_bigint_result(bc, ic); \
-        clear_temp_bigints(tmp, 2); \
-        adjust_nursery(tc, bc); \
-    } \
-    else { \
-        MVMint64 sc; \
-        MVMint64 sa = ba->u.smallint.value; \
-        MVMint64 sb = bb->u.smallint.value; \
-        SMALLINT_OP; \
-        store_int64_result(bc, sc); \
-    } \
-    return result; \
+    {\
+        MVMP6bigintBody *ba = get_bigint_body(tc, a); \
+        MVMP6bigintBody *bb = get_bigint_body(tc, b); \
+        MVMP6bigintBody *bc = get_bigint_body(tc, result); \
+        if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) { \
+            mp_int *ia = force_bigint(tc, ba, 0); \
+            mp_int *ib = force_bigint(tc, bb, 1); \
+            mp_int *ic = MVM_malloc(sizeof(mp_int)); \
+            mp_init(ic); \
+            two_complement_bitop(ia, ib, ic, mp_##opname); \
+            store_bigint_result(bc, ic); \
+            adjust_nursery(tc, bc); \
+        } \
+        else { \
+            MVMint64 sc; \
+            MVMint64 sa = ba->u.smallint.value; \
+            MVMint64 sb = bb->u.smallint.value; \
+            SMALLINT_OP; \
+            store_int64_result(bc, sc); \
+        } \
+        return result; \
+    }\
 }
 
 MVM_BIGINT_UNARY_OP(abs, { sb = labs(sa); })
@@ -471,39 +466,38 @@ MVM_BIGINT_BINARY_OP_SIMPLE(mul, { sc = sa * sb; })
 MVM_BIGINT_BINARY_OP(lcm)
 
 MVMObject *MVM_bigint_gcd(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
-    MVMP6bigintBody *bb = get_bigint_body(tc, b);
-    MVMP6bigintBody *bc;
     MVMObject       *result;
 
     MVMROOT2(tc, a, b, {
         result = MVM_repr_alloc_init(tc, result_type);
     });
 
-    bc = get_bigint_body(tc, result);
+    {
+        MVMP6bigintBody *ba = get_bigint_body(tc, a);
+        MVMP6bigintBody *bb = get_bigint_body(tc, b);
+        MVMP6bigintBody *bc = get_bigint_body(tc, result);
 
-    if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) {
-        mp_int *tmp[2] = { NULL, NULL };
-        mp_int *ia = force_bigint(ba, tmp);
-        mp_int *ib = force_bigint(bb, tmp);
-        mp_int *ic = MVM_malloc(sizeof(mp_int));
-        mp_init(ic);
-        mp_gcd(ia, ib, ic);
-        store_bigint_result(bc, ic);
-        clear_temp_bigints(tmp, 2);
-        adjust_nursery(tc, bc);
-    } else {
-        MVMint32 sa = ba->u.smallint.value;
-        MVMint32 sb = bb->u.smallint.value;
-        MVMint32 t;
-        sa = abs(sa);
-        sb = abs(sb);
-        while (sb != 0) {
-            t  = sb;
-            sb = sa % sb;
-            sa = t;
+        if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) {
+            mp_int *ia = force_bigint(tc, ba, 0);
+            mp_int *ib = force_bigint(tc, bb, 1);
+            mp_int *ic = MVM_malloc(sizeof(mp_int));
+            mp_init(ic);
+            mp_gcd(ia, ib, ic);
+            store_bigint_result(bc, ic);
+            adjust_nursery(tc, bc);
+        } else {
+            MVMint32 sa = ba->u.smallint.value;
+            MVMint32 sb = bb->u.smallint.value;
+            MVMint32 t;
+            sa = abs(sa);
+            sb = abs(sb);
+            while (sb != 0) {
+                t  = sb;
+                sb = sa % sb;
+                sa = t;
+            }
+            store_int64_result(bc, sa);
         }
-        store_int64_result(bc, sa);
     }
 
     return result;
@@ -517,11 +511,9 @@ MVMint64 MVM_bigint_cmp(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
     MVMP6bigintBody *ba = get_bigint_body(tc, a);
     MVMP6bigintBody *bb = get_bigint_body(tc, b);
     if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) {
-        mp_int *tmp[2] = { NULL, NULL };
-        mp_int *ia = force_bigint(ba, tmp);
-        mp_int *ib = force_bigint(bb, tmp);
+        mp_int *ia = force_bigint(tc, ba, 0);
+        mp_int *ib = force_bigint(tc, bb, 1);
         MVMint64 r = (MVMint64)mp_cmp(ia, ib);
-        clear_temp_bigints(tmp, 2);
         return r;
     }
     else {
@@ -532,9 +524,6 @@ MVMint64 MVM_bigint_cmp(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
 }
 
 MVMObject * MVM_bigint_mod(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
-    MVMP6bigintBody *bb = get_bigint_body(tc, b);
-    MVMP6bigintBody *bc;
 
     MVMObject *result;
 
@@ -542,37 +531,40 @@ MVMObject * MVM_bigint_mod(MVMThreadContext *tc, MVMObject *result_type, MVMObje
         result = MVM_repr_alloc_init(tc, result_type);
     });
 
-    bc = get_bigint_body(tc, result);
+    {
+        MVMP6bigintBody *ba = get_bigint_body(tc, a);
+        MVMP6bigintBody *bb = get_bigint_body(tc, b);
+        MVMP6bigintBody *bc;
+        bc = get_bigint_body(tc, result);
 
-    /* XXX the behavior of C's mod operator is not correct
-     * for our purposes. So we rely on mp_mod for all our modulus
-     * calculations for now. */
-    if (1 || MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) {
-        mp_int *tmp[2] = { NULL, NULL };
-        mp_int *ia = force_bigint(ba, tmp);
-        mp_int *ib = force_bigint(bb, tmp);
-        mp_int *ic = MVM_malloc(sizeof(mp_int));
-        int mp_result;
+        /* XXX the behavior of C's mod operator is not correct
+         * for our purposes. So we rely on mp_mod for all our modulus
+         * calculations for now. */
+        if (1 || MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) {
+            mp_int *ia = force_bigint(tc, ba, 0);
+            mp_int *ib = force_bigint(tc, bb, 1);
+            mp_int *ic = MVM_malloc(sizeof(mp_int));
+            int mp_result;
 
-        mp_init(ic);
+            mp_init(ic);
 
-        mp_result = mp_mod(ia, ib, ic);
-        clear_temp_bigints(tmp, 2);
+            mp_result = mp_mod(ia, ib, ic);
 
-        if (mp_result == MP_VAL) {
-            MVM_exception_throw_adhoc(tc, "Division by zero");
+            if (mp_result == MP_VAL) {
+                MVM_exception_throw_adhoc(tc, "Division by zero");
+            }
+            store_bigint_result(bc, ic);
+            adjust_nursery(tc, bc);
+        } else {
+            store_int64_result(bc, ba->u.smallint.value % bb->u.smallint.value);
         }
-        store_bigint_result(bc, ic);
-        adjust_nursery(tc, bc);
-    } else {
-        store_int64_result(bc, ba->u.smallint.value % bb->u.smallint.value);
     }
 
     return result;
 }
 
 MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
+    MVMP6bigintBody *ba;
     MVMP6bigintBody *bb = get_bigint_body(tc, b);
     MVMP6bigintBody *bc;
     mp_int *ia, *ib, *ic;
@@ -592,6 +584,9 @@ MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         result = MVM_repr_alloc_init(tc, result_type);
     });
 
+    ba = get_bigint_body(tc, a);
+    bb = get_bigint_body(tc, b);
+
     bc = get_bigint_body(tc, result);
 
     /* we only care about MP_LT or !MP_LT, so we give MP_GT even for 0. */
@@ -607,9 +602,8 @@ MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
     }
 
     if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) {
-        mp_int *tmp[2] = { NULL, NULL };
-        ia = force_bigint(ba, tmp);
-        ib = force_bigint(bb, tmp);
+        ia = force_bigint(tc, ba, 0);
+        ib = force_bigint(tc, bb, 1);
 
         ic = MVM_malloc(sizeof(mp_int));
         mp_init(ic);
@@ -624,7 +618,6 @@ MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
             if (mp_result == MP_VAL) {
                 mp_clear(&remainder);
                 mp_clear(&intermediate);
-                clear_temp_bigints(tmp, 2);
                 MVM_exception_throw_adhoc(tc, "Division by zero");
             }
             if (mp_iszero(&remainder) == 0) {
@@ -637,12 +630,10 @@ MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         } else {
             mp_result = mp_div(ia, ib, ic, NULL);
             if (mp_result == MP_VAL) {
-                clear_temp_bigints(tmp, 2);
                 MVM_exception_throw_adhoc(tc, "Division by zero");
             }
         }
         store_bigint_result(bc, ic);
-        clear_temp_bigints(tmp, 2);
         adjust_nursery(tc, bc);
     } else {
         MVMint32 num   = ba->u.smallint.value;
@@ -672,9 +663,8 @@ MVMObject * MVM_bigint_pow(MVMThreadContext *tc, MVMObject *a, MVMObject *b,
     MVMP6bigintBody *bb = get_bigint_body(tc, b);
     MVMObject       *r  = NULL;
 
-    mp_int *tmp[2] = { NULL, NULL };
-    mp_int *base        = force_bigint(ba, tmp);
-    mp_int *exponent    = force_bigint(bb, tmp);
+    mp_int *base        = force_bigint(tc, ba, 0);
+    mp_int *exponent    = force_bigint(tc, bb, 1);
     mp_digit exponent_d = 0;
 
     if (mp_iszero(exponent) || (MP_EQ == mp_cmp_d(base, 1))) {
@@ -714,16 +704,15 @@ MVMObject * MVM_bigint_pow(MVMThreadContext *tc, MVMObject *a, MVMObject *b,
         }
     }
     else {
-        MVMnum64 f_base = mp_get_double(base, 0);
-        MVMnum64 f_exp = mp_get_double(exponent, 0);
+        MVMnum64 f_base = MVM_mp_get_double_shift(base, 0);
+        MVMnum64 f_exp = MVM_mp_get_double_shift(exponent, 0);
         r = MVM_repr_box_num(tc, num_type, pow(f_base, f_exp));
     }
-    clear_temp_bigints(tmp, 2);
     return r;
 }
 
 MVMObject *MVM_bigint_shl(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMint64 n) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
+    MVMP6bigintBody *ba;
     MVMP6bigintBody *bb;
     MVMObject       *result;
 
@@ -731,16 +720,15 @@ MVMObject *MVM_bigint_shl(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         result = MVM_repr_alloc_init(tc, result_type);
     });
 
+    ba = get_bigint_body(tc, a);
     bb = get_bigint_body(tc, result);
 
     if (MVM_BIGINT_IS_BIG(ba) || n >= 31) {
-        mp_int *tmp[1] = { NULL };
-        mp_int *ia = force_bigint(ba, tmp);
+        mp_int *ia = force_bigint(tc, ba, 0);
         mp_int *ib = MVM_malloc(sizeof(mp_int));
         mp_init(ib);
         two_complement_shl(ib, ia, n);
         store_bigint_result(bb, ib);
-        clear_temp_bigints(tmp, 1);
         adjust_nursery(tc, bb);
     } else {
         MVMint64 value;
@@ -765,7 +753,7 @@ int BIGINT_IS_NEGATIVE (MVMP6bigintBody *ba) {
     }
 }
 MVMObject *MVM_bigint_shr(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMint64 n) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
+    MVMP6bigintBody *ba;
     MVMP6bigintBody *bb;
     MVMObject       *result;
 
@@ -773,16 +761,15 @@ MVMObject *MVM_bigint_shr(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         result = MVM_repr_alloc_init(tc, result_type);
     });
 
+    ba = get_bigint_body(tc, a);
     bb = get_bigint_body(tc, result);
 
     if (MVM_BIGINT_IS_BIG(ba) || n < 0) {
-        mp_int *tmp[1] = { NULL };
-        mp_int *ia = force_bigint(ba, tmp);
+        mp_int *ia = force_bigint(tc, ba, 0);
         mp_int *ib = MVM_malloc(sizeof(mp_int));
         mp_init(ib);
         two_complement_shl(ib, ia, -n);
         store_bigint_result(bb, ib);
-        clear_temp_bigints(tmp, 1);
         adjust_nursery(tc, bb);
     } else if (n >= 32) {
         store_int64_result(bb, BIGINT_IS_NEGATIVE(ba) ? -1 : 0);
@@ -796,7 +783,7 @@ MVMObject *MVM_bigint_shr(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
 }
 
 MVMObject *MVM_bigint_not(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
+    MVMP6bigintBody *ba;
     MVMP6bigintBody *bb;
     MVMObject       *result;
 
@@ -804,6 +791,7 @@ MVMObject *MVM_bigint_not(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         result = MVM_repr_alloc_init(tc, result_type);
     });
 
+    ba = get_bigint_body(tc, a);
     bb = get_bigint_body(tc, result);
 
     if (MVM_BIGINT_IS_BIG(ba)) {
@@ -831,11 +819,9 @@ MVMObject *MVM_bigint_expmod(MVMThreadContext *tc, MVMObject *result_type, MVMOb
     MVMP6bigintBody *bd;
     MVMObject       *result;
 
-    mp_int *tmp[3] = { NULL, NULL, NULL };
-
-    mp_int *ia = force_bigint(ba, tmp);
-    mp_int *ib = force_bigint(bb, tmp);
-    mp_int *ic = force_bigint(bc, tmp);
+    mp_int *ia = force_bigint(tc, ba, 0);
+    mp_int *ib = force_bigint(tc, bb, 1);
+    mp_int *ic = force_bigint(tc, bc, 2);
     mp_int *id = MVM_malloc(sizeof(mp_int));
     mp_init(id);
 
@@ -847,7 +833,6 @@ MVMObject *MVM_bigint_expmod(MVMThreadContext *tc, MVMObject *result_type, MVMOb
 
     mp_exptmod(ia, ib, ic, id);
     store_bigint_result(bd, id);
-    clear_temp_bigints(tmp, 3);
     adjust_nursery(tc, bd);
 
     return result;
@@ -875,7 +860,10 @@ MVMObject * MVM_coerce_sI(MVMThreadContext *tc, MVMString *s, MVMObject *type) {
     char *buf = NULL;
     int is_malloced = 0;
     MVMStringIndex i;
-    MVMObject *a = MVM_repr_alloc_init(tc, type);
+    MVMObject *a;
+    MVMROOT(tc, s, {
+        a = MVM_repr_alloc_init(tc, type);
+    });
     if (s->body.num_graphs < 120) {
         buf = alloca(s->body.num_graphs + 1);
     }
@@ -1082,7 +1070,7 @@ MVMnum64 MVM_bigint_to_num(MVMThreadContext *tc, MVMObject *a) {
 
     if (MVM_BIGINT_IS_BIG(ba)) {
         mp_int *ia = ba->u.bigint;
-        return mp_get_double(ia, 0);
+        return MVM_mp_get_double_shift(ia, 0);
     } else {
         return (double)ba->u.smallint.value;
     }
@@ -1104,9 +1092,8 @@ MVMnum64 MVM_bigint_div_num(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
     MVMnum64 c;
 
     if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) {
-        mp_int *tmp[2] = { NULL, NULL };
-        mp_int *ia = force_bigint(ba, tmp);
-        mp_int *ib = force_bigint(bb, tmp);
+        mp_int *ia = force_bigint(tc, ba, 0);
+        mp_int *ib = force_bigint(tc, bb, 1);
 
         mp_clamp(ib);
         if (ib->used == 0) { /* zero-denominator special case */
@@ -1132,16 +1119,25 @@ MVMnum64 MVM_bigint_div_num(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
             if (mp_div(&scaled, ib, &scaled, NULL) != MP_OKAY)
                 MVM_exception_throw_adhoc(tc,
                     "Failed to preform bigint division");
-            c = mp_get_double(&scaled, bbits);
+            c = MVM_mp_get_double_shift(&scaled, bbits);
             mp_clear(&scaled);
         }
-        clear_temp_bigints(tmp, 2);
     } else {
         c = (double)ba->u.smallint.value / (double)bb->u.smallint.value;
     }
     return c;
 }
 
+
+/* 
+    The old version of LibTomMath has it publically defined the new one not,
+    so we can take the (non)existance as a marker.
+ */
+#ifndef MP_GEN_RANDOM_MAX
+#define MP_GEN_RANDOM_MAX MP_MASK
+#define MP_NEW_LTM_VERSION
+#endif
+
 MVMObject * MVM_bigint_rand(MVMThreadContext *tc, MVMObject *type, MVMObject *b) {
     MVMObject *result;
     MVMP6bigintBody *ba;
@@ -1164,7 +1160,14 @@ MVMObject * MVM_bigint_rand(MVMThreadContext *tc, MVMObject *type, MVMObject *b)
 
     if (use_small_arithmetic) {
         if (MP_GEN_RANDOM_MAX >= abs(smallint_max)) {
-            mp_digit result_int = MP_GEN_RANDOM();
+            mp_digit result_int;
+#ifdef MP_NEW_LTM_VERSION
+            mp_digit p = MP_GEN_RANDOM_MAX;
+            mp_rand_digit(&p);
+            result_int = p;
+#else
+            result_int = MP_GEN_RANDOM();
+#endif
             result_int = result_int % smallint_max;
             if(have_to_negate)
                 result_int *= -1;
@@ -1181,9 +1184,8 @@ MVMObject * MVM_bigint_rand(MVMThreadContext *tc, MVMObject *type, MVMObject *b)
     }
 
     if (!use_small_arithmetic) {
-        mp_int *tmp[1] = { NULL };
         mp_int *rnd = MVM_malloc(sizeof(mp_int));
-        mp_int *max = force_bigint(bb, tmp);
+        mp_int *max = force_bigint(tc, bb, 0);
 
         MVMROOT2(tc, type, b, {
             result = MVM_repr_alloc_init(tc, type);
@@ -1196,7 +1198,6 @@ MVMObject * MVM_bigint_rand(MVMThreadContext *tc, MVMObject *type, MVMObject *b)
 
         mp_mod(rnd, max, rnd);
         store_bigint_result(ba, rnd);
-        clear_temp_bigints(tmp, 1);
         adjust_nursery(tc, ba);
     }
 
@@ -1210,16 +1211,15 @@ MVMint64 MVM_bigint_is_prime(MVMThreadContext *tc, MVMObject *a, MVMint64 b) {
     MVMP6bigintBody *ba = get_bigint_body(tc, a);
 
     if (MVM_BIGINT_IS_BIG(ba) || ba->u.smallint.value != 1) {
-        mp_int *tmp[1] = { NULL };
-        mp_int *ia = force_bigint(ba, tmp);
+        mp_int *ia = force_bigint(tc, ba, 0);
         if (mp_cmp_d(ia, 1) == MP_EQ) {
-            clear_temp_bigints(tmp, 1);
             return 0;
         }
         else {
-            int result;
-            mp_prime_is_prime(ia, b, &result);
-            clear_temp_bigints(tmp, 1);
+            int result, err;
+            err = mp_prime_is_prime(ia, b, &result);
+            if (err != MP_OKAY)
+                MVM_exception_throw_adhoc(tc, "Invalid number of rounds (%ld), valid range is 0..%d\n", b, PRIME_SIZE);
             return result;
         }
     } else {
