@@ -2,26 +2,20 @@
 
 typedef struct {
     MVMuint32 items;
-    MVMuint32 start;
     MVMuint32 alloc;
 
     MVMProfileCallNode **list;
 } NodeWorklist;
 
 static void add_node(MVMThreadContext *tc, NodeWorklist *list, MVMProfileCallNode *node) {
-    if (list->start + list->items + 1 < list->alloc) {
-        /* Add at the end */
-        list->items++;
-        list->list[list->start + list->items] = node;
-    } else if (list->start > 0) {
-        /* End reached, add to the start now */
-        list->start--;
-        list->list[list->start] = node;
-    } else {
+    if (list->items == list->alloc) {
         /* Filled up the whole list. Make it bigger */
         list->alloc *= 2;
         list->list = MVM_realloc(list->list, list->alloc * sizeof(MVMProfileCallNode *));
     }
+    /* Add at the end */
+    list->list[list->items] = node;
+    list->items++;
 }
 
 static MVMProfileCallNode *take_node(MVMThreadContext *tc, NodeWorklist *list) {
@@ -29,13 +23,8 @@ static MVMProfileCallNode *take_node(MVMThreadContext *tc, NodeWorklist *list) {
     if (list->items == 0) {
         MVM_panic(1, "profiler: tried to take a node from an empty node worklist");
     }
-    if (list->start > 0) {
-        result = list->list[list->start];
-        list->start++;
-    } else {
-        result = list->list[list->start + list->items];
-        list->items--;
-    }
+    list->items--;
+    result = list->list[list->items];
     return result;
 }
 
@@ -323,6 +312,7 @@ typedef struct {
     MVMString *allocations;
     MVMString *spesh;
     MVMString *jit;
+    MVMString *replaced;
     MVMString *type;
     MVMString *count;
     MVMString *gcs;
@@ -335,6 +325,7 @@ typedef struct {
     MVMString *promoted_bytes;
     MVMString *promoted_bytes_unmanaged;
     MVMString *gen2_roots;
+    MVMString *stolen_gen2_roots;
     MVMString *start_time;
     MVMString *first_entry_time;
     MVMString *osr;
@@ -346,6 +337,10 @@ typedef struct {
     MVMString *managed_size;
     MVMString *has_unmanaged_data;
     MVMString *repr;
+    MVMString *deallocs;
+    MVMString *nursery_fresh;
+    MVMString *nursery_seen;
+    MVMString *gen2;
 } ProfDumpStrs;
 
 typedef struct {
@@ -385,19 +380,40 @@ static void bind_extra_info(MVMThreadContext *tc, MVMObject *storage, MVMString
 static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds, const MVMProfileCallNode *pcn, MVMObject *types_array);
 static MVMObject * dump_call_graph_node_loop(ProfTcPdsStruct *tcpds, const MVMProfileCallNode *pcn) {
     MVMuint32 i;
+    MVMuint64 exclusive_time = pcn->total_time;
+    MVMuint64 overhead       = pcn->total_entries * tcpds->tc->instance->profiling_overhead;
     MVMObject *node_hash;
 
+    /* Subtract profiling overhead, unless that would underflow, in which case just clamp to 0. */
+    if (exclusive_time - overhead > exclusive_time)
+        exclusive_time = 0;
+    else
+        exclusive_time -= overhead;
+
     node_hash = dump_call_graph_node(tcpds->tc, tcpds->pds, pcn, tcpds->types_array);
 
     /* Visit successors in the call graph, dumping them and working out the
      * exclusive time. */
     if (pcn->num_succ) {
-        MVMObject *callees        = new_array(tcpds->tc);
-        MVMuint64  exclusive_time = pcn->total_time;
+        MVMObject *callees = new_array(tcpds->tc);
         for (i = 0; i < pcn->num_succ; i++) {
+            MVMuint64 succ_exclusive_time = pcn->succ[i]->total_time;
+            MVMuint64 succ_overhead       = pcn->succ[i]->total_entries * tcpds->tc->instance->profiling_overhead;
+
             MVM_repr_push_o(tcpds->tc, callees,
                 dump_call_graph_node_loop(tcpds, pcn->succ[i]));
-            exclusive_time -= pcn->succ[i]->total_time;
+
+            /* Subtract profiling overhead, unless that would underflow, in which case just clamp to 0. */
+            if (succ_exclusive_time - succ_overhead > succ_exclusive_time)
+                succ_exclusive_time = 0;
+            else
+                succ_exclusive_time -= succ_overhead;
+
+            /* Subtract profiling overhead, unless that would underflow, in which case just clamp to 0. */
+            if (exclusive_time - succ_exclusive_time > exclusive_time)
+                exclusive_time = 0;
+            else
+                exclusive_time -= succ_exclusive_time;
         }
         MVM_repr_bind_key_o(tcpds->tc, node_hash, tcpds->pds->exclusive_time,
             box_i(tcpds->tc, exclusive_time / 1000));
@@ -405,12 +421,24 @@ static MVMObject * dump_call_graph_node_loop(ProfTcPdsStruct *tcpds, const MVMPr
     }
     else {
         MVM_repr_bind_key_o(tcpds->tc, node_hash, tcpds->pds->exclusive_time,
-            box_i(tcpds->tc, pcn->total_time / 1000));
+            box_i(tcpds->tc, exclusive_time / 1000));
     }
 
     return node_hash;
 }
 
+static void add_type_to_types_array(MVMThreadContext *tc, ProfDumpStrs *pds, MVMObject *type, MVMObject *types_array) {
+    MVMObject *type_info  = insert_if_not_exists(tc, pds, types_array, (MVMint64)(uintptr_t)type);
+
+    if (type_info) {
+        bind_extra_info(tc, type_info, pds->managed_size, box_i(tc, STABLE(type)->size));
+        if (REPR(type)->unmanaged_size)
+            bind_extra_info(tc, type_info, pds->has_unmanaged_data, box_i(tc, 1));
+        bind_extra_info(tc, type_info, pds->type, type);
+        bind_extra_info(tc, type_info, pds->repr, box_s(tc, str(tc, REPR(type)->name)));
+    }
+}
+
 /* Dumps a call graph node. */
 static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
                                         const MVMProfileCallNode *pcn, MVMObject *types_array) {
@@ -446,7 +474,7 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
         /* Use static frame memory address to get a unique ID. */
         MVM_repr_bind_key_o(tc, node_hash, pds->id,
             box_i(tc, (MVMint64)(uintptr_t)pcn->sf));
-    } else {
+    } else if (pcn->native_target_name) {
         MVMString *function_name_string =
             MVM_string_utf8_c8_decode(tc, tc->instance->VMString,
                                       pcn->native_target_name, strlen(pcn->native_target_name));
@@ -463,6 +491,18 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
         MVM_repr_bind_key_o(tc, node_hash, pds->id,
             box_i(tc, (MVMint64)(uintptr_t)pcn->native_target_name));
     }
+    else {
+        MVM_repr_bind_key_o(tc, node_hash, pds->name,
+            box_s(tc, pds->call_graph));
+        MVM_repr_bind_key_o(tc, node_hash, pds->file,
+            box_s(tc, pds->call_graph));
+
+        MVM_repr_bind_key_o(tc, node_hash, pds->line,
+            box_i(tc, -10));
+
+        MVM_repr_bind_key_o(tc, node_hash, pds->id,
+            box_i(tc, (MVMint64)(uintptr_t)pcn));
+    }
 
     /* Entry counts. */
     if (pcn->total_entries)
@@ -509,15 +549,8 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
             MVMProfileAllocationCount *alloc = &pcn->alloc[i];
 
             MVMObject *type       = pcn->alloc[i].type;
-            MVMObject *type_info  = insert_if_not_exists(tc, pds, types_array, (MVMint64)(uintptr_t)type);
 
-            if (type_info) {
-                bind_extra_info(tc, type_info, pds->managed_size, box_i(tc, STABLE(type)->size));
-                if (REPR(type)->unmanaged_size)
-                    bind_extra_info(tc, type_info, pds->has_unmanaged_data, box_i(tc, 1));
-                bind_extra_info(tc, type_info, pds->type, type);
-                bind_extra_info(tc, type_info, pds->repr, box_s(tc, str(tc, REPR(type)->name)));
-            }
+            add_type_to_types_array(tc, pds, type, types_array);
 
             MVM_repr_bind_key_o(tc, alloc_info, pds->id, box_i(tc, (MVMint64)(uintptr_t)type));
             if (alloc->allocations_spesh)
@@ -530,6 +563,9 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
                 box_i(tc, alloc->allocations_interp
                           + alloc->allocations_spesh
                           + alloc->allocations_jit));
+            if (alloc->scalar_replaced)
+                MVM_repr_bind_key_o(tc, alloc_info, pds->replaced,
+                    box_i(tc, alloc->scalar_replaced));
             MVM_repr_push_o(tc, alloc_list, alloc_info);
         }
     }
@@ -571,27 +607,63 @@ static MVMObject * dump_thread_data(MVMThreadContext *tc, ProfDumpStrs *pds,
 
     /* Add GCs. */
     for (i = 0; i < ptd->num_gcs; i++) {
+        MVMuint32 tid;
+
         MVMObject *gc_hash = new_hash(tc);
+
+        MVMProfileGC *gc = &(ptd->gcs[i]);
+
         MVM_repr_bind_key_o(tc, gc_hash, pds->time,
-            box_i(tc, ptd->gcs[i].time / 1000));
+            box_i(tc, gc->time / 1000));
         MVM_repr_bind_key_o(tc, gc_hash, pds->full,
-            box_i(tc, ptd->gcs[i].full));
+            box_i(tc, gc->full));
         MVM_repr_bind_key_o(tc, gc_hash, pds->sequence,
-            box_i(tc, ptd->gcs[i].gc_seq_num - 1));
+            box_i(tc, gc->gc_seq_num - 1));
         MVM_repr_bind_key_o(tc, gc_hash, pds->responsible,
-            box_i(tc, ptd->gcs[i].responsible));
+            box_i(tc, gc->responsible));
         MVM_repr_bind_key_o(tc, gc_hash, pds->cleared_bytes,
-            box_i(tc, ptd->gcs[i].cleared_bytes));
+            box_i(tc, gc->cleared_bytes));
         MVM_repr_bind_key_o(tc, gc_hash, pds->retained_bytes,
-            box_i(tc, ptd->gcs[i].retained_bytes));
+            box_i(tc, gc->retained_bytes));
         MVM_repr_bind_key_o(tc, gc_hash, pds->promoted_bytes,
-            box_i(tc, ptd->gcs[i].promoted_bytes));
+            box_i(tc, gc->promoted_bytes));
         MVM_repr_bind_key_o(tc, gc_hash, pds->promoted_bytes_unmanaged,
-            box_i(tc, ptd->gcs[i].promoted_unmanaged_bytes));
+            box_i(tc, gc->promoted_unmanaged_bytes));
         MVM_repr_bind_key_o(tc, gc_hash, pds->gen2_roots,
-            box_i(tc, ptd->gcs[i].num_gen2roots));
+            box_i(tc, gc->num_gen2roots));
+        MVM_repr_bind_key_o(tc, gc_hash, pds->stolen_gen2_roots,
+            box_i(tc, gc->num_stolen_gen2roots));
         MVM_repr_bind_key_o(tc, gc_hash, pds->start_time,
-            box_i(tc, (ptd->gcs[i].abstime - absolute_start_time) / 1000));
+            box_i(tc, (gc->abstime - absolute_start_time) / 1000));
+
+        if (gc->num_dealloc) {
+            MVMObject *deallocs_array = new_array(tc);
+            MVM_repr_bind_key_o(tc, gc_hash, pds->deallocs,
+                deallocs_array);
+
+            for (tid = 0; tid < gc->num_dealloc; tid++) {
+                MVMProfileDeallocationCount *dealloc = &(gc->deallocs[tid]);
+
+                MVMObject *type_hash = new_hash(tc);
+
+                if (dealloc->deallocs_nursery_fresh)
+                    MVM_repr_bind_key_o(tc, type_hash, pds->nursery_fresh,
+                            box_i(tc, dealloc->deallocs_nursery_fresh));
+                if (dealloc->deallocs_nursery_seen)
+                    MVM_repr_bind_key_o(tc, type_hash, pds->nursery_seen,
+                            box_i(tc, dealloc->deallocs_nursery_seen));
+                if (dealloc->deallocs_gen2)
+                    MVM_repr_bind_key_o(tc, type_hash, pds->gen2,
+                            box_i(tc, dealloc->deallocs_gen2));
+
+                add_type_to_types_array(tc, pds, dealloc->type, types_data);
+
+                MVM_repr_bind_key_o(tc, type_hash, pds->id, box_i(tc, (MVMint64)(uintptr_t)(dealloc->type)));
+
+                MVM_repr_push_o(tc, deallocs_array, type_hash);
+            }
+        }
+
         MVM_repr_push_o(tc, thread_gcs, gc_hash);
     }
     MVM_repr_bind_key_o(tc, thread_hash, pds->gcs, thread_gcs);
@@ -644,6 +716,7 @@ void MVM_profile_dump_instrumented_data(MVMThreadContext *tc) {
         pds.count           = str(tc, "count");
         pds.spesh           = str(tc, "spesh");
         pds.jit             = str(tc, "jit");
+        pds.replaced        = str(tc, "replaced");
         pds.gcs             = str(tc, "gcs");
         pds.time            = str(tc, "time");
         pds.full            = str(tc, "full");
@@ -662,7 +735,12 @@ void MVM_profile_dump_instrumented_data(MVMThreadContext *tc) {
         pds.thread          = str(tc, "thread");
         pds.native_lib      = str(tc, "native library");
         pds.managed_size    = str(tc, "managed_size");
+        pds.deallocs        = str(tc, "deallocs");
+        pds.nursery_fresh   = str(tc, "nursery_fresh");
+        pds.nursery_seen    = str(tc, "nursery_seen");
+        pds.gen2            = str(tc, "gen2");
 
+        pds.stolen_gen2_roots  = str(tc, "stolen_gen2_roots");
         pds.has_unmanaged_data = str(tc, "has_unmanaged_data");
         pds.repr               = str(tc, "repr");
 
@@ -672,10 +750,11 @@ void MVM_profile_dump_instrumented_data(MVMThreadContext *tc) {
 
         MVM_repr_push_o(tc, tc->prof_data->collected_data, types_array);
 
-        MVM_repr_push_o(tc, tc->prof_data->collected_data, dump_thread_data(tc, &pds, tc, tc->prof_data, types_array));
         while (tc->prof_data->current_call)
             MVM_profile_log_exit(tc);
 
+        MVM_repr_push_o(tc, tc->prof_data->collected_data, dump_thread_data(tc, &pds, tc, tc->prof_data, types_array));
+
         /* Get all thread's data */
         thread = tc->instance->threads;
 
@@ -745,12 +824,21 @@ static void mark_call_graph_node(MVMThreadContext *tc, MVMProfileCallNode *node,
     for (i = 0; i < node->num_succ; i++)
         add_node(tc, nodelist, node->succ[i]);
 }
+static void mark_gc_entries(MVMThreadContext *tc, MVMProfileThreadData *ptd, MVMGCWorklist *worklist) {
+    MVMuint32 gci;
+    for (gci = 0; gci < ptd->num_gcs; gci++) {
+        MVMProfileGC *gc = &(ptd->gcs[gci]);
+        MVMuint32 ti;
+        for (ti = 0; ti < gc->num_dealloc; ti++) {
+            MVM_gc_worklist_add(tc, worklist, &(gc->deallocs[ti].type));
+        }
+    }
+}
 void MVM_profile_instrumented_mark_data(MVMThreadContext *tc, MVMGCWorklist *worklist) {
     if (tc->prof_data) {
         /* Allocate our worklist on the stack. */
         NodeWorklist nodelist;
         nodelist.items = 0;
-        nodelist.start = 0;
         nodelist.alloc = 256;
         nodelist.list = MVM_malloc(nodelist.alloc * sizeof(MVMProfileCallNode *));
 
@@ -764,6 +852,38 @@ void MVM_profile_instrumented_mark_data(MVMThreadContext *tc, MVMGCWorklist *wor
 
         MVM_gc_worklist_add(tc, worklist, &(tc->prof_data->collected_data));
 
+        mark_gc_entries(tc, tc->prof_data, worklist);
+
         MVM_free(nodelist.list);
     }
 }
+
+static void dump_callgraph_node(MVMThreadContext *tc, MVMProfileCallNode *n, MVMuint16 depth) {
+    MVMuint16 dc = depth;
+    MVMuint32 idx;
+    char *name;
+
+    for (dc = depth; dc > 0; dc--) {
+        fputc(' ', stderr);
+    }
+
+    if (n->sf)
+        name = MVM_string_utf8_encode_C_string(tc, n->sf->body.name);
+
+    fprintf(stderr, "+ [%3d] %s\n", n->num_succ, name);
+    MVM_free(name);
+
+    for (idx = 0; idx < n->num_succ; idx++) {
+        dump_callgraph_node(tc, n->succ[idx], depth + 1);
+    }
+}
+
+void MVM_dump_callgraph(MVMThreadContext *tc) {
+    MVMProfileThreadData *ptd = tc->prof_data;
+    MVMProfileCallNode *pcn = ptd->call_graph;
+    fprintf(stderr, "\n----------\nCall Graph of TC %p\n\n", tc);
+    if (pcn && pcn->num_succ) {
+        dump_callgraph_node(tc, pcn, 0);
+    }
+    fprintf(stderr, "\n<<<<<<\n");
+}
