@@ -1,5 +1,22 @@
 #include "moar.h"
 
+#define CONFPROG_DEBUG_LEVEL_PROFILER_RESULTS 4
+
+static void debugprint(MVMuint8 active, MVMThreadContext *tc, const char *str, ...) {
+    va_list args;
+    va_start(args, str);
+
+    if (active) {
+        fprintf(stderr, "%p: ", tc);
+        vfprintf(stderr, str, args);
+        fprintf(stderr, "\n");
+    }
+
+    va_end(args);
+}
+
+#define DEBUG_LVL(level) ((debug_level) & CONFPROG_DEBUG_LEVEL_ ## level)
+
 /* Gets the current thread's profiling data structure, creating it if needed. */
 static MVMProfileThreadData * get_thread_data(MVMThreadContext *tc) {
     if (!tc->prof_data) {
@@ -34,51 +51,137 @@ static MVMProfileCallNode *make_new_pcn(MVMProfileThreadData *ptd, MVMuint64 cur
 
 /* Log that we're entering a new frame. */
 void MVM_profile_log_enter(MVMThreadContext *tc, MVMStaticFrame *sf, MVMuint64 mode) {
-    MVMProfileThreadData *ptd = get_thread_data(tc);
+    if (tc->instance->profiling) {
+        MVMProfileThreadData *ptd = get_thread_data(tc);
 
-    MVMuint64 current_hrtime = uv_hrtime();
+        MVMuint64 current_hrtime = uv_hrtime();
+        
+        MVMuint8 was_entered_via_confprog = 0;
 
-    /* Try to locate the entry node, if it's in the call graph already. */
-    MVMProfileCallNode *pcn = NULL;
-    MVMuint32 i;
-    if (ptd->current_call)
-        for (i = 0; i < ptd->current_call->num_succ; i++)
-            if (ptd->current_call->succ[i]->sf == sf)
-                pcn = ptd->current_call->succ[i];
+        /* Try to locate the entry node, if it's in the call graph already. */
+        MVMProfileCallNode *pcn = NULL;
+        MVMuint32 i;
+        if (ptd->current_call)
+            for (i = 0; i < ptd->current_call->num_succ; i++)
+                if (ptd->current_call->succ[i]->sf == sf)
+                    pcn = ptd->current_call->succ[i];
 
-    /* If we didn't find a call graph node, then create one and add it to the
-     * graph. */
-    if (!pcn) {
-        pcn = make_new_pcn(ptd, current_hrtime);
-        pcn->sf = sf;
+        if (MVM_UNLIKELY(!ptd->current_call)) {
+            /* debug_level used by the debugprint macros */
+            MVMuint8 debug_level = tc->instance->confprog ? tc->instance->confprog->debugging_level : 0;
+            MVMuint8 has_confprogs = tc->instance->confprog && (MVM_confprog_has_entrypoint(tc, MVM_PROGRAM_ENTRYPOINT_PROFILER_STATIC)
+                        || MVM_confprog_has_entrypoint(tc, MVM_PROGRAM_ENTRYPOINT_PROFILER_STATIC));
+            /* At the very beginning of profiling with a confprog set, the
+             * root call_graph node needs to be created. */
+            if (has_confprogs && !ptd->call_graph) {
+                debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "initialized initial call graph node\n");
+                ptd->call_graph = MVM_calloc(sizeof(MVMProfileCallNode), 1);
+            }
+            /* In that case, we've got to check if the SF in question is
+             * desired as an entry point */
+            if (sf->body.instrumentation && has_confprogs) {
+                MVMStaticFrameInstrumentation *instrumentation = sf->body.instrumentation;
+                if (instrumentation->profiler_confprog_result == MVM_CONFPROG_SF_RESULT_NEVER) {
+                    goto confprog_refused_enter;
+                }
+                else if (instrumentation->profiler_confprog_result == MVM_CONFPROG_SF_RESULT_TO_BE_DETERMINED) {
+                    if (MVM_confprog_has_entrypoint(tc, MVM_PROGRAM_ENTRYPOINT_PROFILER_STATIC)) {
+                        MVMuint8 result;
+                        debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "running 'profiler_static' entrypoint in confprog");
+                        result = MVM_confprog_run(tc, (void *)sf, MVM_PROGRAM_ENTRYPOINT_PROFILER_STATIC, MVM_CONFPROG_SF_RESULT_ALWAYS);
+                        instrumentation->profiler_confprog_result = result;
+                        if (result == MVM_CONFPROG_SF_RESULT_NEVER) {
+                            debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog for SF resulted in 'never profile'");
+                            goto confprog_refused_enter;
+                        }
+                        if (DEBUG_LVL(PROFILER_RESULTS)) {
+                            switch (result) {
+                                case MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_YES:
+                                case MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_NO:
+                                    debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog result: run dynamic program with default value '%s' (result value: %d)", result == MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_YES ? "yes" : "no", result);
+                                    break;
+                                case MVM_CONFPROG_SF_RESULT_ALWAYS:
+                                    debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog result: always profile from this SF (result value: %d)", result);
+                                    break;
+                                case MVM_CONFPROG_SF_RESULT_TO_BE_DETERMINED:
+                                    debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog result: to be determined (result value %d) - will enter this time, but re-run next time", result);
+                                    break;
+                                default:
+                                    debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  unrecognized result value from confprog: %d", result);
+                                    break;
+                            }
+                            debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog for SF resulted in 'never profile'");
+                        }
+                    }
+                    else if (MVM_confprog_has_entrypoint(tc, MVM_PROGRAM_ENTRYPOINT_PROFILER_DYNAMIC)) {
+                        instrumentation->profiler_confprog_result = MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_YES;
+                    }
+                    else {
+                        MVM_oops(tc, "here we are, what now?");
+                    }
+                }
+                if (instrumentation->profiler_confprog_result == MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_NO
+                        || instrumentation->profiler_confprog_result == MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_YES) {
+                    debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "running 'profiler_dynamic' entrypoint in confprog\n");
+                    if (MVM_confprog_has_entrypoint(tc, MVM_PROGRAM_ENTRYPOINT_PROFILER_DYNAMIC)) {
+                        if (!MVM_confprog_run(tc, (void *)tc->cur_frame, MVM_PROGRAM_ENTRYPOINT_PROFILER_DYNAMIC, instrumentation->profiler_confprog_result == MVM_CONFPROG_SF_RESULT_DYNAMIC_SUGGEST_YES)) {
+                            debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog result: no.\n");
+                            goto confprog_refused_enter;
+                        }
+                        debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  confprog result: yes.\n");
+                    }
+                    else {
+                        debugprint(DEBUG_LVL(PROFILER_RESULTS), tc, "  static confprog said to run dynamic confprog, but none is installed - not profiling.\n");
+                        goto confprog_refused_enter;
+                    }
+                }
+                was_entered_via_confprog = 1;
+            }
+        }
+        /*else {*/
+            /*fprintf(stderr, "there actually was a current_call. also, pcn is %p\n", pcn);*/
+        /*}*/
+
+        /* If we didn't find a call graph node, then create one and add it to the
+         * graph. */
+        if (!pcn) {
+            if (was_entered_via_confprog)
+                ptd->current_call = ptd->call_graph;
+            pcn = make_new_pcn(ptd, current_hrtime);
+            pcn->sf = sf;
+        }
+
+        /* Increment entry counts. */
+        pcn->total_entries++;
+        switch (mode) {
+            case MVM_PROFILE_ENTER_SPESH:
+                pcn->specialized_entries++;
+                break;
+            case MVM_PROFILE_ENTER_SPESH_INLINE:
+                pcn->specialized_entries++;
+                pcn->inlined_entries++;
+                break;
+            case MVM_PROFILE_ENTER_JIT:
+                pcn->jit_entries++;
+                break;
+            case MVM_PROFILE_ENTER_JIT_INLINE:
+                pcn->jit_entries++;
+                pcn->inlined_entries++;
+                break;
+        }
+        pcn->entry_mode =  mode;
+
+        /* Log entry time; clear skip time. */
+        pcn->cur_entry_time = current_hrtime;
+        pcn->cur_skip_time  = 0;
+
+        /* The current call graph node becomes this one. */
+        ptd->current_call = pcn;
+
+        return;
+    confprog_refused_enter:
+        ptd->non_calltree_depth++;
     }
-
-    /* Increment entry counts. */
-    pcn->total_entries++;
-    switch (mode) {
-        case MVM_PROFILE_ENTER_SPESH:
-            pcn->specialized_entries++;
-            break;
-        case MVM_PROFILE_ENTER_SPESH_INLINE:
-            pcn->specialized_entries++;
-            pcn->inlined_entries++;
-            break;
-        case MVM_PROFILE_ENTER_JIT:
-            pcn->jit_entries++;
-            break;
-        case MVM_PROFILE_ENTER_JIT_INLINE:
-            pcn->jit_entries++;
-            pcn->inlined_entries++;
-            break;
-    }
-    pcn->entry_mode = mode;
-
-    /* Log entry time; clear skip time. */
-    pcn->cur_entry_time = current_hrtime;
-    pcn->cur_skip_time  = 0;
-
-    /* The current call graph node becomes this one. */
-    ptd->current_call = pcn;
 }
 
 /* Log that we've entered a native routine */
@@ -128,6 +231,9 @@ static void log_exit(MVMThreadContext *tc, MVMuint32 unwind) {
     MVMProfileCallNode *pcn = ptd->current_call;
     if (!pcn) {
         if (tc->instance->profiling) {
+            if (ptd->non_calltree_depth-- > 0) {
+                return;
+            }
             /* No frame but still profiling; corruption. */
             MVM_dump_backtrace(tc);
             MVM_panic(1, "Profiler lost sequence");
@@ -225,6 +331,7 @@ void log_one_allocation(MVMThreadContext *tc, MVMObject *obj, MVMProfileCallNode
     MVMObject *what = STABLE(obj)->WHAT;
     MVMuint32 i;
     MVMuint8 allocation_target;
+    MVM_ASSERT_NOT_FROMSPACE(tc, what);
     if (replaced) {
         allocation_target = 3;
     } else if (pcn->entry_mode == MVM_PROFILE_ENTER_SPESH || pcn->entry_mode == MVM_PROFILE_ENTER_SPESH_INLINE) {
@@ -287,6 +394,51 @@ void MVM_profile_log_allocated(MVMThreadContext *tc, MVMObject *obj) {
         }
     }
 }
+void MVM_profiler_log_gc_deallocate(MVMThreadContext *tc, MVMObject *object) {
+    if (tc->instance->profiling && STABLE(object)) {
+        MVMProfileGC *pgc = &tc->prof_data->gcs[tc->prof_data->num_gcs];
+        MVMObject *what = STABLE(object)->WHAT;
+        MVMCollectable *item = (MVMCollectable *)object;
+        MVMuint32 i;
+
+        MVMuint8 dealloc_target = 0;
+
+        if (what->header.flags & MVM_CF_FORWARDER_VALID)
+            what = (MVMObject *)what->header.sc_forward_u.forwarder;
+
+        MVM_ASSERT_NOT_FROMSPACE(tc, what);
+
+        if (item->flags & MVM_CF_SECOND_GEN)
+            dealloc_target = 2;
+        else if (item->flags & MVM_CF_NURSERY_SEEN)
+            dealloc_target = 1;
+
+        /* See if there's an existing node to update. */
+        for (i = 0; i < pgc->num_dealloc; i++) {
+            if (pgc->deallocs[i].type == what) {
+                if (dealloc_target == 2)
+                    pgc->deallocs[i].deallocs_gen2++;
+                else if (dealloc_target == 1)
+                    pgc->deallocs[i].deallocs_nursery_seen++;
+                else
+                    pgc->deallocs[i].deallocs_nursery_fresh++;
+                return;
+            }
+        }
+
+        /* No entry; create one. */
+        if (pgc->num_dealloc == pgc->alloc_dealloc) {
+            pgc->alloc_dealloc += 8;
+            pgc->deallocs = MVM_realloc(pgc->deallocs,
+                pgc->alloc_dealloc * sizeof(MVMProfileDeallocationCount));
+        }
+        pgc->deallocs[pgc->num_dealloc].type                   = what;
+        pgc->deallocs[pgc->num_dealloc].deallocs_nursery_fresh = dealloc_target == 0;
+        pgc->deallocs[pgc->num_dealloc].deallocs_nursery_seen  = dealloc_target == 1;
+        pgc->deallocs[pgc->num_dealloc].deallocs_gen2          = dealloc_target == 2;
+        pgc->num_dealloc++;
+    }
+}
 
 /* Logs a scalar-replaced allocation. */
 void MVM_profile_log_scalar_replaced(MVMThreadContext *tc, MVMSTable *st) {
@@ -318,6 +470,11 @@ void MVM_profiler_log_gc_start(MVMThreadContext *tc, MVMuint32 full, MVMuint32 t
     gc->responsible   = this_thread_responsible;
     gc->gc_seq_num    = MVM_load(&tc->instance->gc_seq_number);
 
+    gc->num_dealloc = 0;
+    gc->alloc_dealloc = 0;
+    gc->deallocs = NULL;
+    gc->num_stolen_gen2roots = 0;
+
     /* Record start time. */
     ptd->cur_gc_start_time = uv_hrtime();
     /* Also store this time in the GC data */
@@ -364,6 +521,14 @@ void MVM_profiler_log_unmanaged_data_promoted(MVMThreadContext *tc, MVMuint64 am
     ptd->gc_promoted_unmanaged_bytes += amount;
 }
 
+void MVM_profiler_log_gen2_roots(MVMThreadContext *tc, MVMuint64 amount, MVMThreadContext *other) {
+    if (tc != other) {
+        MVMProfileThreadData *ptd = get_thread_data(tc);
+
+        ptd->gcs[ptd->num_gcs].num_stolen_gen2roots += amount;
+    }
+}
+
 /* Log that we're starting some work on bytecode specialization or JIT. */
 void MVM_profiler_log_spesh_start(MVMThreadContext *tc) {
     /* Record start time. */
