@@ -85,9 +85,129 @@ static void pos_box(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
 /* Gets the primitive boxed by a type. */
 static MVMuint16 prim_spec(MVMThreadContext *tc, MVMSpeshStatsType *type_tuple, MVMint32 i) {
     MVMObject *type = type_tuple ? type_tuple[i].type : NULL;
-    return type
-        ? REPR(type)->get_storage_spec(tc, STABLE(type))->boxed_primitive
-        : 0;
+    const MVMStorageSpec *ss;
+    if (!type)
+        return 0;
+    ss = REPR(type)->get_storage_spec(tc, STABLE(type));
+    if (ss->boxed_primitive)
+        return ss->boxed_primitive;
+    return 0;
+}
+
+static MVMuint8 cmp_prim_spec(MVMThreadContext *tc, MVMSpeshStatsType *type_tuple, MVMint32 i, MVMint32 wanted_prim_spec) {
+    MVMObject *type = NULL;
+    MVMint32 concrete = 0;
+    const MVMStorageSpec *ss;
+    if (type_tuple) {
+        if (type_tuple[i].decont_type) {
+            type = type_tuple[i].decont_type;
+            concrete = type_tuple[i].decont_type_concrete;
+        }
+        else {
+            type = type_tuple[i].type;
+            concrete = type_tuple[i].type_concrete;
+        }
+    }
+    if (!type || !concrete)
+        return 0;
+    ss = REPR(type)->get_storage_spec(tc, STABLE(type));
+    if (ss->boxed_primitive) {
+        return ss->boxed_primitive == wanted_prim_spec;
+    }
+    if (ss->can_box) {
+        if (wanted_prim_spec == MVM_STORAGE_SPEC_BP_INT) {
+            return ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_INT;
+        }
+        else if (wanted_prim_spec == MVM_STORAGE_SPEC_BP_NUM) {
+            return ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_NUM;
+        }
+        else if (wanted_prim_spec == MVM_STORAGE_SPEC_BP_STR) {
+            return ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_STR;
+        }
+        else
+            MVM_panic(1, "Spesh error: unexpected wanted_prim_spec %d", wanted_prim_spec);
+    }
+    return 0;
+}
+
+static void insert_getarg_and_box(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
+                                  MVMSpeshIns *insert_point, MVMCallsiteFlags arg_flags,
+                                  MVMint32 argument_idx, MVMSpeshOperand value_temp) {
+    /* Instruction to get value depends on argument type. */
+    if ((arg_flags & MVM_CALLSITE_ARG_MASK) == MVM_CALLSITE_ARG_OBJ) {
+        /* It's already a boxed object, so just fetch it into the value
+         * register. */
+        MVMSpeshIns *fetch_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
+        fetch_ins->info = MVM_op_get_op(MVM_OP_sp_getarg_o);
+        fetch_ins->operands = MVM_spesh_alloc(tc, g, 2 * sizeof(MVMSpeshOperand));
+        fetch_ins->operands[0] = value_temp;
+        fetch_ins->operands[1].lit_ui16 = argument_idx;
+        MVM_spesh_manipulate_insert_ins(tc, bb, insert_point, fetch_ins);
+    }
+    else {
+        MVMSpeshIns *box_ins, *hlltype_ins, *fetch_ins;
+
+        /* We need to box it. Get a temporary register to box into, and one
+         * for the HLL type. */
+        MVMSpeshOperand unboxed_temp, box_type_temp;
+        MVMuint16 box_op;
+        MVMuint16 hlltype_op;
+        MVMuint16 fetch_op;
+        switch (arg_flags & MVM_CALLSITE_ARG_MASK) {
+            case MVM_CALLSITE_ARG_INT:
+                unboxed_temp = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_int64);
+                box_op = MVM_OP_box_i;
+                hlltype_op = MVM_OP_hllboxtype_i;
+                fetch_op = MVM_OP_sp_getarg_i;
+                break;
+            case MVM_CALLSITE_ARG_NUM:
+                unboxed_temp = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_num64);
+                box_op = MVM_OP_box_n;
+                hlltype_op = MVM_OP_hllboxtype_n;
+                fetch_op = MVM_OP_sp_getarg_n;
+                break;
+            case MVM_CALLSITE_ARG_STR:
+                unboxed_temp = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_str);
+                box_op = MVM_OP_box_s;
+                hlltype_op = MVM_OP_hllboxtype_s;
+                fetch_op = MVM_OP_sp_getarg_s;
+                break;
+            default:
+                MVM_panic(1, "Spesh args: unexpected named argument type %d", arg_flags);
+        }
+
+        /* Emit instruction to box value. */
+        box_type_temp = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_obj);
+        box_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
+        box_ins->info = MVM_op_get_op(box_op);
+        box_ins->operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
+        box_ins->operands[0] = value_temp;
+        box_ins->operands[1] = unboxed_temp;
+        box_ins->operands[2] = box_type_temp;
+        MVM_spesh_manipulate_insert_ins(tc, bb, insert_point, box_ins);
+        MVM_spesh_get_facts(tc, g, value_temp)->writer = box_ins;
+
+        /* Prepend the instruction get box type. */
+        hlltype_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
+        hlltype_ins->info = MVM_op_get_op(hlltype_op);
+        hlltype_ins->operands = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshOperand));
+        hlltype_ins->operands[0] = box_type_temp;
+        MVM_spesh_manipulate_insert_ins(tc, bb, insert_point, hlltype_ins);
+        MVM_spesh_get_facts(tc, g, box_type_temp)->writer = hlltype_ins;
+
+        /* Prepend fetch instruction. */
+        fetch_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
+        fetch_ins->info = MVM_op_get_op(fetch_op);
+        fetch_ins->operands = MVM_spesh_alloc(tc, g, 2 * sizeof(MVMSpeshOperand));
+        fetch_ins->operands[0] = unboxed_temp;
+        fetch_ins->operands[1].lit_ui16 = argument_idx;
+        MVM_spesh_manipulate_insert_ins(tc, bb, insert_point, fetch_ins);
+        MVM_spesh_get_facts(tc, g, unboxed_temp)->writer = fetch_ins;
+
+        /* Can release the temporary register now. */
+        MVM_spesh_manipulate_release_temp_reg(tc, g, unboxed_temp);
+        MVM_spesh_manipulate_release_temp_reg(tc, g, box_type_temp);
+    }
 }
 
 /* Puts a single named argument into a slurpy hash, boxing if needed. */
@@ -115,81 +235,7 @@ static void slurp_named_arg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *
     bindkey_ins->operands[2] = value_temp;
     MVM_spesh_manipulate_insert_ins(tc, bb, hash_ins, bindkey_ins);
 
-    /* Instruction to get value depends on argument type. */
-    if ((flags & MVM_CALLSITE_ARG_MASK) == MVM_CALLSITE_ARG_OBJ) {
-        /* It's already a boxed object, so just fetch it into the value
-         * register. */
-        MVMSpeshIns *fetch_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
-        fetch_ins->info = MVM_op_get_op(MVM_OP_sp_getarg_o);
-        fetch_ins->operands = MVM_spesh_alloc(tc, g, 2 * sizeof(MVMSpeshOperand));
-        fetch_ins->operands[0] = value_temp;
-        fetch_ins->operands[1].lit_ui16 = arg_idx;
-        MVM_spesh_manipulate_insert_ins(tc, bb, hash_ins, fetch_ins);
-    }
-    else {
-        MVMSpeshIns *box_ins, *hlltype_ins, *fetch_ins;
-
-        /* We need to box it. Get a temporary register to box into, and one
-         * for the HLL type. */
-        MVMSpeshOperand unboxed_temp, box_type_temp;
-        MVMuint16 box_op;
-        MVMuint16 hlltype_op;
-        MVMuint16 fetch_op;
-        switch (flags & MVM_CALLSITE_ARG_MASK) {
-            case MVM_CALLSITE_ARG_INT:
-                unboxed_temp = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_int64);
-                box_op = MVM_OP_box_i;
-                hlltype_op = MVM_OP_hllboxtype_i;
-                fetch_op = MVM_OP_sp_getarg_i;
-                break;
-            case MVM_CALLSITE_ARG_NUM:
-                unboxed_temp = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_num64);
-                box_op = MVM_OP_box_n;
-                hlltype_op = MVM_OP_hllboxtype_n;
-                fetch_op = MVM_OP_sp_getarg_n;
-                break;
-            case MVM_CALLSITE_ARG_STR:
-                unboxed_temp = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_str);
-                box_op = MVM_OP_box_s;
-                hlltype_op = MVM_OP_hllboxtype_s;
-                fetch_op = MVM_OP_sp_getarg_s;
-                break;
-            default:
-                MVM_panic(1, "Spesh args: unexpected named argument type %d", flags);
-        }
-
-        /* Emit instruction to box value. */
-        box_type_temp = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_obj);
-        box_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
-        box_ins->info = MVM_op_get_op(box_op);
-        box_ins->operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
-        box_ins->operands[0] = value_temp;
-        box_ins->operands[1] = unboxed_temp;
-        box_ins->operands[2] = box_type_temp;
-        MVM_spesh_manipulate_insert_ins(tc, bb, hash_ins, box_ins);
-        MVM_spesh_get_facts(tc, g, value_temp)->writer = box_ins;
-
-        /* Prepend the instruction get box type. */
-        hlltype_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
-        hlltype_ins->info = MVM_op_get_op(hlltype_op);
-        hlltype_ins->operands = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshOperand));
-        hlltype_ins->operands[0] = box_type_temp;
-        MVM_spesh_manipulate_insert_ins(tc, bb, hash_ins, hlltype_ins);
-        MVM_spesh_get_facts(tc, g, box_type_temp)->writer = hlltype_ins;
-
-        /* Prepend fetch instruction. */
-        fetch_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
-        fetch_ins->info = MVM_op_get_op(fetch_op);
-        fetch_ins->operands = MVM_spesh_alloc(tc, g, 2 * sizeof(MVMSpeshOperand));
-        fetch_ins->operands[0] = unboxed_temp;
-        fetch_ins->operands[1].lit_ui16 = arg_idx;
-        MVM_spesh_manipulate_insert_ins(tc, bb, hash_ins, fetch_ins);
-        MVM_spesh_get_facts(tc, g, unboxed_temp)->writer = fetch_ins;
-
-        /* Can release the temporary register now. */
-        MVM_spesh_manipulate_release_temp_reg(tc, g, unboxed_temp);
-        MVM_spesh_manipulate_release_temp_reg(tc, g, box_type_temp);
-    }
+    insert_getarg_and_box(tc, g, bb, hash_ins, flags, arg_idx, value_temp);
 
     /* Insert key fetching instruciton; we just store the string in a spesh
      * slot. */
@@ -205,6 +251,43 @@ static void slurp_named_arg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *
     MVM_spesh_manipulate_release_temp_reg(tc, g, value_temp);
 }
 
+static void slurp_positional_arg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
+                                 MVMSpeshIns *array_ins, MVMint32 pos_idx, MVMint32 target_idx) {
+    MVMSpeshIns *index_ins;
+
+    /* Look up arg flags and name, and compute index. */
+    MVMCallsiteFlags flags = g->cs->arg_flags[pos_idx];
+
+    /* Allocate temporary registers for the key and value. */
+    MVMSpeshOperand index_temp = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_int64);
+    MVMSpeshOperand value_temp = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_obj);
+
+    /* Insert bind key instruction after slurpy hash creation instruction (we
+     * do it first as below we prepend instructions to obtain the key and the
+     * value. */
+    MVMSpeshIns *bindpos_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
+    bindpos_ins->info = MVM_op_get_op(MVM_OP_bindpos_o);
+    bindpos_ins->operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
+    bindpos_ins->operands[0] = array_ins->operands[0];
+    bindpos_ins->operands[1] = index_temp;
+    bindpos_ins->operands[2] = value_temp;
+    MVM_spesh_manipulate_insert_ins(tc, bb, array_ins, bindpos_ins);
+
+    insert_getarg_and_box(tc, g, bb, array_ins, flags, pos_idx, value_temp);
+
+    /* Insert index setting instruciton */
+    index_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
+    index_ins->info = MVM_op_get_op(MVM_OP_const_i64_16);
+    index_ins->operands = MVM_spesh_alloc(tc, g, 2 * sizeof(MVMSpeshOperand));
+    index_ins->operands[0] = index_temp;
+    index_ins->operands[1].lit_i16 = target_idx;
+    MVM_spesh_manipulate_insert_ins(tc, bb, array_ins, index_ins);
+
+    /* Release temporary registers after. */
+    MVM_spesh_manipulate_release_temp_reg(tc, g, index_temp);
+    MVM_spesh_manipulate_release_temp_reg(tc, g, value_temp);
+}
+
 /* Takes information about the incoming callsite and arguments, and performs
  * various optimizations based on that information. */
 void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
@@ -217,6 +300,8 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
     MVMSpeshBB   *paramnamesused_bb  = NULL;
     MVMSpeshIns  *param_sn_ins       = NULL;
     MVMSpeshBB   *param_sn_bb        = NULL;
+    MVMSpeshIns  *param_sp_ins       = NULL;
+    MVMSpeshBB   *param_sp_bb        = NULL;
 
     MVMSpeshIns **pos_ins    = MVM_calloc(MAX_POS_ARGS, sizeof(MVMSpeshIns *));
     MVMSpeshBB  **pos_bb     = MVM_calloc(MAX_POS_ARGS, sizeof(MVMSpeshBB *));
@@ -259,8 +344,11 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                 MVMint16 idx = ins->operands[1].lit_i16;
                 if (idx < 0 || idx >= MAX_POS_ARGS)
                     goto cleanup;
-                if (pos_ins[idx]) /* Dupe; weird. */
+                if (pos_ins[idx]) { /* Dupe; weird. */
+                    MVM_spesh_graph_add_comment(tc, g, ins,
+                            "bailed argument spesh: duplicate access to argument %ld", idx);
                     goto cleanup;
+                }
                 pos_ins[idx] = ins;
                 pos_bb[idx]  = bb;
                 if (idx > req_max)
@@ -276,8 +364,11 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                 MVMint16 idx = ins->operands[1].lit_i16;
                 if (idx < 0 || idx >= MAX_POS_ARGS)
                     goto cleanup;
-                if (pos_ins[idx]) /* Dupe; weird. */
+                if (pos_ins[idx]) { /* Dupe; weird. */
+                    MVM_spesh_graph_add_comment(tc, g, ins,
+                            "bailed argument spesh: duplicate access to argument %ld", idx);
                     goto cleanup;
+                }
                 pos_ins[idx] = ins;
                 pos_bb[idx]  = bb;
                 if (idx > opt_max)
@@ -304,6 +395,8 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                 num_named++;
                 break;
             case MVM_OP_param_sp:
+                param_sp_ins = ins;
+                param_sp_bb = bb;
                 break;
             case MVM_OP_param_sn:
                 param_sn_ins = ins;
@@ -312,6 +405,8 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
             case MVM_OP_usecapture:
             case MVM_OP_savecapture:
                 /* Require full args processing context for now; bail. */
+                MVM_spesh_graph_add_comment(tc, g, ins,
+                        "bailed argument spesh: support for %s NYI", ins->info->name);
                 goto cleanup;
             case MVM_OP_paramnamesused:
                 if (paramnamesused_ins)
@@ -330,6 +425,8 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
             case MVM_OP_param_rn2_u:
             case MVM_OP_param_on2_u:
                 /* Don't understand how to specialize these yet. */
+                MVM_spesh_graph_add_comment(tc, g, ins,
+                        "bailed argument spesh: support for %s NYI", ins->info->name);
                 goto cleanup;
             default:
                 break;
@@ -341,12 +438,18 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
     }
 
     /* If we didn't find a checkarity instruction, bail. */
-    if (!checkarity_ins)
+    if (!checkarity_ins) {
+        MVM_spesh_graph_add_comment(tc, g, g->entry->first_ins,
+                "bailed argument spesh: no checkarity instruction found");
         goto cleanup;
+    }
 
     /* If required and optional aren't contiguous, bail. */
-    if (opt_min >= 0 && req_max + 1 != opt_min)
+    if (opt_min >= 0 && req_max + 1 != opt_min) {
+        MVM_spesh_graph_add_comment(tc, g, g->entry->first_ins,
+                "bailed argument spesh: required and optional args aren't contiguous");
         goto cleanup;
+    }
 
     /* If the number of passed args is in range... */
     if (cs->num_pos >= req_max + 1 && (opt_max < 0 || cs->num_pos <= opt_max + 1)) {
@@ -355,46 +458,65 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
         MVMint32 i;
         for (i = 0; i < cs->num_pos; i++) {
             MVMCallsiteEntry arg_flag = cs->arg_flags[i];
-            if (!pos_ins[i])
-                goto cleanup;
-            switch (pos_ins[i]->info->opcode) {
-            case MVM_OP_param_rp_i:
-            case MVM_OP_param_op_i:
-                if (arg_flag != MVM_CALLSITE_ARG_INT)
-                    if (arg_flag != MVM_CALLSITE_ARG_OBJ ||
-                            prim_spec(tc, type_tuple, i) != MVM_STORAGE_SPEC_BP_INT)
-                        goto cleanup;
-                break;
-            case MVM_OP_param_rp_u:
-            case MVM_OP_param_op_u:
-                if (arg_flag != MVM_CALLSITE_ARG_INT)
-                    if (arg_flag != MVM_CALLSITE_ARG_OBJ ||
-                            prim_spec(tc, type_tuple, i) != MVM_STORAGE_SPEC_BP_INT) {
+            if (pos_ins[i]) {
+                switch (pos_ins[i]->info->opcode) {
+                case MVM_OP_param_rp_i:
+                case MVM_OP_param_op_i:
+                    if (arg_flag != MVM_CALLSITE_ARG_INT)
+                        if (arg_flag != MVM_CALLSITE_ARG_OBJ ||
+                                !cmp_prim_spec(tc, type_tuple, i, MVM_STORAGE_SPEC_BP_INT)) {
+                            MVM_spesh_graph_add_comment(tc, g, pos_ins[i],
+                                    "bailed argument spesh: expected arg flag %ld to be int or box an int; type at position was %s", i, type_tuple && type_tuple[i].type ? MVM_6model_get_debug_name(tc, type_tuple[i].type) : "null type tuple");
+                            goto cleanup;
+                        }
+                    break;
+                case MVM_OP_param_rp_u:
+                case MVM_OP_param_op_u:
+                    if (arg_flag != MVM_CALLSITE_ARG_INT)
+                        if (arg_flag != MVM_CALLSITE_ARG_OBJ ||
+                                !cmp_prim_spec(tc, type_tuple, i, MVM_STORAGE_SPEC_BP_INT)) {
+                            MVM_spesh_graph_add_comment(tc, g, pos_ins[i],
+                                    "bailed argument spesh: expected arg flag %ld to be int or box an int; type at position was %s", i, type_tuple && type_tuple[i].type ? MVM_6model_get_debug_name(tc, type_tuple[i].type) : "null type tuple");
+                            goto cleanup;
+                        }
+                    break;
+                case MVM_OP_param_rp_n:
+                case MVM_OP_param_op_n:
+                    if (arg_flag != MVM_CALLSITE_ARG_NUM)
+                        if (arg_flag != MVM_CALLSITE_ARG_OBJ ||
+                                !cmp_prim_spec(tc, type_tuple, i, MVM_STORAGE_SPEC_BP_NUM)) {
+                            MVM_spesh_graph_add_comment(tc, g, pos_ins[i],
+                                    "bailed argument spesh: expected arg flag %ld to be num or box a num", i);
+                            goto cleanup;
+                        }
+                    break;
+                case MVM_OP_param_rp_s:
+                case MVM_OP_param_op_s:
+                    if (arg_flag != MVM_CALLSITE_ARG_STR)
+                        if (arg_flag != MVM_CALLSITE_ARG_OBJ ||
+                                !cmp_prim_spec(tc, type_tuple, i, MVM_STORAGE_SPEC_BP_STR)) {
+                            MVM_spesh_graph_add_comment(tc, g, pos_ins[i],
+                                    "bailed argument spesh: expected arg flag %ld to be str or box a str type at position was %s", i, type_tuple && type_tuple[i].type ? MVM_6model_get_debug_name(tc, type_tuple[i].type) : "null type tuple");
+                            goto cleanup;
+                        }
+                    break;
+                case MVM_OP_param_rp_o:
+                case MVM_OP_param_op_o:
+                    if (arg_flag != MVM_CALLSITE_ARG_OBJ && arg_flag != MVM_CALLSITE_ARG_INT &&
+                        arg_flag != MVM_CALLSITE_ARG_NUM && arg_flag != MVM_CALLSITE_ARG_STR) {
+                        MVM_spesh_graph_add_comment(tc, g, pos_ins[i],
+                                "bailed argument spesh: expected arg flag %ld to be obj or int or num or str", i);
                         goto cleanup;
                     }
-                break;
-            case MVM_OP_param_rp_n:
-            case MVM_OP_param_op_n:
-                if (arg_flag != MVM_CALLSITE_ARG_NUM)
-                    if (arg_flag != MVM_CALLSITE_ARG_OBJ ||
-                            prim_spec(tc, type_tuple, i) != MVM_STORAGE_SPEC_BP_NUM)
-                        goto cleanup;
-                break;
-            case MVM_OP_param_rp_s:
-            case MVM_OP_param_op_s:
-                if (arg_flag != MVM_CALLSITE_ARG_STR)
-                    if (arg_flag != MVM_CALLSITE_ARG_OBJ ||
-                            prim_spec(tc, type_tuple, i) != MVM_STORAGE_SPEC_BP_STR)
-                        goto cleanup;
-                break;
-            case MVM_OP_param_rp_o:
-            case MVM_OP_param_op_o:
-                if (arg_flag != MVM_CALLSITE_ARG_OBJ && arg_flag != MVM_CALLSITE_ARG_INT &&
-                    arg_flag != MVM_CALLSITE_ARG_NUM && arg_flag != MVM_CALLSITE_ARG_STR)
-                    goto cleanup;
-                break;
-            default:
-                break;
+                    break;
+                default:
+                    break;
+                }
+            }
+            else if (!pos_ins[i] && !param_sp_ins) {
+                MVM_spesh_graph_add_comment(tc, g, g->entry->first_ins,
+                        "bailed argument spesh: no positional arg fetch op found for %ld", i);
+                goto cleanup;
             }
         }
 
@@ -405,6 +527,8 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
          * facts. */
         for (i = 0; i < cs->num_pos; i++) {
             MVMCallsiteEntry arg_flag = cs->arg_flags[i];
+            if (!pos_ins[i])
+                continue;
             switch (pos_ins[i]->info->opcode) {
                 /* FIXME add support for param_rp_u here */
             case MVM_OP_param_rp_i:
@@ -413,7 +537,7 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                     pos_ins[i]->info = MVM_op_get_op(MVM_OP_sp_getarg_i);
                 }
                 else {
-                    pos_unbox(tc, g, pos_bb[i], pos_ins[i], MVM_op_get_op(MVM_OP_unbox_i));
+                    pos_unbox(tc, g, pos_bb[i], pos_ins[i], MVM_op_get_op(MVM_OP_decont_i));
                     pos_added[i]++;
                     if (type_tuple && type_tuple[i].type)
                         add_facts(tc, g, i, type_tuple[i], pos_ins[i]);
@@ -425,7 +549,7 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                     pos_ins[i]->info = MVM_op_get_op(MVM_OP_sp_getarg_i);
                 }
                 else {
-                    pos_unbox(tc, g, pos_bb[i], pos_ins[i], MVM_op_get_op(MVM_OP_unbox_i));
+                    pos_unbox(tc, g, pos_bb[i], pos_ins[i], MVM_op_get_op(MVM_OP_decont_i));
                     pos_added[i]++;
                     if (type_tuple && type_tuple[i].type)
                         add_facts(tc, g, i, type_tuple[i], pos_ins[i]);
@@ -437,7 +561,7 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                     pos_ins[i]->info = MVM_op_get_op(MVM_OP_sp_getarg_n);
                 }
                 else {
-                    pos_unbox(tc, g, pos_bb[i], pos_ins[i], MVM_op_get_op(MVM_OP_unbox_n));
+                    pos_unbox(tc, g, pos_bb[i], pos_ins[i], MVM_op_get_op(MVM_OP_decont_n));
                     pos_added[i]++;
                     if (type_tuple && type_tuple[i].type)
                         add_facts(tc, g, i, type_tuple[i], pos_ins[i]);
@@ -449,7 +573,7 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                     pos_ins[i]->info = MVM_op_get_op(MVM_OP_sp_getarg_s);
                 }
                 else {
-                    pos_unbox(tc, g, pos_bb[i], pos_ins[i], MVM_op_get_op(MVM_OP_unbox_s));
+                    pos_unbox(tc, g, pos_bb[i], pos_ins[i], MVM_op_get_op(MVM_OP_decont_s));
                     pos_added[i]++;
                     if (type_tuple && type_tuple[i].type)
                         add_facts(tc, g, i, type_tuple[i], pos_ins[i]);
@@ -547,8 +671,10 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
             /* Now go by instruction. */
             switch (named_ins[i]->info->opcode) {
             case MVM_OP_param_rn_i:
-                if (found_idx == -1)
+                if (found_idx == -1) {
+                    MVM_spesh_graph_add_comment(tc, g, named_ins[i], "bailed argument spesh: required named argument not found!");
                     goto cleanup;
+                }
                 if (found_flag & MVM_CALLSITE_ARG_INT) {
                     named_ins[i]->info = MVM_op_get_op(MVM_OP_sp_getarg_i);
                     named_ins[i]->operands[1].lit_i16 = found_idx + 1;
@@ -557,14 +683,16 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                 else if (found_flag & MVM_CALLSITE_ARG_OBJ
                         && prim_spec(tc, type_tuple, found_flag_idx) == MVM_STORAGE_SPEC_BP_INT) {
                     named_ins[i]->operands[1].lit_i16 = found_idx + 1;
-                    pos_unbox(tc, g, named_bb[i], named_ins[i], MVM_op_get_op(MVM_OP_unbox_i));
+                    pos_unbox(tc, g, named_bb[i], named_ins[i], MVM_op_get_op(MVM_OP_decont_i));
                     named_used_bit_field |= (MVMuint64)1 << cur_named;
                 }
                 named_used++;
                 break;
             case MVM_OP_param_rn_n:
-                if (found_idx == -1)
+                if (found_idx == -1) {
+                    MVM_spesh_graph_add_comment(tc, g, named_ins[i], "bailed argument spesh: required named argument not found!");
                     goto cleanup;
+                }
                 if (found_flag & MVM_CALLSITE_ARG_NUM) {
                     named_ins[i]->info = MVM_op_get_op(MVM_OP_sp_getarg_n);
                     named_ins[i]->operands[1].lit_i16 = found_idx + 1;
@@ -573,14 +701,16 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                 else if (found_flag & MVM_CALLSITE_ARG_OBJ
                         && prim_spec(tc, type_tuple, found_flag_idx) == MVM_STORAGE_SPEC_BP_NUM) {
                     named_ins[i]->operands[1].lit_i16 = found_idx + 1;
-                    pos_unbox(tc, g, named_bb[i], named_ins[i], MVM_op_get_op(MVM_OP_unbox_n));
+                    pos_unbox(tc, g, named_bb[i], named_ins[i], MVM_op_get_op(MVM_OP_decont_n));
                     named_used_bit_field |= (MVMuint64)1 << cur_named;
                 }
                 named_used++;
                 break;
             case MVM_OP_param_rn_s:
-                if (found_idx == -1)
+                if (found_idx == -1) {
+                    MVM_spesh_graph_add_comment(tc, g, named_ins[i], "bailed argument spesh: required named argument not found!");
                     goto cleanup;
+                }
                 if (found_flag & MVM_CALLSITE_ARG_STR) {
                     named_ins[i]->info = MVM_op_get_op(MVM_OP_sp_getarg_s);
                     named_ins[i]->operands[1].lit_i16 = found_idx + 1;
@@ -589,14 +719,16 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                 else if (found_flag & MVM_CALLSITE_ARG_OBJ
                         && prim_spec(tc, type_tuple, found_flag_idx) == MVM_STORAGE_SPEC_BP_STR) {
                     named_ins[i]->operands[1].lit_i16 = found_idx + 1;
-                    pos_unbox(tc, g, named_bb[i], named_ins[i], MVM_op_get_op(MVM_OP_unbox_s));
+                    pos_unbox(tc, g, named_bb[i], named_ins[i], MVM_op_get_op(MVM_OP_decont_s));
                     named_used_bit_field |= (MVMuint64)1 << cur_named;
                 }
                 named_used++;
                 break;
             case MVM_OP_param_rn_o:
-                if (found_idx == -1)
+                if (found_idx == -1) {
+                    MVM_spesh_graph_add_comment(tc, g, named_ins[i], "bailed argument spesh: required named argument not found!");
                     goto cleanup;
+                }
                 if (found_flag & MVM_CALLSITE_ARG_OBJ) {
                     MVMuint16 arg_idx = found_idx + 1;
                     named_ins[i]->info = MVM_op_get_op(MVM_OP_sp_getarg_o);
@@ -642,7 +774,7 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                 else if (found_flag & MVM_CALLSITE_ARG_OBJ
                         && prim_spec(tc, type_tuple, found_flag_idx) == MVM_STORAGE_SPEC_BP_INT) {
                     named_ins[i]->operands[1].lit_i16 = found_idx + 1;
-                    pos_unbox(tc, g, named_bb[i], named_ins[i], MVM_op_get_op(MVM_OP_unbox_i));
+                    pos_unbox(tc, g, named_bb[i], named_ins[i], MVM_op_get_op(MVM_OP_decont_i));
                     MVM_spesh_manipulate_insert_goto(tc, g, named_bb[i], named_ins[i]->next,
                         named_ins[i]->operands[2].ins_bb);
                     MVM_spesh_manipulate_remove_successor(tc, named_bb[i],
@@ -669,7 +801,7 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                 else if (found_flag & MVM_CALLSITE_ARG_OBJ
                         && prim_spec(tc, type_tuple, found_flag_idx) == MVM_STORAGE_SPEC_BP_NUM) {
                     named_ins[i]->operands[1].lit_i16 = found_idx + 1;
-                    pos_unbox(tc, g, named_bb[i], named_ins[i], MVM_op_get_op(MVM_OP_unbox_n));
+                    pos_unbox(tc, g, named_bb[i], named_ins[i], MVM_op_get_op(MVM_OP_decont_n));
                     MVM_spesh_manipulate_insert_goto(tc, g, named_bb[i], named_ins[i]->next,
                         named_ins[i]->operands[2].ins_bb);
                     MVM_spesh_manipulate_remove_successor(tc, named_bb[i],
@@ -696,7 +828,7 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                 else if (found_flag & MVM_CALLSITE_ARG_OBJ
                         && prim_spec(tc, type_tuple, found_flag_idx) == MVM_STORAGE_SPEC_BP_STR) {
                     named_ins[i]->operands[1].lit_i16 = found_idx + 1;
-                    pos_unbox(tc, g, named_bb[i], named_ins[i], MVM_op_get_op(MVM_OP_unbox_s));
+                    pos_unbox(tc, g, named_bb[i], named_ins[i], MVM_op_get_op(MVM_OP_decont_s));
                     MVM_spesh_manipulate_insert_goto(tc, g, named_bb[i], named_ins[i]->next,
                         named_ins[i]->operands[2].ins_bb);
                     MVM_spesh_manipulate_remove_successor(tc, named_bb[i],
@@ -779,7 +911,7 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
         if (param_sn_ins) {
             /* Construct it as a hash. */
             MVMObject *hash_type = g->sf->body.cu->body.hll_config->slurpy_hash_type;
-            if (REPR(hash_type)->ID == MVM_REPR_ID_MVMHash) {
+            if (REPR(hash_type)->ID == MVM_REPR_ID_MVMHash && !REPR(hash_type)->initialize) {
                 MVMSpeshOperand target    = param_sn_ins->operands[0];
                 param_sn_ins->info        = MVM_op_get_op(MVM_OP_sp_fastcreate);
                 param_sn_ins->operands    = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
@@ -787,6 +919,7 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                 param_sn_ins->operands[1].lit_i16 = sizeof(MVMHash);
                 param_sn_ins->operands[2].lit_i16 = MVM_spesh_add_spesh_slot(tc, g,
                     (MVMCollectable *)STABLE(hash_type));
+                MVM_spesh_graph_add_comment(tc, g, param_sn_ins, "argument spesh: slurpy nameds optimized");
             }
             else {
                 MVM_oops(tc, "Arg spesh: slurpy hash type was not a VMHash as expected");
@@ -800,6 +933,30 @@ void MVM_spesh_args(MVMThreadContext *tc, MVMSpeshGraph *g, MVMCallsite *cs,
                         slurp_named_arg(tc, g, param_sn_bb, param_sn_ins, i);
         }
 
+        /* If we have a slurpy array ... */
+        if (param_sp_ins) {
+            MVMint32 start_slurping_at = param_sp_ins->operands[1].lit_i16;
+            /* Construct it as a hash. */
+            MVMObject *array_type = g->sf->body.cu->body.hll_config->slurpy_array_type;
+            if (REPR(array_type)->ID == MVM_REPR_ID_VMArray && !REPR(array_type)->initialize) {
+                MVMSpeshOperand target    = param_sp_ins->operands[0];
+                param_sp_ins->info        = MVM_op_get_op(MVM_OP_sp_fastcreate);
+                param_sp_ins->operands    = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
+                param_sp_ins->operands[0] = target;
+                param_sp_ins->operands[1].lit_i16 = sizeof(MVMArray);
+                param_sp_ins->operands[2].lit_i16 = MVM_spesh_add_spesh_slot(tc, g,
+                    (MVMCollectable *)STABLE(array_type));
+                MVM_spesh_graph_add_comment(tc, g, param_sp_ins, "argument spesh: slurpy positional optimized");
+            }
+            else {
+                MVM_oops(tc, "Arg spesh: slurpy array type was not a VMArray as expected");
+            }
+
+            /* Populate it with all args after the numeral argument */
+            for (i = cs->num_pos - 1; i >= start_slurping_at; i--)
+                slurp_positional_arg(tc, g, param_sp_bb, param_sp_ins, i, i - start_slurping_at);
+        }
+
         /* Stash the named used bit field in the graph; will need to make it
          * into the candidate and all the way to deopt. */
         g->deopt_named_used_bit_field = named_used_bit_field;
