@@ -92,8 +92,8 @@ static void decont_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *in
      * info. */
     MVMint32 in_flags = in_facts->flags;
     if ((in_flags & MVM_SPESH_FACT_TYPEOBJ) ||
-            (in_flags & MVM_SPESH_FACT_KNOWN_TYPE) &&
-            !in_facts->type->st->container_spec) {
+            ((in_flags & MVM_SPESH_FACT_KNOWN_TYPE) &&
+            !in_facts->type->st->container_spec)) {
         copy_facts(tc, g, out_orig, out_i, in_orig, in_i);
         return;
     }
@@ -127,8 +127,10 @@ static void wval_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 tgt_ori
 static void wvalfrom_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 tgt_orig,
                            MVMuint16 tgt_i, MVMuint16 sslot, MVMint64 idx) {
     MVMSerializationContext *sc = (MVMSerializationContext *)g->spesh_slots[sslot];
-    MVMObject *target = MVM_sc_get_object(tc, sc, idx);
-    object_facts(tc, g, tgt_orig, tgt_i, MVM_sc_try_get_object(tc, sc, idx));
+    if (MVM_sc_is_object_immediately_available(tc, sc, idx)) {
+        MVMObject *target = MVM_sc_get_object(tc, sc, idx);
+        object_facts(tc, g, tgt_orig, tgt_i, MVM_sc_try_get_object(tc, sc, idx));
+    }
 }
 
 /* Let's figure out what exact type of iter we'll get from an iter op */
@@ -376,7 +378,7 @@ static void log_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
         guard->operands[1] = preguard_reg;
         guard->operands[2].lit_i16 = MVM_spesh_add_spesh_slot_try_reuse(tc, g,
             (MVMCollectable *)agg_type->st);
-        guard->operands[3].lit_ui32 = g->deopt_addrs[2 * deopt_one_ann->data.deopt_idx];
+        guard->operands[3].lit_ui32 = deopt_one_ann->data.deopt_idx;
         if (ins->next)
             MVM_spesh_manipulate_insert_ins(tc, bb, ins, guard);
         else
@@ -401,9 +403,14 @@ static void log_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
         deopt_one_ann->next = NULL;
         guard->annotations = deopt_one_ann;
 
-        /* Move deopt usages to the preguard register. */
-        pre_facts->usage.deopt_users = facts->usage.deopt_users;
-        facts->usage.deopt_users = NULL;
+        /* Copy deopt usages to the preguard register. */
+        {
+            MVMSpeshDeoptUseEntry *due = facts->usage.deopt_users;
+            while (due) {
+                MVM_spesh_usages_add_deopt_usage(tc, g, pre_facts, due->deopt_idx);
+                due = due->next;
+            }
+        }
 
         /* Add entry in log guards table, and mark facts as depending on it. */
         if (g->num_log_guards % 16 == 0) {
@@ -510,6 +517,11 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 ins->operands[2].reg.orig, ins->operands[2].reg.i);
             break;
+        case MVM_OP_nativecallcast:
+            create_facts(tc, g,
+                ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                ins->operands[2].reg.orig, ins->operands[2].reg.i);
+            break;
         case MVM_OP_trunc_u16:
         case MVM_OP_trunc_i16:
             trunc_i16_facts(tc, g, ins);
