@@ -47,7 +47,7 @@ void MVM_spesh_graph_grow_deopt_table(MVMThreadContext *tc, MVMSpeshGraph *g) {
 }
 
 /* Records a de-optimization annotation and mapping pair. */
-void MVM_spesh_graph_add_deopt_annotation(MVMThreadContext *tc, MVMSpeshGraph *g,
+MVMint32 MVM_spesh_graph_add_deopt_annotation(MVMThreadContext *tc, MVMSpeshGraph *g,
                                           MVMSpeshIns *ins_node, MVMuint32 deopt_target,
                                           MVMint32 type) {
     /* Add an annotations. */
@@ -61,6 +61,7 @@ void MVM_spesh_graph_add_deopt_annotation(MVMThreadContext *tc, MVMSpeshGraph *g
     MVM_spesh_graph_grow_deopt_table(tc, g);
     g->deopt_addrs[2 * g->num_deopt_addrs] = deopt_target;
     g->num_deopt_addrs++;
+    return ann->data.deopt_idx;
 }
 
 MVM_FORMAT(printf, 4, 5)
@@ -134,7 +135,8 @@ static MVMint32 already_succs(MVMThreadContext *tc, MVMSpeshBB *bb, MVMSpeshBB *
 #define MVM_CFG_BB_END      2
 #define MVM_CFG_BB_JUMPLIST 4
 static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf,
-                      MVMint32 *existing_deopts, MVMint32 num_existing_deopts) {
+                      MVMint32 *existing_deopts, MVMint32 num_existing_deopts,
+                      MVMint32 *deopt_usage_info, MVMSpeshIns ***deopt_usage_ins_out) {
     MVMSpeshBB  *cur_bb, *prev_bb;
     MVMSpeshIns *last_ins;
     MVMint64     i;
@@ -714,6 +716,24 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
         }
     }
 
+    /* We may also need to reconstruct deopt usage info, to allow optimizing of
+     * inlinees. That is persisted using bytecode offsets, and this is the last
+     * place we can map those into instructions, so do it here if needed. */
+    if (deopt_usage_ins_out && deopt_usage_info) {
+        MVM_VECTOR_DECL(MVMSpeshIns *, usage_ins);
+        MVMuint32 idx = 0;
+        MVM_VECTOR_INIT(usage_ins, 32);
+        while (1) {
+            MVMint32 offset = deopt_usage_info[idx];
+            if (offset == -1)
+                break;
+            MVM_VECTOR_PUSH(usage_ins, ins_flat[byte_to_ins_flags[offset] >> 3]);
+            idx++;
+            idx += deopt_usage_info[idx] + 1; /* Skip over deopt indices */
+        }
+        *deopt_usage_ins_out = usage_ins;
+    }
+
     /* Clear up the temporary arrays. */
     MVM_free(byte_to_ins_flags);
     MVM_free(ins_flat);
@@ -790,7 +810,7 @@ static void dfs(MVMSpeshBB **rpo, MVMint32 *insert_pos, MVMuint8 *seen, MVMSpesh
     bb->rpo_idx = *insert_pos;
     (*insert_pos)--;
 }
-static MVMSpeshBB ** reverse_postorder(MVMThreadContext *tc, MVMSpeshGraph *g) {
+MVMSpeshBB ** MVM_spesh_graph_reverse_postorder(MVMThreadContext *tc, MVMSpeshGraph *g) {
     MVMSpeshBB **rpo  = MVM_calloc(g->num_bbs, sizeof(MVMSpeshBB *));
     MVMuint8    *seen = MVM_calloc(g->num_bbs, 1);
     MVMint32     ins  = g->num_bbs - 1;
@@ -1223,7 +1243,7 @@ static void ssa(MVMThreadContext *tc, MVMSpeshGraph *g) {
     MVMint32 i, num_locals;
 
     /* Compute dominance frontiers. */
-    MVMSpeshBB **rpo  = reverse_postorder(tc, g);
+    MVMSpeshBB **rpo  = MVM_spesh_graph_reverse_postorder(tc, g);
     MVMint32    *doms = compute_dominators(tc, g, rpo);
     add_children(tc, g, rpo, doms);
     add_dominance_frontiers(tc, g, rpo, doms);
@@ -1274,7 +1294,7 @@ MVMSpeshGraph * MVM_spesh_graph_create(MVMThreadContext *tc, MVMStaticFrame *sf,
     }
 
     /* Build the CFG out of the static frame, and transform it to SSA. */
-    build_cfg(tc, g, sf, NULL, 0);
+    build_cfg(tc, g, sf, NULL, 0, NULL, NULL);
     if (insert_object_nulls)
         insert_object_null_instructions(tc, g);
     if (!cfg_only) {
@@ -1289,7 +1309,8 @@ MVMSpeshGraph * MVM_spesh_graph_create(MVMThreadContext *tc, MVMStaticFrame *sf,
 
 /* Takes a static frame and creates a spesh graph for it. */
 MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStaticFrame *sf,
-                                                 MVMSpeshCandidate *cand, MVMuint32 cfg_only) {
+                                                 MVMSpeshCandidate *cand, MVMuint32 cfg_only,
+                                                 MVMSpeshIns ***deopt_usage_ins_out) {
     /* Create top-level graph object. */
     MVMSpeshGraph *g     = MVM_calloc(1, sizeof(MVMSpeshGraph));
     g->sf                = sf;
@@ -1305,6 +1326,7 @@ MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStatic
     g->num_deopt_addrs   = cand->num_deopts;
     g->alloc_deopt_addrs = cand->num_deopts;
     g->deopt_named_used_bit_field = cand->deopt_named_used_bit_field;
+    g->deopt_pea         = cand->deopt_pea;
     g->local_types       = cand->local_types;
     g->lexical_types     = cand->lexical_types;
     g->num_spesh_slots   = cand->num_spesh_slots;
@@ -1323,7 +1345,8 @@ MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStatic
     }
 
     /* Build the CFG out of the static frame, and transform it to SSA. */
-    build_cfg(tc, g, sf, cand->deopts, cand->num_deopts);
+    build_cfg(tc, g, sf, cand->deopts, cand->num_deopts, cand->deopt_usage_info,
+            deopt_usage_ins_out);
     if (!cfg_only) {
         MVM_spesh_eliminate_dead_bbs(tc, g, 0);
         add_predecessors(tc, g);
@@ -1353,7 +1376,7 @@ void MVM_spesh_graph_recompute_dominance(MVMThreadContext *tc, MVMSpeshGraph *g)
 
     /* Now form the new dominance tree. */
     add_predecessors(tc, g);
-    rpo = reverse_postorder(tc, g);
+    rpo = MVM_spesh_graph_reverse_postorder(tc, g);
     doms = compute_dominators(tc, g, rpo);
     add_children(tc, g, rpo, doms);
     MVM_free(rpo);
