@@ -53,6 +53,10 @@ static int is_static_frame_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inlin
         *no_inline_reason = "inlining is disabled";
         return 0;
     }
+    if (tc->instance->debugserver) {
+        *no_inline_reason = "inlining not supported when debugging";
+        return 0;
+    }
 
     /* Check frame is not marked as not being allowed to be inlined. */
     if (target_sf->body.no_inline) {
@@ -79,6 +83,16 @@ static int is_static_frame_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inlin
         return 0;
     }
 
+    /* Ensure that we haven't hit size limits. */
+    if (inliner->num_locals > MVM_SPESH_INLINE_MAX_LOCALS) {
+        *no_inline_reason = "inliner has too many locals";
+        return 0;
+    }
+    if (inliner->num_inlines > MVM_SPESH_INLINE_MAX_INLINES) {
+        *no_inline_reason = "inliner has too many inlines";
+        return 0;
+    }
+
     return 1;
 }
 
@@ -87,10 +101,12 @@ static int is_static_frame_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inlin
  * and returns zero. Optionally builds up usage counts on the graph. */
 static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                                MVMStaticFrame *target_sf, MVMSpeshIns *invoke_ins,
-                               MVMSpeshGraph *ig, char **no_inline_reason) {
+                               MVMSpeshGraph *ig, char **no_inline_reason, MVMOpInfo const **no_inline_info) {
     MVMSpeshBB *bb = ig->entry;
     MVMint32 same_hll = target_sf->body.cu->body.hll_config ==
             inliner->sf->body.cu->body.hll_config;
+    if (no_inline_info)
+        *no_inline_info = NULL;
     while (bb) {
         MVMSpeshIns *ins = bb->first_ins;
         while (ins) {
@@ -101,12 +117,16 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
              * which case we're done. */
             if (!is_phi && ins->info->no_inline) {
                 *no_inline_reason = "target has a :noinline instruction";
+                if (no_inline_info)
+                    *no_inline_info = ins->info;
                 return 0;
             }
 
             /* XXX TODO: See if the workaround below is still required */
             if (opcode == MVM_OP_throwpayloadlexcaller && tc->instance->profiling) {
                 *no_inline_reason = "target has throwpayloadlexcaller, which currently causes problems when profiling is on";
+                if (no_inline_info)
+                    *no_inline_info = ins->info;
                 return 0;
             }
 
@@ -114,6 +134,8 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
              * cannot inline. */
             if (!same_hll && ins->info->uses_hll) {
                 *no_inline_reason = "target has a :useshll instruction and HLLs are different";
+                if (no_inline_info)
+                    *no_inline_info = ins->info;
                 return 0;
             }
 
@@ -133,6 +155,8 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
             if (opcode == MVM_OP_bindlex) {
                 if (ins->operands[0].lex.outers > 0) {
                     *no_inline_reason = "target has bind to outer lexical";
+                    if (no_inline_info)
+                        *no_inline_info = ins->info;
                     return 0;
                 }
             }
@@ -166,7 +190,7 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
 
 /* Gets the effective size for inlining considerations of a specialization,
  * which is its code size minus the code size of its inlines. */
-MVMint32 get_effective_size(MVMThreadContext *tc, MVMSpeshCandidate *cand) {
+static MVMint32 get_effective_size(MVMThreadContext *tc, MVMSpeshCandidate *cand) {
     MVMint32 result = cand->bytecode_size;
     MVMuint32 i;
     for (i = 0; i < cand->num_inlines; i++)
@@ -176,6 +200,33 @@ MVMint32 get_effective_size(MVMThreadContext *tc, MVMSpeshCandidate *cand) {
     return (MVMuint32)result;
 }
 
+/* Add deopt usage info to the inlinee. */
+static void add_deopt_usages(MVMThreadContext *tc, MVMSpeshGraph *g, MVMint32 *deopt_usage_info,
+                      MVMSpeshIns **deopt_usage_ins) {
+    MVMuint32 usage_idx = 0;
+    MVMuint32 ins_idx = 0;
+    while (deopt_usage_info[usage_idx] != -1) {
+        MVMSpeshIns *ins = deopt_usage_ins[ins_idx++];
+        MVMint32 count = deopt_usage_info[usage_idx + 1];
+        MVMint32 i;
+        usage_idx += 2;
+        for (i = 0; i < count; i++) {
+            MVMint32 deopt_idx = deopt_usage_info[usage_idx++];
+            MVMSpeshFacts *facts = MVM_spesh_get_facts(tc, g, ins->operands[0]);
+            MVMSpeshDeoptUseEntry *entry = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshDeoptUseEntry));
+            entry->deopt_idx = deopt_idx;
+            entry->next = facts->usage.deopt_users;
+            facts->usage.deopt_users = entry;
+        }
+    }
+}
+
+/* Get the maximum inline size applicable to the specified static frame (it can
+ * be configured by language). */
+int MVM_spesh_inline_get_max_size(MVMThreadContext *tc, MVMStaticFrame *sf) {
+    return sf->body.cu->body.hll_config->max_inline_size;
+}
+
 /* Sees if it will be possible to inline the target code ref, given we could
  * already identify a spesh candidate. Returns NULL if no inlining is possible
  * or a graph ready to be merged if it will be possible. */
@@ -184,12 +235,14 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
                                                MVMSpeshCandidate *cand,
                                                MVMSpeshIns *invoke_ins,
                                                char **no_inline_reason,
-                                               MVMuint32 *effective_size) {
+                                               MVMuint32 *effective_size,
+                                               MVMOpInfo const **no_inline_info) {
     MVMSpeshGraph *ig;
+    MVMSpeshIns **deopt_usage_ins = NULL;
 
     /* Check bytecode size is within the inline limit. */
     *effective_size = get_effective_size(tc, cand);
-    if (*effective_size > MVM_SPESH_MAX_INLINE_SIZE) {
+    if (*effective_size > MVM_spesh_inline_get_max_size(tc, target_sf)) {
         *no_inline_reason = "bytecode is too large to inline";
         return NULL;
     }
@@ -200,13 +253,14 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
 
     /* Build graph from the already-specialized bytecode and check if we can
      * inline the graph. */
-    ig = MVM_spesh_graph_create_from_cand(tc, target_sf, cand, 0);
-    if (is_graph_inlineable(tc, inliner, target_sf, invoke_ins, ig, no_inline_reason)) {
+    ig = MVM_spesh_graph_create_from_cand(tc, target_sf, cand, 0, &deopt_usage_ins);
+    if (is_graph_inlineable(tc, inliner, target_sf, invoke_ins, ig, no_inline_reason, no_inline_info)) {
         /* We can inline it. Do facts discovery, which also sets usage counts.
          * We also need to bump counts for any inline's code_ref_reg to make
          * sure it stays available for deopt. */
         MVMuint32 i;
         MVM_spesh_facts_discover(tc, ig, NULL, 1);
+        add_deopt_usages(tc, ig, cand->deopt_usage_info, deopt_usage_ins);
         for (i = 0; i < ig->num_inlines; i++) {
             /* We can't be very precise about this, because we don't know the
              * SSA version in effect. So bump usages of all version of the
@@ -216,10 +270,12 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
             for (j = 0; j < ig->fact_counts[reg]; j++)
                 MVM_spesh_usages_add_unconditional_deopt_usage(tc, ig, &(ig->facts[reg][j]));
         }
+        MVM_free(deopt_usage_ins);
         return ig;
     }
     else {
         /* Not possible to inline. Clear it up. */
+        MVM_free(deopt_usage_ins);
         MVM_spesh_graph_destroy(tc, ig);
         return NULL;
     }
@@ -228,7 +284,7 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
 /* Tries to get a spesh graph for a particular unspecialized candidate. */
 MVMSpeshGraph * MVM_spesh_inline_try_get_graph_from_unspecialized(MVMThreadContext *tc,
         MVMSpeshGraph *inliner, MVMStaticFrame *target_sf, MVMSpeshIns *invoke_ins,
-        MVMSpeshCallInfo *call_info, MVMSpeshStatsType *type_tuple, char **no_inline_reason) {
+        MVMSpeshCallInfo *call_info, MVMSpeshStatsType *type_tuple, char **no_inline_reason, MVMOpInfo const **no_inline_info) {
     MVMSpeshGraph *ig;
 
     /* Cannot inline with flattening args. */
@@ -251,7 +307,7 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph_from_unspecialized(MVMThreadConte
     MVM_spesh_optimize(tc, ig, NULL);
 
     /* See if it's inlineable; clean up if not. */
-    if (is_graph_inlineable(tc, inliner, target_sf, invoke_ins, ig, no_inline_reason)) {
+    if (is_graph_inlineable(tc, inliner, target_sf, invoke_ins, ig, no_inline_reason, no_inline_info)) {
         return ig;
     }
     else {
@@ -395,6 +451,22 @@ static void rewrite_outer_lookup(MVMThreadContext *tc, MVMSpeshGraph *g,
     MVM_spesh_usages_add_by_reg(tc, g, code_ref_reg, ins);
 }
 
+/* Rewrites a lexical bind to an outer to be done via. a register holding
+ * the outer coderef. */
+static void rewrite_outer_bind(MVMThreadContext *tc, MVMSpeshGraph *g,
+                               MVMSpeshIns *ins, MVMuint16 num_locals,
+                               MVMuint16 op, MVMSpeshOperand code_ref_reg) {
+    MVMSpeshOperand *new_operands = MVM_spesh_alloc(tc, g, 4 * sizeof(MVMSpeshOperand));
+    new_operands[0].lit_ui16 = ins->operands[0].lex.idx;
+    new_operands[1].lit_ui16 = ins->operands[0].lex.outers;
+    new_operands[2] = code_ref_reg;
+    new_operands[3] = ins->operands[1];
+    new_operands[3].reg.orig += num_locals;
+    ins->info = MVM_op_get_op(op);
+    ins->operands = new_operands;
+    MVM_spesh_usages_add_by_reg(tc, g, code_ref_reg, ins);
+}
+
 static void rewrite_hlltype(MVMThreadContext *tc, MVMSpeshGraph *inlinee, MVMSpeshIns *ins) {
     MVMObject *selected_type;
     MVMHLLConfig *hll = inlinee->sf->body.cu->body.hll_config;
@@ -417,29 +489,79 @@ static void rewrite_hlltype(MVMThreadContext *tc, MVMSpeshGraph *inlinee, MVMSpe
     ins->info = MVM_op_get_op(MVM_OP_sp_getspeshslot);
 }
 
+static void tweak_guard_deopt_idx(MVMSpeshIns *ins, MVMSpeshAnn *ann) {
+    /* Twiddle guard opcode to point to the correct deopt index */
+    switch (ins->info->opcode) {
+    case MVM_OP_sp_guard:
+    case MVM_OP_sp_guardconc:
+    case MVM_OP_sp_guardtype:
+    case MVM_OP_sp_guardobj:
+    case MVM_OP_sp_guardnotobj:
+    case MVM_OP_sp_rebless:
+        ins->operands[3].lit_ui32 = ann->data.deopt_idx;
+        break;
+    case MVM_OP_sp_guardsf:
+    case MVM_OP_sp_guardsfouter:
+    case MVM_OP_sp_guardjustconc:
+    case MVM_OP_sp_guardjusttype:
+        ins->operands[2].lit_ui32 = ann->data.deopt_idx;
+        break;
+    default:
+        break;
+    }
+}
+
 /* Merges the inlinee's spesh graph into the inliner. */
-MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
+static MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                  MVMSpeshGraph *inlinee, MVMStaticFrame *inlinee_sf,
                  MVMSpeshBB *invoke_bb, MVMSpeshIns *invoke_ins,
                  MVMSpeshOperand code_ref_reg,
-                 MVMuint32 *inline_boundary_handler, MVMuint16 bytecode_size) {
+                 MVMuint32 *inline_boundary_handler, MVMuint16 bytecode_size,
+                 MVMCallsite *cs) {
     MVMSpeshFacts **merged_facts;
     MVMuint16      *merged_fact_counts;
-    MVMint32        i, orig_inlines, total_inlines, orig_deopt_addrs;
+    MVMint32        i, j, orig_inlines, total_inlines, orig_deopt_addrs,
+                    orig_deopt_pea_mat_infos, impl_deopt_idx;
     MVMuint32       total_handlers = inliner->num_handlers + inlinee->num_handlers + 1;
     MVMSpeshBB     *inlinee_first_bb = NULL, *inlinee_last_bb = NULL;
     MVMuint8        may_cause_deopt = 0;
+    MVMSpeshBB *bb;
+    MVM_VECTOR_DECL(MVMSpeshOperand, regs_for_deopt);
 
-    /* If the inliner and inlinee are from different compilation units, we
-     * potentially have to fix up extra things. */
+    /* If the inliner and inlinee are from different compilation units or
+     * HLLs, we potentially have to fix up extra things. */
     MVMint32 same_comp_unit = inliner->sf->body.cu == inlinee->sf->body.cu;
-
     MVMint32 same_hll = same_comp_unit || inliner->sf->body.cu->body.hll_config ==
             inlinee_sf->body.cu->body.hll_config;
 
+    /* Gather all of the registers that have a deopt usage at the point of
+     * the invoke. We'll need to distribute deopt usages inside of the
+     * inline to them for correctness of later analyses that use the deopt
+     * usage information. */
+    impl_deopt_idx = return_deopt_idx(tc, invoke_ins);
+    MVM_VECTOR_INIT(regs_for_deopt, 0);
+    for (i = 0; i < inliner->sf->body.num_locals; i++) {
+        MVMint32 vers = inliner->fact_counts[i];
+        for (j = 0; j < vers; j++) {
+            MVMSpeshDeoptUseEntry *due = inliner->facts[i][j].usage.deopt_users;
+            while (due) {
+                if (due->deopt_idx == impl_deopt_idx) {
+                    MVMSpeshOperand o;
+                    o.reg.orig = i;
+                    o.reg.i = j;
+                    MVM_VECTOR_PUSH(regs_for_deopt, o);
+                }
+                due = due->next;
+            }
+        }
+    }
+
     /* Renumber the locals, lexicals, and basic blocks of the inlinee; also
-     * re-write any indexes in annotations that need it. */
-    MVMSpeshBB *bb = inlinee->entry;
+     * re-write any indexes in annotations that need it. Since in this pass
+     * we identify all deopt points, we also take the opportunity to add
+     * deopt users to all registers that have deopt usages thanks to the
+     * invoke. */
+    bb = inlinee->entry;
     while (bb) {
         MVMSpeshIns *ins = bb->first_ins;
         while (ins) {
@@ -452,6 +574,10 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                 case MVM_SPESH_ANN_DEOPT_INLINE:
                 case MVM_SPESH_ANN_DEOPT_OSR:
                     ann->data.deopt_idx += inliner->num_deopt_addrs;
+                    tweak_guard_deopt_idx(ins, ann);
+                    for (i = 0; i < MVM_VECTOR_ELEMS(regs_for_deopt); i++)
+                        MVM_spesh_usages_add_deopt_usage_by_reg(tc, inliner,
+                                regs_for_deopt[i], ann->data.deopt_idx);
                     break;
                 case MVM_SPESH_ANN_INLINE_START:
                 case MVM_SPESH_ANN_INLINE_END:
@@ -476,6 +602,14 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                 rewrite_outer_lookup(tc, inliner, ins, inliner->num_locals,
                     MVM_OP_sp_getlexvia_ins, code_ref_reg);
             }
+            else if (opcode == MVM_OP_sp_bindlex_in && ins->operands[0].lex.outers > 0) {
+                rewrite_outer_bind(tc, inliner, ins, inliner->num_locals,
+                    MVM_OP_sp_bindlexvia_in, code_ref_reg);
+            }
+            else if (opcode == MVM_OP_sp_bindlex_os && ins->operands[0].lex.outers > 0) {
+                rewrite_outer_bind(tc, inliner, ins, inliner->num_locals,
+                    MVM_OP_sp_bindlexvia_os, code_ref_reg);
+            }
             else if (opcode == MVM_OP_getlex && ins->operands[1].lex.outers > 0) {
                 MVMuint16 outers = ins->operands[1].lex.outers;
                 MVMStaticFrame *outer = inlinee_sf;
@@ -551,17 +685,7 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
             inlinee_last_bb = bb;
         bb = bb->linear_next;
     }
-
-    /* If we saw that we may deopt anywhere, mark the basic blocks of the
-     * inline as maybe deopting; this prevents us doing too aggressive
-     * optimizations within them. */
-    if (may_cause_deopt) {
-        bb = inlinee->entry;
-        while (bb) {
-            bb->inlined_may_cause_deopt = 1;
-            bb = bb->linear_next;
-        }
-    }
+    MVM_VECTOR_DESTROY(regs_for_deopt);
 
     /* Link inlinee BBs into the linear next chain. */
     bb = invoke_bb->linear_next;
@@ -581,7 +705,42 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         MVM_spesh_manipulate_add_successor(tc, inliner, inliner->entry, move);
     }
 
-    /* Merge facts. */
+    /* Merge materialization deopt info (for scalar-replaced entities). */
+    orig_deopt_pea_mat_infos = MVM_VECTOR_ELEMS(inliner->deopt_pea.materialize_info);
+    for (i = 0; i < MVM_VECTOR_ELEMS(inlinee->deopt_pea.materialize_info); i++) {
+        MVMSpeshPEAMaterializeInfo mi_orig = inlinee->deopt_pea.materialize_info[i];
+        MVMSpeshPEAMaterializeInfo mi_new;
+        mi_new.stable_sslot = mi_orig.stable_sslot + inliner->num_spesh_slots;
+        mi_new.num_attr_regs = mi_orig.num_attr_regs;
+        if (mi_new.num_attr_regs) {
+            mi_new.attr_regs = MVM_malloc(mi_new.num_attr_regs * sizeof(MVMuint16));
+            for (j = 0; j < mi_new.num_attr_regs; j++)
+                mi_new.attr_regs[j] = mi_orig.attr_regs[j] + inliner->num_locals;
+        }
+        else {
+            mi_new.attr_regs = NULL;
+        }
+        MVM_VECTOR_PUSH(inliner->deopt_pea.materialize_info, mi_new);
+    }
+    for (i = 0; i < MVM_VECTOR_ELEMS(inlinee->deopt_pea.deopt_point); i++) {
+        MVMSpeshPEADeoptPoint dp_orig = inlinee->deopt_pea.deopt_point[i];
+        MVMSpeshPEADeoptPoint dp_new;
+        dp_new.deopt_point_idx = dp_orig.deopt_point_idx + inliner->num_deopt_addrs;
+        dp_new.materialize_info_idx = dp_orig.materialize_info_idx + orig_deopt_pea_mat_infos;
+        dp_new.target_reg = dp_orig.target_reg + inliner->num_locals;
+        MVM_VECTOR_PUSH(inliner->deopt_pea.deopt_point, dp_new);
+    }
+
+    /* Merge facts, fixing up deopt indexes in usage chains. */
+    for (i = 0; i < inlinee->num_locals; i++) {
+        for (j = 0; j < inlinee->fact_counts[i]; j++) {
+            MVMSpeshDeoptUseEntry *due = inlinee->facts[i][j].usage.deopt_users;
+            while (due) {
+                due->deopt_idx += inliner->num_deopt_addrs;
+                due = due->next;
+            }
+        }
+    }
     merged_facts = MVM_spesh_alloc(tc, inliner,
         (inliner->num_locals + inlinee->num_locals) * sizeof(MVMSpeshFacts *));
     memcpy(merged_facts, inliner->facts,
@@ -684,6 +843,7 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     inliner->inlines[total_inlines - 1].unreachable = 0;
     inliner->inlines[total_inlines - 1].deopt_named_used_bit_field =
         inlinee->deopt_named_used_bit_field;
+    inliner->inlines[total_inlines - 1].cs = cs;
     inliner->inlines[total_inlines - 1].may_cause_deopt = may_cause_deopt;
     inliner->inlines[total_inlines - 1].bytecode_size   = bytecode_size;
     inliner->num_inlines = total_inlines;
@@ -1192,7 +1352,8 @@ void MVM_spesh_inline(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     /* Merge inlinee's graph into the inliner. */
     MVMuint32 inline_boundary_handler;
     MVMSpeshBB *inlinee_last_bb = merge_graph(tc, inliner, inlinee, inlinee_sf,
-        invoke_bb, invoke_ins, code_ref_reg, &inline_boundary_handler, bytecode_size);
+        invoke_bb, invoke_ins, code_ref_reg, &inline_boundary_handler, bytecode_size,
+        call_info->cs);
 
     /* If we're profiling, note it's an inline. */
     first_ins = find_first_instruction(tc, inlinee);
