@@ -59,8 +59,6 @@ MVMuint32 MVM_string_utf16_decodestream_main(MVMThreadContext *tc, MVMDecodeStre
     MVMint32 last_accept_pos, last_was_cr;
     MVMuint32 reached_stopper;
     int low, high;
-    MVMint32  pos = cur_bytes == ds->bytes_head ? ds->bytes_head_pos : 0;
-    MVMuint8 *bytes = (unsigned char *)cur_bytes->bytes;
     /* Set to 1 to remove the BOM even when big endian or little endian are
      * explicitly specified. */
     int remove_bom = 0;
@@ -163,7 +161,7 @@ MVMuint32 MVM_string_utf16_decodestream_main(MVMThreadContext *tc, MVMDecodeStre
             last_accept_pos = pos += 2;
             total++;
             if (MVM_string_decode_stream_maybe_sep(tc, seps, value) ||
-                    stopper_chars && *stopper_chars == total) {
+                    (stopper_chars && *stopper_chars == total)) {
                 reached_stopper = 1;
                 goto done;
             }
@@ -242,7 +240,7 @@ static MVMString * MVM_string_utf16_decode_main(MVMThreadContext *tc,
     }
 
     if (bytes % 2) {
-        MVM_exception_throw_adhoc(tc, "Malformed UTF-16; odd number of bytes");
+        MVM_exception_throw_adhoc(tc, "Malformed UTF-16; odd number of bytes (%"PRIu64")", (MVMuint64)bytes);
     }
 
     utf16_end = utf16 + bytes;
@@ -338,9 +336,9 @@ char * MVM_string_utf16_encode_substr_main(MVMThreadContext *tc, MVMString *str,
 #endif
     /* must check start first since it's used in the length check */
     if (start < 0 || start > strgraphs)
-        MVM_exception_throw_adhoc(tc, "start out of range");
+        MVM_exception_throw_adhoc(tc, "start (%"PRId64") out of range (0..%"PRIu32")", start, strgraphs);
     if (start + lengthu > strgraphs)
-        MVM_exception_throw_adhoc(tc, "length out of range");
+        MVM_exception_throw_adhoc(tc, "length (%"PRId64") out of range (0..%"PRIu32")", length, strgraphs);
 
     if (replacement)
         repl_bytes = (MVMuint8 *) MVM_string_utf16_encode_substr(tc,
