@@ -235,9 +235,28 @@ MVMString * MVM_string_utf8_decode(MVMThreadContext *tc, const MVMObject *result
                     else /* non-line ending codepoint */
                         col++;
                     break;
-                case UTF8_REJECT:
+                case UTF8_REJECT: {
+                    size_t error_pos = orig_bytes - bytes;
                     MVM_free(buffer);
-                    MVM_exception_throw_adhoc(tc, "Malformed UTF-8 at line %u col %u", line, col);
+                    /* Different error messages for the first few bytes, so that
+                     * we print as much context as possible */
+                    if (error_pos >= 3) {
+                        unsigned char a = orig_utf8[error_pos - 2], b = orig_utf8[error_pos - 1], c = orig_utf8[error_pos];
+                        MVM_exception_throw_adhoc(tc, "Malformed UTF-8 near bytes %02hhx %02hhx %02hhx at line %u col %u", a, b, c, line, col);
+                    }
+                    else if (error_pos == 2) {
+                        unsigned char a = orig_utf8[error_pos - 1], b = orig_utf8[error_pos];
+                        MVM_exception_throw_adhoc(tc, "Malformed UTF-8 near bytes %02hhx %02hhx at line %u col %u", a, b, line, col);
+                    }
+                    else if (error_pos == 1) {
+                        unsigned char a = orig_utf8[error_pos];
+                        MVM_exception_throw_adhoc(tc, "Malformed UTF-8 near byte %02hhx at line %u col %u", a, line, col);
+                    }
+                    else {
+                        MVM_exception_throw_adhoc(tc, "Malformed UTF-8 at line %u col %u", line, col);
+                    }
+                    break;
+                }
                 }
             }
             MVM_free(buffer);
@@ -303,6 +322,24 @@ MVMString * MVM_string_utf8_decode_strip_bom(MVMThreadContext *tc, const MVMObje
     return MVM_string_utf8_decode(tc, result_type, utf8, bytes);
 }
 
+static void encoding_error(MVMThreadContext *tc, char *bytes, int error_pos) {
+    if (error_pos >= 3) {
+        unsigned char a = bytes[error_pos - 2], b = bytes[error_pos - 1], c = bytes[error_pos];
+        MVM_exception_throw_adhoc(tc, "Malformed UTF-8 near bytes %02hhx %02hhx %02hhx", a, b, c);
+    }
+    else if (error_pos == 2) {
+        unsigned char a = bytes[error_pos - 1], b = bytes[error_pos];
+        MVM_exception_throw_adhoc(tc, "Malformed UTF-8 near bytes %02hhx %02hhx", a, b);
+    }
+    else if (error_pos == 1) {
+        unsigned char a = bytes[error_pos];
+        MVM_exception_throw_adhoc(tc, "Malformed UTF-8 near byte %02hhx", a);
+    }
+    else {
+        MVM_exception_throw_adhoc(tc, "Malformed UTF-8");
+    }
+}
+
 /* Decodes using a decodestream. Decodes as far as it can with the input
  * buffers, or until a stopper is reached. */
 MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds,
@@ -383,11 +420,12 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                     lag_last_accept_pos = pos;
                     break;
                 }
-                case UTF8_REJECT:
+                case UTF8_REJECT: {
                     MVM_free(buffer);
-                    MVM_exception_throw_adhoc(tc, "Malformed UTF-8");
+                    encoding_error(tc, bytes, pos - 1);
                     break;
                 }
+                }
             }
 
             while (pos < cur_bytes->length) {
@@ -432,11 +470,12 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                     lag_last_accept_pos = pos;
                     break;
                 }
-                case UTF8_REJECT:
+                case UTF8_REJECT: {
                     MVM_free(buffer);
-                    MVM_exception_throw_adhoc(tc, "Malformed UTF-8");
+                    encoding_error(tc, bytes, pos - 1);
                     break;
                 }
+                }
             }
 
             /* If we fall out of the loop and have a lagged codepoint, but
@@ -486,11 +525,12 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                     }
                     break;
                 }
-                case UTF8_REJECT:
+                case UTF8_REJECT: {
                     MVM_free(buffer);
-                    MVM_exception_throw_adhoc(tc, "Malformed UTF-8");
+                    encoding_error(tc, bytes, pos - 1);
                     break;
                 }
+                }
             }
         }
         cur_bytes = cur_bytes->next;
@@ -522,11 +562,11 @@ char * MVM_string_utf8_encode_substr(MVMThreadContext *tc,
     MVMuint64        repl_length;
 
     if (start < 0 || start > strgraphs)
-        MVM_exception_throw_adhoc(tc, "start out of range");
+        MVM_exception_throw_adhoc(tc, "start (%"PRId64") out of range (0..%"PRIu32")", start, strgraphs);
     if (length == -1)
         length = strgraphs;
     if (length < 0 || start + length > strgraphs)
-        MVM_exception_throw_adhoc(tc, "length out of range");
+        MVM_exception_throw_adhoc(tc, "length (%"PRId64") out of range (0..%"PRIu32")", length, strgraphs);
 
     if (replacement)
         repl_bytes = (MVMuint8 *) MVM_string_utf8_encode_substr(tc,
