@@ -576,9 +576,9 @@ char * MVM_string_windows125X_encode_substr(MVMThreadContext *tc, MVMString *str
 
     /* must check start first since it's used in the length check */
     if (start < 0 || strgraphs < start)
-        MVM_exception_throw_adhoc(tc, "start out of range");
+        MVM_exception_throw_adhoc(tc, "start (%"PRId64") out of range (0..%"PRIu32")", start, strgraphs);
     if (length < -1 || strgraphs < start + lengthu)
-        MVM_exception_throw_adhoc(tc, "length out of range");
+        MVM_exception_throw_adhoc(tc, "length (%"PRId64") out of range (-1..%"PRIu32")", length, strgraphs);
 
     if (replacement)
         repl_bytes = (MVMuint8 *) MVM_string_windows125X_encode_substr(tc,
